/*
 * nciddhangup.c - This file is part of ncidd.
 *
 * Copyright (c) 2005-2021
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncidd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncidd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncidd.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ncidd.h"

/* globals */
int pickup = 1;
int regex;
const char * blacklist = BLACKLIST;
const char * listname;
const char * whitelist = WHITELIST;
const char * audiofmt  = XPRESS;

struct list *blkHead = NULL, *blkCurrent = NULL,
            *whtHead = NULL, *whtCurrent = NULL;

int doList(const char * filename, list_t **listHead, list_t **listCurrent);
int onList(const char * namep, const char * nmbrp, int flag, list_t **listHead);
int hangupCall(int mode, const char * recording);
int checkWhitelist(const char * namep, const char * nmbrp);
int checkBlacklist(const char * namep, const char * nmbrp);
int onBlackWhite (const char * namep, const char * nmbrp);

void addEntry(const char * fnptr, const char * inptr, char * wdptr, int lc, list_t **listHead, list_t **listCurrent);
void nextEntry(list_t **listHead, list_t **listCurrent);
void rmEntries(list_t **listHead, list_t **listCurrent);

#ifndef __CYGWIN__
    extern char *strsignal();
#endif

/*
 * Process the blacklist or whitelist file
 * Returns: 0 - no errors
 *          # of errors
 */
int doList(const char * filename, list_t **listHead, list_t **listCurrent)
{
    char input[BUFSIZ], word[BUFSIZ], msgbuf[BUFSIZ], *ptr;
    const char * inptr;
    struct list *node, *nextnode;
    int lc, i, bl;
    int max_temp = 0, max_entry = 0;
    FILE *fp;

    bl = !strcmp(blacklist, filename) ? 1 : 0;

    if ((fp = fopen(filename, "r")) == NULL)
    {
        logMsg(LEVELE, "%s file missing: %s\n",
                bl ? "Blacklist" : "Whitelist", filename);
        return 1;
    }

    /* read each line of file, one line at a time */
    for (lc = 1; fgets(input, BUFSIZ, fp) != NULL; lc++)
    {
        /* line containing only <NL> or <CR><NL> or is a comment line */
        if (input[0] == '\n' || input[0] == '\r' || input[0] == '#') continue;

        /* remove '\n' and '\r' if there */
        if ((ptr = strchr(input, (int) '\n'))) *ptr = '\0';
        if ((ptr = strchr(input, (int) '\r'))) *ptr = '\0';

        inptr = getWord(filename, input, word, lc);
        if (inptr == 0) continue;

        /* get search strings on line */
        addEntry(filename, inptr, word, lc, listHead, listCurrent);
    }
    (void) fclose(fp);
    logMsg(LEVEL1, "Processed %s file: %s\n",
            bl ? "blacklist" : "whitelist", filename);

    logMsg(LEVEL1, "%s Table:\n", bl ? "Blacklist" : "Whitelist");

    /* log the number of entries at LEVEL1 and list them at LEVEL7 */
    if (!errorStatus && *listHead)
    {
        node = *listHead;
        for (i = 0; node != NULL; i++)
        {
            nextnode = node->next;
            node = nextnode;
        }

        logMsg(LEVEL1, "    Number of Entries: %d\n", i);

        /*
         * Make the List display pretty by determining maximum column width.
         */

        node = *listHead;
        while(node != NULL)
        {
            nextnode = node->next;
            max_temp = strlen(node->entry);
            if (max_temp > max_entry) max_entry = max_temp;
            node = nextnode;
        }

        logMsg(LEVEL7, "    %-4s %s%-*s %s\n",
            "SLOT",
            "ENTRY", max_entry - (int) strlen("ENTRY") + 2, "",
            "MATCH NAME");

        logMsg(LEVEL7, "    %-4s %s%-*s %s\n",
            "----",
            "-----", max_entry - (int) strlen("ENTRY") + 2, "",
            "----------");

        node = *listHead;
        for (i = 0; node != NULL; i++)
        {
            nextnode = node->next;
            sprintf(msgbuf, "    %-4.3d \"%s\"%-*s ",
                    i,
                    node->entry,
                    max_entry - (int) strlen(node->entry), "");

            if (node->newname)
                strcat(strcat(strcat(msgbuf, "\""), node->newname), "\"");
            strcat(msgbuf, NL);

            logMsg(LEVEL7, msgbuf);
            node = nextnode;
        }
    }
    else
    {
        logMsg(LEVEL1, "    Number of Entries: 0\n");
    }

    return errorStatus;
}

/*
 * Process blacklist or whitelist file lines
 */
void addEntry(const char * fnptr, const char * inptr, char * wdptr, int lc, list_t **listHead, list_t **listCurrent)
{
    int err;
    char errbuf[80], *cptr, *eptr;

    /* process the blacklist or whitelist words */
    do
    {
        /*
         * If inQuotes = 1, the word is inside "" and is not a comment.
         * Allows an entry in the blacklist or whitelist to begin with a "#".
         */
        if (!inQuotes && *wdptr == '#')    break; /* rest of line is comment */
        nextEntry(listHead, listCurrent);
        if (strlen(wdptr) > ENTRYSIZE - 1) {
            configError(fnptr, lc, wdptr, ERRLONG);
            return;
        }
        /* store the entry to be matched */
        (*listCurrent)->entry = strdup(wdptr);
        /* cheats const char inptr here */
        if ((cptr = strstr(inptr, "#=")))
        {
            /* store the name that comes after #= for later use
             * if the caller is found in white or blacklist
             */
            cptr += 2;

            /* remove any leading spaces */
            while (*cptr && (*cptr == ' ' || *cptr == '\t')) ++cptr;

            /* remove a trailing comment from the #= new name */
            if ((eptr = strstr(cptr, "#"))) *eptr = '\0';
            /* remove any trailing spaces */
            eptr = cptr;
            while (*eptr) eptr++;
            for (--eptr; *eptr && (*eptr == ' ' || *eptr == '\t'); --eptr) *eptr = '\0';

            /* store the #= newname with * delimiters removed */
            (*listCurrent)->newname = strdup(starfree(cptr, CIDSIZE));
        }
        else (*listCurrent)->newname = 0;

        if (regex == 1)
        {
          /* expressions are regex */
          (*listCurrent)->preg = (regex_t*)malloc(sizeof(regex_t));
          err = regcomp((*listCurrent)->preg, wdptr, REG_EXTENDED|REG_NOSUB);
          if (err)
          {
            regerror(err, (*listCurrent)->preg, errbuf, sizeof(errbuf));
            configError(fnptr, lc, wdptr, errbuf);
          }
        }
        else if (regex == 2)
        {
          /* expressions are pcre */
          (*listCurrent)->pcreg = pcre_compile(wdptr, 0, &pcreErrorMessage, &pcreErrorOffset, NULL);

          if (pcreErrorMessage)
          {
            sprintf(errbuf, "pcre_compile failed (offset: %d), %s\n", pcreErrorOffset, pcreErrorMessage);
            configError(fnptr, lc, wdptr, errbuf);
          }
        }
        else /* (regex == 0) */
        {
          /* NCID Simple expressions */
          (*listCurrent)->ncidreg = store_bwstrmatch(wdptr);
        }
    }
    while ((inptr = getWord(fnptr, inptr, wdptr, lc)) != 0);
}

/*
 * Adds a node to the list for a new entry
 */
void nextEntry(list_t **listHead, list_t **listCurrent)
{
    list_t *node;

    if (!(node = (list_t *) malloc(sizeof(list_t)))) errorExit(-1, progname, 0);
    node->next = NULL;
    if (*listHead == NULL) *listHead = node;
    else (*listCurrent)->next = node;
    *listCurrent = node;
}

/*
 * Frees all list nodes
 */
void rmEntries(list_t **listHead, list_t **listCurrent)
{
    struct list *node = *listHead;
    struct list *nextnode;

    while (node != NULL)
    {
        nextnode = node->next;
        if (regex == 1) { regfree(node->preg); free(node->preg); }
        else if (regex == 2) pcre_free(node->pcreg);
        else free_abwstr(node->ncidreg);
        if (node->entry) free(node->entry);
        if (node->newname) free(node->newname);
        free(node);
        node = nextnode;
    }

    *listHead = NULL;
    *listCurrent = NULL;
}

/*
 * Check if call is in whitelist file
 *
 * cflag == 0 if call not in whitelist
 * cflag == 1 if call is in whitelist
 *
 * if cflag == 0 call doHangup()
 *
 */

int checkWhitelist(const char * namep, const char * nmbrp)
{
    return onList(namep, nmbrp, 0, &whtHead);
}

/*
 * Check if call is in blacklist file
 *
 * A blacklist match must never be on the whitelist.
 *
 * Must call checkWhitelist() first to make sure call
 * is not on the whitelist.
 */

int checkBlacklist(const char * namep, const char * nmbrp)
{
    return onList(namep, nmbrp, 0, &blkHead);
}


/*
 * Hangup Call -- go off hook
 * return = 0 modem is off hook
 * return != 0 error
 */
int hangupCall_offhook(int mode)
{
    int ret = 0, pret;

    /* try to create the tty lock file.
       if it's already present, don't try to use the modem */
    if (makeLockfile()) return 1;


    /* put tty port in raw mode */
    if (!(ret = tcsetattr(ttyfd, TCSANOW, &rtty) < 0))
    {
        /* Send AT to get modem OK after switch to raw mode */
        (void) initModem(GETOK, HANGUPTRY);

        if (mode == 1)    /* hangup mode */
        {
            /* Pick up the call */
            ret = initModem(PICKUP, HANGUPTRY);
            logMsg(LEVEL5, "hangup mode %d: PICKUP sent, return code is %d\n",mode,ret);
        }
        else if (mode == 2)   /* FAX hangup mode */
        {
            /* Set FAX mode */
            ret = initModem(FAXMODE, HANGUPTRY);    /* AT+FCLASS=1 */
            logMsg(LEVEL5, "hangup mode %d: FAXMODE sent, return code is %d\n",mode,ret);

            if (pickup)
            {
                /* PICKUP required for some (mostly USB) modems */
                ret = initModem(PICKUP, HANGUPTRY);
                logMsg(LEVEL5, "hangup mode %d: PICKUP sent, return code is %d\n",mode,ret);
            }
        }
        else /* Voice Hangup Mode (mode == 3) */
        {
          /*
           *  Set Voice Mode:             AT+FCLASS=8  (return == 0) OK
           *  speaker volume normal:      AT+VGT=128   (return == 0) OK
           *  Select Compress Method:     AT+VSM=128   (return == 0) OK
           *  Answer Call:                AT+VLS=1     (return == 0) OK
           *  Select Voice Transfer Mode: AT+VTX       (return == 1) CONNECT
           *  send voice data to modem
           *  end of voice data:          <DLE><ETX>   (return == 0) OK
           *  Go on hook in data mode:   ATH          (return == 0) OK
           *
           *  NOTE: AT+VSM=128 is the default used for the CONEXANT
           *        CX93001 chipset used in a lot of modems.
           */

            ret = initModem(VOICEMODE, HANGUPTRY);  /* AT+FCLASS=8 */
            logMsg(LEVEL5, "hangup mode %d: VOICEMODE sent, return code is %d\n",mode,ret);

            ret = initModem(SPKRVOL, HANGUPTRY);    /* AT+VGT=128 */
            ret = initModem(audiofmt, HANGUPTRY);   /* AT+VSM=130 */
            ret = initModem(ANSCALL, HANGUPTRY);    /* AT+VLS=1   */
            ret = initModem(NOECHO, HANGUPTRY);     /* Echo Off   */
            ret = initModem(VOICEXFR, HANGUPTRY);   /* AT+VTX     */
            if (ret == 1 ) ret = 0; /* returned CONNECT */
        }
    }

    if (ret)
    {
       /* unable to go off-hook, try to hang up and leave */
        pret = initModem(HANGUP, HANGUPTRY);
        logMsg(LEVEL5, "hangup mode %d: HANGUP sent, return code is %d\n", mode, pret);

        /* if Hangup does not indicate OK try 1 more time  */
        if (pret != 0)
        {
            /* Send AT to get modem OK */
            pret = initModem(GETOK, HANGUPTRY);
            logMsg(LEVEL5, "hangup mode %d try 2: GETOK sent, return code is %d\n", mode, pret);

            /* Send HANGUP to get modem OK */
            pret = initModem(HANGUP, HANGUPTRY);
            logMsg(LEVEL5, "hangup mode %d try 2: HANGUP sent, return code is %d\n", mode, pret);
        }

        if (mode != 1)    /* FAX or VOICE hangup mode */
        {
            /* set data mode */
            pret = initModem(DATAMODE, HANGUPTRY);
            logMsg(LEVEL5, "hangup mode %d: DATAMODE sent, return code is %d\n", mode, pret);
        }

        /* take tty port out of raw mode */
        (void) tcsetattr(ttyfd, TCSAFLUSH, &ntty);


        /* remove TTY port lockfile */
        removeLockfile();
    }

    return ret;
}


int hangupCall(int mode, const char * recording)
{
    int ret = 0, msgfd, rcnt, wcnt, tcnt = 0;
    unsigned int hangupdelay;
    char voicebuf[BUFSIZ];

    (void) wcnt;

    if (mode == 1)    /* hangup mode */
    {
        /* set off-hook delay to make sure to hangup call */
        hangupdelay = HANGUPDELAY;
    }
    else if (mode == 2)   /* FAX hangup mode */
    {
        /* generate FAX tones */
        ret = initModem(FAXANS, HANGUPTRY);
        logMsg(LEVEL5, "hangup mode %d: FAXANS sent, return code is %d\n", mode, ret);

        /* set off-hook delay so caller hears annoying fax tones */
        hangupdelay = FAXDELAY;
    }
    else /* Voice Hangup Mode (mode == 3) */
    {
       /*  send voice data to modem
        *  end of voice data:          <DLE><ETX>   (return == 0) OK
        *  Go on hook in data mode:   ATH          (return == 0) OK
        */

        /* send voice file */
        if ((msgfd = open(recording, O_RDONLY)) == -1)
        {
            logMsg(LEVELE, "%d: %s\n", msgfd, strerror(errno));
            ret = initModem(DLEETX, VOICETRY); /* <DLE><ETX> */
            hangupdelay = HANGUPDELAY;
        }
        else
        {
            /*
             * rcnt == BUFSIZ or remaining bytes of file at end
             */
            logMsg(LEVEL3, "Sending file %s, to modem in %d byte chunks\n", recording, BUFSIZ);
            do
            {
                rcnt = read(msgfd, voicebuf, BUFSIZ);
                if (rcnt > 0 ) wcnt = write(ttyfd, voicebuf, rcnt);
                tcnt = tcnt + rcnt;
            }
            while (rcnt > 0);
            close(msgfd);
            logMsg(LEVEL3, "Sent %d bytes to modem for announcement\n", tcnt);
            ret = initModem(DLEETX, VOICETRY); /* <DLE><ETX> */

            /* set delay after announcement */
            hangupdelay = VOICEDELAY;
        }
    }

    /* off-hook delay for all hangup modes */
    usleep(hangupdelay * 1000000);
    logMsg(LEVEL3, "off-hook for %d %s\n", hangupdelay, hangupdelay == 1 ? "second" : "seconds");

    /* Hangup the call */
    ret = initModem(HANGUP, HANGUPTRY);
    logMsg(LEVEL5, "hangup mode %d: HANGUP sent, return code is %d\n", mode, ret);

    /* if Hangup does not indicate OK try 1 more time  */
    if (ret != 0)
    {
        /* Send AT to get modem OK */
        ret = initModem(GETOK, HANGUPTRY);
        logMsg(LEVEL5, "hangup mode %d try 2: GETOK sent, return code is %d\n", mode, ret);

        /* Send HANGUP to get modem OK */
        ret = initModem(HANGUP, HANGUPTRY);
        logMsg(LEVEL5, "hangup mode %d try 2: HANGUP sent, return code is %d\n", mode, ret);
    }

    if (mode != 1)    /* FAX or VOICE hangup mode */
    {
        /* set data mode */
        ret = initModem(DATAMODE, HANGUPTRY);
        logMsg(LEVEL5, "hangup mode %d: DATAMODE sent, return code is %d\n", mode, ret);
    }

    /* take tty port out of raw mode */
    (void) tcsetattr(ttyfd, TCSAFLUSH, &ntty);


    /* remove TTY port lockfile */
    removeLockfile();

    return ret;
}

/*
 * Compare blacklist or whitelist strings to name and number
 *
 * If "flag" is zero, log a match message
 *   Return = 1  if the number or matches
 *   Return = 0  if no match
 *
 * If "flag" is one, never log a match message
 *   Return = 3  if the number matches
 *   Return = 1  if the name matches
 *   Return = 0  if no match
 */
int onList(const char * namep, const char * nmbrp, int flag, list_t **listHead)
{
    int ret = 0, i, pos, nbrMatch = 0;
    char *ptr, *listptr;
    list_t *node, *nextnode;
    void * regexp = NULL;

    listptr = (listHead == &blkHead ? "Blacklist" : "Whitelist");
    node = *listHead;
    if (!node)
    {
        logMsg(LEVEL3, "%s empty\n", listptr);
        return ret;
    }

    logMsg(LEVEL4, "Begin: Search %s file nmbr=%s name=%s [%s]\n", listptr, nmbrp==0?"0":nmbrp, namep==0?"0":namep, strdate(ONLYTIME));

    for (i = 0; node != 0; i++)
    {
        nextnode = node->next;
        ptr = node->entry;
        pos = i;

        if (regex == 2)
        {
           /* perl regular expression */
           regexp = (void*)node->pcreg;
        }
        else if (regex == 1)
        {
           /* regex regular expression */
           regexp = (void*)node->preg;
        }
        else /* (regex == 0) */
        {
           /* ncid simple expression */
           regexp = (void*)node->ncidreg;
        }

        if (strmatch(namep, regexp)) { ret = 1; break; }
        if (strmatch(nmbrp, regexp)) { ret = 1, nbrMatch = 2; break; }

        node = nextnode;
    }

    logMsg(LEVEL3, "Checked %s for match flag=%d ret=%d nmbrmatch=%d\n",
        listptr, flag, ret, nbrMatch);
    if (flag) return (ret + nbrMatch);
    if (ret)
    {
        listname = node->newname;
        logMsg(LEVEL3, "%s Match #%.2d: %s    number: %s    name: %s\n",
                listptr, pos, ptr, nmbrp, namep);
    }

    return ret;
}

/*
 * Check for a name or a number being in the black or white list
 *
 * Return = 0  not on either list
 * Return = 1  name is on the blacklist and not on the whitelist
 * Return = 5  number is on the blacklist and not on the whitelist
 * Return = 2  name is on the whitelist; may or may not be on the blacklist
 * Return = 6  number is on the whitelist; may or may not be on the blacklist
 *
 * Bits 0 and 1:
 *      0 - On neither list
 *      1 - Name or number on blacklist
 *      2 - Name or number on whitelist
 *
 * Bit 2:
 *      0 - Name is on the list
 *      1 - Number is on the list
 */
int onBlackWhite (const char * namep, const char * nmbrp)
{
    int ret;

    /* ret will be either 0, 1, or 3 */
    if ((ret = onList(namep, nmbrp, 1, &whtHead))) return 2 * ret;
    if ((ret = onList(namep, nmbrp, 1, &blkHead))) return 2 * ret - 1;
    return 0;
}


/*
 * remove existing whitelist, blacklist entries to free memory used
 */
void rmBWlists(void)
{
    rmEntries(&blkHead, &blkCurrent);

    rmEntries(&whtHead, &whtCurrent);
}

/*
 * read the whitelist, blacklist file entries into memory
 */
void readBWlists(void)
{
    /* reload blacklist file but quit on error */
    if (doList(blacklist, &blkHead, &blkCurrent)) errorExit(-114, 0, 0);

    /* reload whitelist file but quit on error */
    if (doList(whitelist, &whtHead, &whtCurrent)) errorExit(-114, 0, 0);
}
