/*
 * ncidd.c - This file is part of ncidd.
 *
 * Copyright (c) 2005-2021
 * by John L. Chmielewski <jlc@users.sourceforge.net>
 *
 * ncidd is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ncidd is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ncidd.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ncidd.h"
#include "nciddPN.h"

/* globals */
const char * cidlog   = CIDLOG;
const char * datalog  = DATALOG;
const char * country  = DEFAULT_COUNTRY;
const char * language = DEFAULT_LANGUAGE;
const char * initstr  = INITSTR;
const char * initcid  = INITCID1;
const char * logfile  = LOGFILE;
const char * serial_lineid   = ONELINE;
const char * interfaceaddr   = "";
const char * pidfile;
const char * lockfile;        /* the modem lock file name */

char * progname;
char hostname[HOST_NAME_MAX];
int modemModes;
int debug, setcid, sendlog, sendinfo, cidnoname;
int modem_locked;
int mainsock;       /* our listen socket for client connections */
int port = PORT;    /* listen port number */
volatile int HUP_flag, USR1_flag, USR2_flag, TERM_flag; /* interrupt flags */

/* modem and per-call info */
const char *  ttyport  = TTYPORT;     /* the modem device name */
const char * TTYspeed;
int ttyspeed   = TTYSPEED;
int ttyfd = -1;            /* the modem file descriptor */
int pollpos = -1;          /* remember modem pollfd during lock */
int pollevents;            /* remember modem pollfd during lock */
int ringcnt;               /* one ringie-dingie ... */
int widflag;               /* modem supports call waiting (WID) */
time_t ringtimer = 0;      /* to detect end of ringing */
time_t modemlosttimer = 0; /* to try modem init again */
time_t abw_lastread  = 0;  /* time of last ABW reads */
time_t checkABWtimer = 0; /* to check for manual ABW changes */
int modemlostcount = 0;

int cidinput = -1;
int hangup;
int hupmode;
int nomodem, noserial; /* deprecated options, use cidinput */
int clocal, gencid = 0;
int verbose = 1;
int ignore1;
int OSXlaunchd, sigpipe, sendlog_override, dialOUT;
long unsigned int cidlogmax = LOGMAX;
pid_t pid;

const char * announce = RECORDING;
char annpath[BUFSIZ];
const char * hupname = HUPEXTENSION;
char huppath[BUFSIZ];
const char * huprmd;
char rmdpath[BUFSIZ];

char tmpIPaddr[MAXIPBUF];
char tmpHostName[MAXIPBUF];
char modembuf[BUFSIZ];

struct pollfd polld[MAXCONNECT];
struct termios otty, rtty, ntty;
FILE *logptr;
time_t hanguptimer = 0;
int dialclientpos = -1;
int dialdelay = DIAL_TIMEOUT;
int modem_lidx;
#define MAXMODEMLINES 20
char * modem_ilines[MAXMODEMLINES];

char LineIDS[CIDSIZE * 4]; /* room for 4 lineid's */

/* IPinfo[pos] is for same client/gateway as in polld[pos] */
struct ipinfo {
    char addr[MAXIPBUF];
    char name[MAXIPBUF];
    char ident[MAXIPBUF];
    int  dialout;
} IPinfo[MAXCONNECT];

/* ack[pos] is for same client/gateway as in polld[pos] */
int ack[MAXCONNECT]; /* only for clients */

/* modem and per-call info */
struct cid
{
    int status;
    int calltype;   /* CID HUP OUT etc */
    int cidsent;    /* used to give gencid on 2nd ring */
    char ciddate[CIDSIZE];
    char cidtime[CIDSIZE];
    char cidnmbr[CIDSIZE];
    char cidname[CIDSIZE];
    char cidmesg[CIDSIZE];
    char cidline[CIDSIZE];
    char * infoBuffer;
};
struct cid modem_cid = {0, CID, 0, "", "", "", "", NOMESG, ONELINE};
struct cid gwcid = {0, CID, 0, "", "", "", "", NOMESG, "-"};


/* this is filled by getINFO() to return values from MSG: and NOT: */
struct mesg
{
    char date[CIDSIZE];
    char time[CIDSIZE];
    char nmbr[CIDSIZE];
    char name[CIDSIZE];
    char line[CIDSIZE];
    char type[CIDSIZE];
} mesg;

/* used in handleGWCallinfo() while building the cidcall line */
struct end
{
    char htype[CIDSIZE];
    char ctype[CIDSIZE];
    char  date[CIDSIZE];
    char  time[CIDSIZE];
    char scall[CIDSIZE];
    char ecall[CIDSIZE];
    char  line[CIDSIZE];
    char  nmbr[CIDSIZE];
    char  name[CIDSIZE];
} endcall;

struct setword setword[] = {
    /* char *word; int type; char **buf; int *value; int min; int max */
    {"announce",   WORDSTR,  &announce,  0,          0,       0},
    {"audiofmt",   WORDSTR,  &audiofmt,  0,          0,       0},
    {"blacklist",  WORDSTR,  &blacklist, 0,          0,       0},
    {"cidalias",   WORDSTR,  &cidalias,  0,          0,       0},
    {"cidinput",   WORDNUM,  0,          &cidinput,  0,       3},
    {"cidlog",     WORDSTR,  &cidlog,    0,          0,       0},
    {"cidlogmax",  WORDNUM,  0,          (int *) &cidlogmax, 1, LOGMAXNUM},
    {"cidnoname",  WORDNUM,  0,          &cidnoname, OFF,     ON}, /*unused*/
    {"country",    WORDSTR,  &country,   0,          0,       0},
    {"datalog",    WORDSTR,  &datalog,   0,          0,       0},
    {"dialdelay",  WORDNUM,  0,          &dialdelay, 2,       10},
    {"hangup",     WORDNUM,  0,          &hangup,    0,       3},
    {"hupmode",    WORDNUM,  0,          &hupmode,   0,       3},
    {"hupname",    WORDSTR,  &hupname,   0,          0,       0},
    {"huprmd",     WORDSTR,  &huprmd,    0,          0,       0},
    {"gencid",     WORDNUM,  0,          &gencid,    OFF,     ON},
    {"ignore1",    WORDNUM,  0,          &ignore1,   OFF,     ON},
    {"initcid",    WORDSTR|WORDFLAG, &initcid, &setcid, 0,    0},
    {"initstr",    WORDSTR,  &initstr,   0,          0,       0},
    {"ifaddr",     WORDSTR,  &interfaceaddr,    0,          0,       0},
    {"language",   WORDSTR,  &language,  0,          0,       0},
    {"lineid",     WORDSTR,  &serial_lineid,    0,          0,       0},
    {"lockfile",   WORDSTR,  &lockfile,  0,          0,       0},
    {"nomodem",    WORDNUM,  0,          &nomodem,   OFF,     ON},
    {"noserial",   WORDNUM,  0,          &noserial,  OFF,     ON},
    {"pickup",     WORDNUM,  0,          &pickup,    OFF,     ON},
    {"pidfile",    WORDSTR,  &pidfile,   0,          0,       0},
    {"port",       WORDNUM,  0,          &port,      0,       0},
    {"regex",      WORDNUM,  0,          &regex,     0,       2},
    {"ttyclocal",  WORDNUM,  0,          &clocal,    OFF,     ON},
    {"ttyport",    WORDSTR,  &ttyport,   0,          0,       0},
    {"ttyspeed",   WORDSTR,  &TTYspeed,  0,          0,       0},
    {"nanp_format",  WORDNUM,  0,          &nanp_format, 0,       3},
    {"verbose",    WORDNUM,  0,          &verbose,   1,       MAXLEVEL - 2},
    {"whitelist",  WORDSTR,  &whitelist, 0,          0,       0},
    {0,            0,        0,          0,          0,       0}
};

/* associate variables for config file "send cidlog" and "send cidinfo" */
struct sendclient sendclient[] = {
    /* char *word; int *value */
    {  "cidlog",   &sendlog   },
    {  "cidinfo",  &sendinfo  },
    {  0,          0          }
};

/* All line labels of interest to a client, new types must be added */
/* sorted alphabetically here for readability */
const char *lineTags[] =
{
    BLKLINE,
    CIDLINE,
    ENDLINE,
    HUPLINE,
    MSGLINE,
    MWILINE,
    NOTLINE,
    OUTLINE,
    PIDLINE,
    PUTLINE,
    RIDLINE,
    RLYLINE,
    WIDLINE,
    NULL
};

/* server lines sent to clients */
/* sorted alphabetically here for readability */
char *serverTags[] =
{
    "+BLK:",
    "+CID:",
    "+CIDINFO:",
    "+END:",
    "+HUP:",
    "+MSG:",
    "+MWI:",
    "+NOT:",
    "+OUT:",
    "+PID:",
    "+PUT:",
    "+RID:",
    "+RLY:",
    "+WID:",
    NULL
};

/* List of popular country codes as returned by QUERYATGCI ("AT+GCI?")
 * http://doc.slitaz.org/en:handbook:pstn:countries
 * http://www.cisco.com/c/en/us/td/docs/routers/access/modem/AT/wic/command/reference/atwic/atwic3.html
 *
 * Country codes have their origin based on fascimile standards.
 * The definitive guide for all country codes is maintained by the
 * International Telegraph Union (ITU), a specialized agency of the
 * United Nations. Not all modem manufacturers use a consistent list
 * of country codes and not all modem manufactures even use the
 * complete list. The ITU's complete list is available here:
 *     http://www.itu.int/dms_pub/itu-t/opb/sp/T-SP-T.35-2012-OAS-PDF-E.pdf
 */
char *modemGCI[] =
{
     "00 Japan",
     "04 Germany",
     "09 Australia",
     "0A Austria",
     "0F Belgium",
     "16 Brazil",
     "1B Bulgaria",
     "20 Canada",
     "26 China",
     "2D Cyprus",
     "2E Czech Republic",
     "31 Denmark",
     "3C Finland",
     "3D France",
     "42 Germany",
     "46 Greece",
     "50 Hong Kong",
     "51 Hungary",
     "52 Iceland",
     "53 India",
     "57 Ireland",
     "58 Israel",
     "59 Italy",
     "61 Korea",
     "68 Liechtenstein",
     "69 Luxembourg",
     "6C Malaysia",
     "73 Mexico",
     "7B Netherlands",
     "7E New Zealand",
     "82 Norway",
     "89 Philippines",
     "8A Poland",
     "8B Portugal",
     "8C Singapore",
     "9F South Africa",
     "A0 Spain",
     "A5 Sweden",
     "A6 Switzerland",
     "A9 Thailand",
     "AE Turkey",
     "B4 United Kingdom",
     "B5 United States",
     "B8 Russia",
     "F6 TBR-21(Default)",
     "F7 Lithuania",
     "F8 Latvia",
     "F9 Estonia",
     "FB Slovakia",
     "FC Slovenia",
     "FE Taiwan",
     NULL
};

#ifndef __CYGWIN__
    extern char *strsignal();
#endif

void writeClients(const char * inbuf);
void writeLog(const char * logf, const char * logbuf);
void cleanup(void);
void getINFO(const char * bufptr);
void getField(const char * bufptr, const char * field_name, char * mesgptr, const char * noval);
void normalExit(void);
void showConnected(void);
void checkHangupMode(int *mode);
void changeTTY(int raw);
void modem_warning(void);
void doIDENT(int pos, const char * buf);
void initPoll(void);

void openandCheckSerial(void);
void checkSerialOptions(void);
void set_ttyspeed(void);
void make_lockfile_name(void);
void sendACK(int pos, const char * buf);
void sendCall(const char * linelabel, const struct cid * scid, const char * nameptr);
void doPoll(int events);
void sendInfo(int ring, const char * lineid);
void errorExit(int error, const char * msg, const char * arg);
void doHangupTimer(void);
void doRingTimer(void);
void doModemLockTimer(void);
void doModemLostTimer(void);
void startModemLostTimer(const char * msg);
void doCheckABWTimer(void);
void doHELLO(int pos);
void doOptLine(int pos);
void sendLog(int pos);
void reload(void);
void hex2uchar(const char * buf, unsigned char *ucbuf);
void setup_signal_handlers(void);
void USRhandler(int sig);
void TERMhandler(int sig);
void SEGVhandler(int sig);
void TERMexit(void);
void handleGWCallinfo(int pos, const char * buf);
void handleClientRequest(int pos, const char * buf);
void handleClientWork(const char * buf);
void formatCID(char ** lines);
void gatewayCID(const char * buf);
void finishCID(struct cid * fcid);
void TCIserial(const char * buf);
void xdmfCID(const char * buf);
void modemCID(const char * buf);
void checkHangupExtension(void);
int check_ABW_dates(const char * filename);
int runExternalHangup(const char ** linelabel, const struct cid * ecid, const char ** nameptr, const char ** rmdptr);

int getOptions(int argc, char *argv[]);
int CheckForLockfile(void);
int makeLockfile(void);
int removeLockfile(void);
int CheckLockfilePath(void);
int tcpOpen(void);
int openSerial(void);
int checkModem(void);
int checkModemCapabilities(void);
int setModem(void);
int initModem(const char * ptr, int maxwait);
int reAttachModem(void);
int tcpAccept(void);
int openTTY(void);
int checkVoiceFile(int *mode, char * path, const char * recording);
int replyToSock(int sd, char *buf);
int nothex(const char * mesg);
int setTTY(void);
int addPoll(int addfd);
int checkSCID(void);

char * trimWhitespace(char *sptr, char *eptr);
const char * getLineLabel(int calltype);
const char * getSpeedLabel(int ttyspeed);
unsigned char checksum(unsigned char *ptr, size_t size);

/*
 *  The main program
 */

int main(int argc, char *argv[])
{
    int events, argind, i, fd, errnum, ret;
    struct stat statbuf;
    struct utsname utsbuf;
    char msgbuf[BUFSIZ], aliasbuf[CIDSIZE];

    /* set up signal handlers for HUP, USR1, USR2, INT, etc */
    setup_signal_handlers();

    /* global containing name of program */
    progname = strrchr(argv[0], (int) '/');
    progname = progname ? progname + 1 : argv[0];

    /* get the nostname */
    if (gethostname(hostname, HOST_NAME_MAX) == -1)
    {
        logMsg(LEVEL1, "gethostname(): %s\n", strerror(errno));
    }
    hostname[HOST_NAME_MAX-1] = '\0';

    /* Get the default config file */
    config = CONFIG;

    /* process options from the command line */
    argind = getOptions(argc, argv);

    /* should not be any arguments */
    if (argc - argind != 0)
    {
        fprintf(stderr, NOOPT, progname, argv[argind]);
        fprintf(stderr, USAGE, progname);
        exit(0);
    }

    /* open or create logfile */
    logptr = fopen(logfile, "a+");
    errnum = errno;

    logMsg(LEVEL1, "Started: %s\nServer: %s %s\n%s\n",strdate(WITHSEP),
            progname, VERSION, API);

    /* uname system call information */
    if ((ret = uname(&utsbuf) == -1))
    {
        logMsg(LEVELE, "uname system call failed with errno = %d\n", errno);
    }
    else
    {
        logMsg(LEVEL1, "Sysname:  %s\nNodename: %s\nMachine:  %s\nRelease:  %s\nVersion:  %s\n",
             utsbuf.sysname, utsbuf.nodename, utsbuf.machine, utsbuf.release, utsbuf.version);
    }

    /* log command line and any options on separate lines */
    sprintf(msgbuf, "Command line: %s", argv[0]);
    for (i = 1; i < argc; i++)
    {
        if (*argv[i] == '-')
            strcat(strcat(msgbuf, "\n              "), argv[i]);
        else strcat(strcat(msgbuf, " "), argv[i]);
    }
    strcat(msgbuf, NL);
    logMsg(LEVEL1, msgbuf);

    /* check status of logfile */
    if (logptr)
    {
        /* logfile opened */
        logMsg(LEVEL1, "Logfile: %s\n", logfile);
    }
    else
    {
        /* logfile open failed */
        logMsg(LEVELE, "%s: %s\n", logfile, strerror(errnum));
    }

    /*
     * read config file, if present, exit on any errors
     * do not override any options set on the command line
     */
    if ((ret = doConf()))
    {
        if (ret > 0) errorExit(-104, 0, 0);
        else errorExit(-1, progname, 0);
    }

    logMsg(LEVEL1, "Verbose level: %d\n", verbose);

    /* I have CIDSIZE 256 BUFSIZ 8192, safesend clips incoming messages at 1024 */
    logMsg(LEVEL4, "CIDSIZE %d BUFSIZ %d\n", CIDSIZE, BUFSIZ);

    /* for compatibility with ncidd <= 1.6 if cidinput == -1 */
    if (cidinput < 0)
    {
        if (noserial)
        {
            if (hangup || hupmode) cidinput = 0x2;
            else cidinput = 0x3;
        }
        else cidinput = (nomodem & 0x1);
        if (noserial || nomodem)
        {
            logMsg(LEVEL1, "noserial = %d and nomodem = %d were used to create cidinput\n", noserial, nomodem);
        }
    }

    if (noserial || nomodem)
    {
        logMsg(LEVELE, "***WARNING*** noserial and nomodem are obsolete, use cidinput instead\n");
    }

    logMsg(LEVEL1, "cidinput = %d\n", cidinput);

    logMsg(LEVEL1, "sendinfo = %d\n", sendinfo);

    /* check hangup/hupmode against cidinput */
    if (cidinput & 0x1)
    {
        /* no modem */

        if (hangup)
        {
            sprintf(msgbuf,
            "A modem must be used with the internal hangup option. cidinput = %d", cidinput);
            errorExit(-110, "Fatal", msgbuf);
        }

        if (hupmode)
        {
            sprintf(msgbuf,
            "A modem must be used with a hangup extension. cidinput = %d", cidinput);
            errorExit(-110, "Fatal", msgbuf);
        }
    }

    /*
     * indicate what is configured to send to the clients
     */
    for (i = 0; sendclient[i].word; i++)
        if (*sendclient[i].value)
        {
            logMsg(LEVEL1, "Configured to send '%s' to clients.\n",
                sendclient[i].word);
        }

    /*
     * indicate location of helper scripts
     */
    logMsg(LEVEL1, "Helper tools:\n    %s\n    %s\n", NCIDUPDATE, NCIDUTIL);

    /*
     *     check if country code is valid
     */
    if (findRegionCode(country) != 0)
    {
      logMsg(LEVEL1, "country code set to : %s \n", country);
    }
    else
    {
      sprintf(msgbuf, "%s is not a valid country code\n", country);
      errorExit(-110, "Fatal", msgbuf);
    }

    /* NANP country format */
    switch (nanp_format)
    {
      case 0:
        logMsg(LEVEL1, "no format (%d): xxxxxxxxxx\n", nanp_format);
        break;
      case 1:
        logMsg(LEVEL1, "google standard format (%d): (xxx) xxx-xxxx\n", nanp_format);
        break;
      case 2:
        logMsg(LEVEL1, "all-hyphen format (%d): xxx-xxx-xxxx\n", nanp_format);
        break;
      case 3:
        logMsg(LEVEL1, "all-dot format (%d): xxx.xxx.xxxx\n", nanp_format);
        break;
    }


    switch (regex)
    {
        case 0:
            logMsg(LEVEL1, "Using simple expressions for alias/blacklist/whitelist entries\n");
            break;
        case 1:
            logMsg(LEVEL1, "Using posix regular expressions for alias/blacklist/whitelist entries\n");
            break;
        case 2:
            logMsg(LEVEL1, "Using Perl regular expressions for alias/blacklist/whitelist entries\n");
            break;
        default:
            sprintf(msgbuf, "Invalid regex number: %d", regex);
            errorExit(-107, "Fatal", msgbuf);
            break;
    }

    logMsg(LEVEL1,
      "\nBegin: Loading alias, blacklist and whitelist files [%s]\n", strdate(ONLYTIME));

    /* read alias file, if present, exit on any errors */
    if (readAliasFile()) errorExit(-109, 0, 0);

    /* read blacklist and whitelist files, exit on any errors */
    readBWlists();

    if (hangup) logMsg(LEVEL1, "%s\n%s\n", BLMSG, WLMSG);
    logMsg(LEVEL1, "%s\n", ignore1 ? IGNORE1 : NOIGNORE1);

    /* time of last ABW reads */
    abw_lastread  = time(0);
    checkABWtimer = time(0) + ABWCHECK;

    logMsg(LEVEL1,
      "End: Loaded alias, blacklist and whitelist files [%s]\n\n", strdate(ONLYTIME));
    if (verbose == LEVEL8) normalExit();

    if (stat(cidlog, &statbuf) == 0)
    {
      logMsg(LEVEL1, "CID logfile: %s\nCID logfile maximum size: %lu bytes\n",
        cidlog, cidlogmax);
    }
    else
    {
        /* Create the call log file if not present
           because writeLog won't create files */
        if ((fd = open(cidlog, O_WRONLY | O_APPEND | O_CREAT,
             S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0)
        {
            logMsg(LEVELE, "%s: %s\n", cidlog, strerror(errno));
        }
        else
        {
          close(fd);
          logMsg(LEVEL1,
            "Created CID logfile: %s\nCID logfile maximum size: %lu bytes\n",
            cidlog, cidlogmax);
        }
    }

    if (stat(datalog, &statbuf) == 0)
    {
        logMsg(LEVEL1, "Data logfile: %s\n", datalog);
    }
    else
    {
        logMsg(LEVEL1, "Data logfile not present: %s\n", datalog);
    }

    /*
     * lineid could have been changed in either ncidd.conf or ncidd.alias
     * this sets cid.cidline to lineid after any changes to it
     */
    strncpy(aliasbuf, serial_lineid, CIDSIZE -1);
    aliasbuf[CIDSIZE-1] = '\0';
    userAlias(0, 0, aliasbuf);
    serial_lineid = strdup(aliasbuf);

    /*initialize the modem cid */
    strcpy(modem_cid.cidmesg, NOMESG); /* default message */
    strcpy(modem_cid.cidline, serial_lineid); /* default line indicator */
    modem_cid.status = 0;
    modem_cid.cidsent = 0;
    modem_cid.calltype = CID;
    modem_cid.infoBuffer = NULL;

    /* Add lineid's to LineIDS */
    if (strchr(serial_lineid, ' ') || strchr(serial_lineid, '\t')) /* if space or tab */
        sprintf(LineIDS, "\"%s\"", serial_lineid);
    else sprintf(LineIDS, "%s", serial_lineid);

    logMsg(LEVEL1, "Maximum number of clients/gateways: %d\n",
            (cidinput == 0x3) ? MAXCLIENTS + 1 : MAXCLIENTS);

    logMsg(LEVEL1, "Telephone Line Identifier: %s\n", serial_lineid);

    logMsg(LEVEL1, "Hostname & Port: %s:%d\n", hostname, port);

    if (cidinput == 0x3)
    {
      /* gateway without modem support */
      logMsg(LEVEL1, "Caller ID from gateways without modem support\n");
    }
    else
    {
      /*
       * cidinput == 0x0, 0x1, or 0x2
       * TTY port needs to be set
       */
      openandCheckSerial();
    }

    if (debug || OSXlaunchd)
    {
        if (debug) logMsg(LEVEL1, "Debug Mode\n");
        else logMsg(LEVEL1, "OSX Launchd Mode\n");
    }
    else
    {
        /* fork and exit parent */
        if(fork() != 0) return 0;

        /* close stdin and make fd 0 unavailable */
        close(0);
        if (open("/dev/null",  O_WRONLY | O_SYNC) < 0)
        {
            errorExit(-1, "/dev/null", 0);
        }

        /* become session leader */
        setsid();
    }

    /*
     * Create a pid file
     */
    if (doPID(&pid, pidfile))
    {
        sprintf(msgbuf,"%s already exists", pidfile);
        errorExit(-110, "Fatal", msgbuf);
    }

    /* initialize the polld table */
    initPoll();
    /* initialize the safeSend client table */
    initQueuedMsg(MAXCONNECT);

    /* add the modem to the polld table, if configured */
    if (cidinput != 0x3 && ttyfd >= 0) {
        /* using a modem or device */
        pollpos = addPoll(ttyfd);
        logMsg(LEVEL3,"%s is fd %d\n", (cidinput==0x1)?"Caller ID Device":"Modem", ttyfd);
    }
    else {
        pollpos = -1;
    }

    /* initialize server socket */
    if ((mainsock = tcpOpen()) < 0) errorExit(-1, "socket", 0);

    /* add the server listen socket to the polld table, eg port 3333 */
    ret = addPoll(mainsock);
    logMsg(LEVEL3,"NCID connection socket is sd %d pos %d\n", mainsock, ret);

    logMsg(LEVEL1,"End of startup: %s\n",
            strdate(WITHSEP));

    /*
     * loop forever waiting for events from devices and sockets
     */
    while (1)
    {
        switch (events = poll(polld, MAXCONNECT, POLL_TIMEOUT))
        {
            case -1:    /* error */
                if (errno != EINTR) /* No error for SIGHUP */
                    errorExit(-1, "poll", 0);
                /* fall through to treat interrupts the same as a time out */

            case 0:    /* time out, without an event */
                /* check to see if DIAL hangup timer is in use */
                if (hanguptimer) doHangupTimer();
                /* check to see if ringing has stopped */
                if (ringtimer) doRingTimer();
                /* try modem init again */
                if (modemlosttimer) doModemLostTimer();
                /* see if ABW files have been edited */
                if (checkABWtimer) doCheckABWTimer();

                if (TERM_flag)
                {
                   /* a TERM signal asked ncidd to exit */
                   TERMexit();
                }
                if (USR1_flag)
                {
                    /* got signal SIGUSR1, replace cidcall.log file with cidcall.log.new */
                    USR1_flag = 0;
                    sprintf (msgbuf, "%s.new", cidlog);
                    if (access (msgbuf, F_OK) == 0)
                    {
                        rename (msgbuf, cidlog);
                        logMsg (LEVEL1,
                        "Replaced %s with %s.new: %s\n", cidlog, cidlog, strdate(ONLYTIME));
                    }
                }
                if (HUP_flag)
                {
                   /* got signal SIGHUP, reload the alias, whitelist, blacklist */
                   HUP_flag = 0;
                   reload();
                }
                if (USR2_flag)
                {
                   /* got signal SIGUSR2, list the connected clients */
                   USR2_flag = 0;
                   showConnected();
                }

                /* if no modem, skip TTY lockfile code */
                if (!(cidinput & 0x1))
                {
                   /* 0x0 and 0x2 */
                   doModemLockTimer();
                }
                break;

            default:    /* 1 or more events */
                if (TERM_flag)
                {
                   /* a TERM signal asked ncidd to exit */
                   TERMexit();
                }
                /* check to see if DIAL hangup timer is in use */
                if (hanguptimer) doHangupTimer();
                /* check to see if ringing has stopped */
                if (ringtimer) doRingTimer();

                doPoll(events);
                break;
        }
    }
}

/* hang up an outgoing DIAL call after a few seconds (dialdelay) */
void doHangupTimer(void)
{
  char msgbuf[BUFSIZ];

  if (hanguptimer < time(0))
  {
    /* turn the timer off */
    hanguptimer = 0;

    /* hang up the call, log it and send notice to client */
    initModem(HANGUP, READTRY);
    logMsg(LEVEL2, "dial - hungup %s on line \"%s\" at %s\n",
            modem_cid.cidnmbr, modem_cid.cidline, strdate(ONLYTIME));
    sprintf(msgbuf, DIAL_HANGUP CRLF, modem_cid.cidnmbr, modem_cid.cidline);
    safeSend(dialclientpos, msgbuf);

    /* Reset caller id info in modem_cid */
    strcpy(modem_cid.cidmesg, NOMESG); /* default message */
    strcpy(modem_cid.cidline, serial_lineid); /* default line indicator */
    modem_cid.status = 0;
    modem_cid.cidsent = 0;
    modem_cid.calltype = CID;
    if (modem_cid.infoBuffer!=NULL) free(modem_cid.infoBuffer);
    modem_cid.infoBuffer = NULL;

  }
}

/* the modem hasn't said RING in a while (RINGWAIT seconds) */
void doRingTimer(void)
{
   if (ringtimer < time(0))
   {
       /* turn the timer off */
       ringtimer = 0;
       logMsg(LEVEL5, "ring: %d timeout at: %s\n", ringcnt, strdate(ONLYTIME));
       ringcnt = 0;

       /* tell clients that ringing stopped */
       sendInfo(0, modem_cid.cidline);

       /* Reset caller id info in modem_cid */
       strcpy(modem_cid.cidmesg, NOMESG); /* default message */
       strcpy(modem_cid.cidline, serial_lineid); /* default line indicator */
       modem_cid.status = 0;
       modem_cid.cidsent = 0;
       modem_cid.calltype = CID;
       if (modem_cid.infoBuffer!=NULL) free(modem_cid.infoBuffer);
       modem_cid.infoBuffer = NULL;

   }
}

/* try the modem init again in a while */
void startModemLostTimer(const char * msg)
{
   char msgbuf[BUFSIZ];

   logMsg(LEVEL1, "%s Modem Not Connected - %s at: %s\n", ttyport, msg, strdate(ONLYTIME));

   /* stop polling, but reserve the slot in polld[] */
   if (pollpos >= 0) reservePoll(pollpos);
   if (ttyfd >= 0) close(ttyfd);
   ttyfd = -1;

   if (modemlostcount == 0) {
      sprintf(msgbuf, "%s %s lost", ttyport, (cidinput==0x1)?"Caller ID Device":"Modem");
      sysMesg(msgbuf, serial_lineid);
   }
   if (modemlostcount > 5) {
      if (cidinput == 0x2) {
         /* support modem open fail not fatal, remove it */
         modem_warning();
         modemlosttimer = 0;
         return;
      }
      else {
         sprintf(msgbuf, "%s %s still lost, server shutting down now", ttyport, (cidinput==0x1)?"Caller ID Device":"Modem");
         sysMesg(msgbuf, serial_lineid);
         errorExit(-104, msg, "Modem Lost");
      }
   }
   modemlostcount++;
   modemlosttimer = time(0) + MODEMWAIT;
}

/* try the modem init again */
void doModemLostTimer(void)
{
   char msgbuf[BUFSIZ];

   if (modemlosttimer < time(0))
   {
       /* turn the timer off */
       modemlosttimer = 0;
       logMsg(LEVEL5, "Modem Lost timer expires, trying again: %s\n", strdate(ONLYTIME));

       if (reAttachModem() == 0 && modemlosttimer == 0) {
           modemlostcount = 0;

           /* tell clients that the modem is back */
           sprintf(msgbuf, "%s %s connected", ttyport, (cidinput==0x1)?"Caller ID Device":"Modem");
           sysMesg(msgbuf, serial_lineid);
       }
   }
}

/* try the modem init again */
int reAttachModem(void)
{
   if (openSerial() != 0) return -1;

   if (cidinput == 0)
   {
       /* set modem to do caller ID */
       if (setModem() != 0) return -1;
   }
   else if (cidinput == 0x2)
   {
       /* modem used to support gateway */
       if (initModem(initstr, READTRY) != 0) {
          startModemLostTimer("Failed to initialize modem");
          return -1;
       }
   }

    /* has the modem been here before? */
    if (pollpos >= 0) {
       /* put the modem back into the polld table */
       restorePoll(pollpos, ttyfd);
    }
    else {
       /* first time finding the modem */
       if (checkModemCapabilities() < 0) return -1;

       /* add the modem to the polld table */
       pollpos = addPoll(ttyfd);
       logMsg(LEVEL3,"%s is fd %d\n", (cidinput==0x1)?"Caller ID Device":"Modem", ttyfd);
    }
    return 0;
}


/*
 * Configure serial device or modem
 * for cidinput == 0x0, 0x1, or 0x2
 */
void openandCheckSerial(void)
{

    checkSerialOptions();
    if (openSerial() != 0) return;
    checkModemCapabilities();
}


/*
 * Check the cidinput mode against the hangup/hupmode setting
 * and make sure the supporting voice/extension files are present
 * Describe the settings for the log file.
 */
void checkSerialOptions(void)
{
    int ret;

    /*
     * noserial and nomodem are obsolete (now used to set cidinput)
     *      noserial = 1: serial port not used
     *      noserial = 0: serial port used for Caller ID
     *      nomodem  = 1: serial device
     *      nomodem  = 0: modem
     *
     * cidinput = 0: Caller ID from a modem and optional gateways
     * cidinput = 1: Caller ID from a serial or USB device and optional gateways
     * cidinput = 2: Caller ID from a gateway with modem support
     * cidinput = 3: Caller ID from gateways without modem support
     */

    /* If the tty port speed was set, map it to the correct integer. */
    set_ttyspeed();

    /* Create lock file name from TTY port device name */
    make_lockfile_name();

    /* set or reset the TTYspeed string */
    getSpeedLabel(ttyspeed);

    logMsg(LEVEL1, "TTY port speed: %s\n", TTYspeed);
    logMsg(LEVEL1, "TTY lock file: %s\n", lockfile);
    logMsg(LEVEL1, "TTY port control signals %s\n", clocal ? "disabled" : "enabled");

    if (cidinput == 0x1)
    {
      /* serial device is only wanted for caller ID */
      logMsg(LEVEL1, "Caller ID from a serial or USB device and optional gateways\n");
    }
    else
    {
      /*
       * noserial == 0 && nomodem == 0
       * noserial == 1 && nomodem == 0
       * cidinput == 0x0 or 0x2
      */

      if (cidinput == 0x2)
      {
        /* modem used for gateway support; hangup and dial */

        if (!hangup && !hupmode)
            logMsg(LEVEL1, "Modem will only be used to dial number\n");
        else
            logMsg(LEVEL1, "Modem will only be used to dial number and hangup calls\n");
      }
      else /* cidinput == 0 */
      {
        /* modem used for callerid, hangup and dial */

        /* get modem capabilities */
        logMsg(LEVEL1, "Modem will be used for Caller ID\n");

        if (gencid)
           logMsg(LEVEL1, "Handles modem calls without Caller ID\n");
        else
           logMsg(LEVEL1, "Does not handle modem calls without Caller ID\n");

        logMsg(LEVEL1, "Caller ID from a modem and optional gateways\n");
        logMsg(LEVEL1, "Modem also used to %s\n", (hangup || hupmode)
                ? "dial number and terminate calls" : "dial number");
      }

      /* cidinput == 0x0 or 0x2 do dial and hangup */
      logMsg(LEVEL1, "Dial Modem: %s\nDial Timeout: %d seconds\n",
              serial_lineid, dialdelay);
      logMsg(LEVEL1, "LineIDS: %s\n", LineIDS);

      if (hangup)
      {
        /* check hangup mode against modem abilities */
        if (hangup == 3)
        {
          checkVoiceFile(&hangup, annpath, announce);
        }
      }

      if (hupmode)
      {
        /* check to make sure the hangup extension is present and executable */
        checkHangupExtension();

        logMsg(LEVEL1, "Hangup Extension name: %s\nHangup Extension path: %s\n",
                hupname, huppath);

        if (hupmode == 3)
        {
          ret = 1;
          if (huprmd) ret = checkVoiceFile(&hupmode, rmdpath, huprmd);
          if (ret)
          {
            /* no external hangup voice file, try again with the one for hangup */
            hupmode = 3;
            checkVoiceFile(&hupmode, rmdpath, announce);
            logMsg(LEVEL1, "Default Hangup Extension Recording: %s\n", rmdpath);
          }
        }
      }
    }
}


/*
 * Configure serial device or modem
 * for cidinput == 0x0, 0x1, or 0x2
 */
int openSerial(void)
{
    /* Open tty port; start a timer if it fails */
    if (openTTY() < 0)
    {
        /* perhaps ModemManager was interfering at boot time */
        startModemLostTimer("serial device open TTY failed");
        return -1;
    }

    logMsg(LEVEL1, "TTY port opened: %s\n", ttyport);

    if (setTTY() < 0)
    {
        logMsg(LEVELE, "Set TTY error:  %s\n", strerror(errno));
        startModemLostTimer("serial device setTTY");
        return -1;
    }

    return 0;
}

/*
 * Check modem capabilities
 * for cidinput== 0x0, tells modem to do caller ID
 */
int checkModemCapabilities(void)
{
    int ret;
    static int checked = 0;

    /*
     * noserial == 0 && nomodem == 0
     * noserial == 1 && nomodem == 0
     * cidinput == 0x0 or 0x2
    */
    if (cidinput == 0x1)
    {
      /* serial callerid device, nothing more to do  */
    }
    else if (cidinput == 0x2)
    {
      /* modem used for gateway support; hangup and dial */

      ret = initModem(initstr, READTRY);
      if (ret)
      {
        startModemLostTimer("Gateway support modem not found");
        return -1;
      }
      if (!hangup && !hupmode)
      {
         logMsg(LEVEL1, "Modem detected\nModem only used to dial number\n");
      }
      else
      {
        logMsg(LEVEL1, "Modem detected\nModem only used to dial number and hangup calls\n");
        /* get modem capabilities */
        if (checkModem() < 0) return -1;
      }
    }
    else /* cidinput == 0 */
    {
      /* modem used for callerid, hangup and dial */

      /* get modem capabilities */
      if (checkModem() != 0) return -1;
      /* set modem to do caller ID */
      if (setModem() != 0) return -1;

      logMsg(LEVEL1, "Modem detected\nModem used for Caller ID\n");
      logMsg(LEVEL1, "ncidd set for country %s\n", country);

      if (isNANPACountry(country))
      {
        logMsg(LEVEL1, "ncidd US phonenumber format %d:", nanp_format);
        if (nanp_format == 0) logMsg(LEVEL1, " not formatted\n");
        else if (nanp_format == 1) logMsg(LEVEL1, " (xxx) xxx-xxxx\n");
        else logMsg(LEVEL1, " xxx-xxx-xxxx\n");
      }
    }

    if (checked) return 0;

    /* do this section just once */
    if (hangup)
    {
      /* check hangup mode against modem abilities from checkModem() */
      checkHangupMode(&hangup);
    }

    if (hupmode)
    {
      /* check hangup mode against modem abilities from checkModem() */
      checkHangupMode(&hupmode);
    }
    checked = 1;
    return 0;
}


/*
 * If the tty port speed was set, map it to the correct integer.
 * set the modem ttyspeed from the string TTYspeed
 */
void set_ttyspeed(void)
{
      if (TTYspeed)
      {
        if (!strcmp(TTYspeed, "115200")) ttyspeed = B115200;
        else if (!strcmp(TTYspeed, "38400")) ttyspeed = B38400;
        else if (!strcmp(TTYspeed, "19200")) ttyspeed = B19200;
        else if (!strcmp(TTYspeed, "9600")) ttyspeed = B9600;
        else if (!strcmp(TTYspeed, "4800")) ttyspeed = B4800;
        else if (!strcmp(TTYspeed, "2400")) ttyspeed = B2400;
        else if (!strcmp(TTYspeed, "1200")) ttyspeed = B1200;
        else errorExit(-108, "Invalid TTY port speed set in config file", TTYspeed);
      }
}


const char * getSpeedLabel(int ttyspeed)
{

   if (TTYspeed != NULL) free((void*)TTYspeed);

   switch(ttyspeed)
   {
     case B1200:
       TTYspeed = strdup("1200");
       break;
     case B2400:
       TTYspeed = strdup("2400");
       break;
     case B4800:
       TTYspeed = strdup("4800");
       break;
     case B9600:
       TTYspeed = strdup("9600");
       break;
     case B19200:
       TTYspeed = strdup("19200");
       break;
     case B38400:
       TTYspeed = strdup("38400");
       break;
     case B115200:
       TTYspeed = strdup("115200");
       break;
     default:
       TTYspeed = strdup("TTYspeed?");
       break;
   }
   return TTYspeed;
}


/*
 * Create lock file name from TTY port device name
 * and check for an existing lock file
 */

void make_lockfile_name(void)
{
   if (!lockfile)
   {
      const char * ptr;
      char * newlockfile;
      if ((ptr = strrchr(ttyport, '/'))) ptr++;
      else ptr = ttyport;

      newlockfile = (char *) malloc(strlen(LOCKFILE) + strlen(ptr) + 1);
      if (newlockfile !=  NULL) {
         strcat(strcpy(newlockfile, LOCKFILE), ptr);
         lockfile = newlockfile;
      }
      else errorExit(-1, progname, 0);
   }

   /* check TTY port lock file */
   if (CheckForLockfile()) errorExit(-102, "Exiting - TTY lockfile exists", lockfile);
   if ((hangup || hupmode) && CheckLockfilePath()) errorExit(-102, "Exiting - TTY lockfile not writable", lockfile);
}


/*
 * modem_warning only called when cidinput == 0x2
 * modem used to support gateways so can be removed
 * change cidinput to gateways without modem support
 * make sure hangup is disabled
 */
void modem_warning(void)
{
    int oldcidinput, oldhangup;

    oldcidinput = cidinput;
    oldhangup = hangup;
    cidinput += 0x1; /* remove modem */
    logMsg(LEVELE, "***Warning*** No modem or modem busy: %s\n\
              Changed \"cidinput\" from %d to %d\n\
              Set \"cidinput\" properly in ncidd.conf\n",
              ttyport, oldcidinput, cidinput);
    hangup = 0;      /* make sure hangup is disabled */
    if (oldhangup)
    {
        logMsg(LEVEL1, "Changed \"hangup\" from %d to %d\n\
              Set \"hangup\" to 0 (default) in ncidd.conf\n",
              oldhangup, hangup);
    }
}

int getOptions(int argc, char *argv[])
{
    int c, num;
    int option_index = 0;
    static struct option long_options[] = {
        {"alias", 1, 0, 'A'},
        {"announce", 1, 0, 'a'},
        {"audiofmt", 1, 0, 'f'},
        {"blacklist", 1, 0, 'B'},
        {"cidinput", 1, 0, 'j'},
        {"cidlog", 1, 0, 'c'},
        {"cidlogmax", 1, 0, 'M'},
        {"config", 1, 0, 'C'},
        {"datalog", 1, 0, 'd'},
        {"debug", 0, 0, 'D'},
        {"gencid", 1, 0, 'g'},
        {"help", 0, 0, 'h'},
        {"hangup", 1, 0, 'H'},
        {"hupmode", 1, 0, 'm'},
        {"initcid", 1, 0, 'i'},
        {"initstr", 1, 0, 'I'},
        {"lineid", 1, 0, 'e'},
        {"lockfile", 1, 0, 'l'},
        {"logfile", 1, 0, 'L'},
        {"nomodem", 1, 0, 'n'},
        {"noserial", 1, 0, 'N'},
        {"pidfile", 1, 0, 'P'},
        {"port", 1, 0, 'p'},
        {"ifaddr", 1, 0, 'Q'},
        {"regex", 1, 0, 'r'},
        {"send", 1, 0, 's'},
        {"ttyspeed", 1, 0, 'S'},
        {"ttyclocal", 1, 0, 'T'},
        {"ttyport", 1, 0, 't'},
        {"verbose", 1, 0, 'v'},
        {"version", 0, 0, 'V'},
        {"whitelist", 1, 0, 'W'},
        {"osx-launchd", 0, 0, '0'},
        {0, 0, 0, 0}
    };

    while ((c = getopt_long (argc, argv, "a:c:d:e:f:g:hi:j:l:m:n:p:r:s:t:v:A:B:C:DH:I:L:M:N:P:Q:S:T:VW:",
        long_options, &option_index)) != -1)
    {
        switch (c)
        {
            case '0':
                ++OSXlaunchd;
                break;
            case 'A':
                if (!(cidalias = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("cidalias")) >= 0) setword[num].type = 0;
                break;
            case 'B':
                if (!(blacklist = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("blacklist")) >= 0) setword[num].type = 0;
                break;
            case 'C':
                if (!(config = strdup(optarg))) errorExit(-1, progname, 0);
                break;
            case 'D':
                ++debug;
                break;
            case 'H':
                hangup = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(hangup == 0 && *optarg == '0') && hangup > 3))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("hangup")) >= 0) setword[num].type = 0;
                break;
            case 'I':
                if (!(initstr = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("initstr")) >= 0) setword[num].type = 0;
                break;
            case 'L':
                if (!(logfile = strdup(optarg))) errorExit(-1, progname, 0);
                break;
            case 'M':
                cidlogmax = atoi(optarg);
                if ((num = findWord("cidlogmax")) >= 0)
                {
                    if (cidlogmax < (unsigned) setword[num].min ||
                        cidlogmax > (unsigned) setword[num].max)
                        errorExit(-107, "Invalid number", optarg);
                    setword[num].type = 0;
                }
                break;
            case 'N':
                noserial = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(noserial == 0 && *optarg == '0') && noserial != 1))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("noserial")) >= 0) setword[num].type = 0;
                break;
            case 'P':
                if (!(pidfile = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("pidfile")) >= 0) setword[num].type = 0;
                break;
            case 'S':
                if (!(TTYspeed = strdup(optarg))) errorExit(-1, progname, 0);
                set_ttyspeed();
                if ((num = findWord("ttyspeed")) >= 0) setword[num].type = 0;
                break;
            case 'T':
                clocal = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(clocal == 0 && *optarg == '0') && clocal != 1))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("ttyclocal")) >= 0) setword[num].type = 0;
                break;
            case 'V': /* version */
                fprintf(stderr, SHOWVER, progname, hostname, port, VERSION, API);
                exit(0);
            case 'W':
                if (!(whitelist = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("whitelist")) >= 0) setword[num].type = 0;
                break;
            case 'a':
                if (!(announce = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("announce")) >= 0) setword[num].type = 0;
                break;
            case 'c':
                if (!(cidlog = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("cidlog")) >= 0) setword[num].type = 0;
                break;
            case 'd':
                if (!(datalog = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("datalog")) >= 0) setword[num].type = 0;
                break;
            case 'e':
                if (!(serial_lineid = strdup(optarg))) errorExit(-1, progname, 0);
                if (strlen(serial_lineid) > CIDSIZE -1)
                    errorExit(-113, "string too long", optarg);
                if ((num = findWord("serial_lineid")) >= 0) setword[num].type = 0;
                break;
            case 'f':
                if (!(audiofmt = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("audiofmt")) >= 0) setword[num].type = 0;
                break;
            case 'g':
                gencid = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(gencid == 0 && *optarg == '0') && gencid != 1))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("gencid")) >= 0) setword[num].type = 0;
                break;
            case 'h': /* help message */
                fprintf(stderr, DESC, progname);
                fprintf(stderr, USAGE, progname);
                exit(0);
            case 'i':
                if (!(initcid = strdup(optarg))) errorExit(-1, progname, 0);
                ++setcid;
                if ((num = findWord("initcid")) >= 0) setword[num].type = 0;
                break;
            case 'j':
                cidinput = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(cidinput == 0 && *optarg == '0') && cidinput > 3))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("cidinput")) >= 0) setword[num].type = 0;
                break;
            case 'l':
                if (!(lockfile = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("lockfile")) >= 0) setword[num].type = 0;
                break;
            case 'm':
                hupmode = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(hupmode == 0 && *optarg == '0') && hupmode > 3))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("hupmode")) >= 0) setword[num].type = 0;
                break;
            case 'n':
                nomodem = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(nomodem == 0 && *optarg == '0') && nomodem != 1))
                    errorExit(-107, "Invalid number", optarg);
                if ((num = findWord("nomodem")) >= 0) setword[num].type = 0;
                break;
            case 'p':
                if((port = atoi(optarg)) == 0)
                    errorExit(-101, "Invalid port number", optarg);
                if ((num = findWord("port")) >= 0) setword[num].type = 0;
                break;
            case 'Q':
                if (!(interfaceaddr = strdup(optarg))) errorExit(-1, progname, 0);
                if (strlen(interfaceaddr) > CIDSIZE -1)
                    errorExit(-113, "string too long", optarg);
                if ((num = findWord("interfaceaddr")) >= 0) setword[num].type = 0;
                break;
            case 'r':
                regex = atoi(optarg);
                if (strlen(optarg) != 1 ||
                    (!(regex == 0 && *optarg == '0') && regex != 1 && regex != 2))
                    errorExit(-107, "Invalid regex number", optarg);
                if ((num = findWord("regex")) >= 0) setword[num].type = 0;
                break;
            case 's':
                if ((num = findSend(optarg)) < 0)
                    errorExit(-106, "Invalid send data type", optarg);
                ++(*sendclient[num].value);
                break;
            case 't':
                if (!(ttyport = strdup(optarg))) errorExit(-1, progname, 0);
                if ((num = findWord("ttyport")) >= 0) setword[num].type = 0;
                break;
            case 'v':
                verbose = atoi(optarg);
                /* valid range: 1-9 */
                if (strlen(optarg) != 1 || (verbose == 0))
                {
                    verbose = 1;
                    errorExit(-107, "Invalid number", optarg);
                }
                if ((num = findWord("verbose")) >= 0) setword[num].type = 0;
                break;
            case '?': /* bad option */
                fprintf(stderr, USAGE, progname);
                errorExit(-100, 0, 0);
        }
    }
    return optind;
}

/*
 * Open tty port; tries to make sure the open does
 * not hang if port in use, or not restored after use
 */

int openTTY(void)
{
    /*
     * open the modem tty device
     * O_RDWR    read/write
     * O_NOCTTY  don't become a controlling tty
     * O_NDELAY  don't hang here at the open
     */
    if ((ttyfd = open(ttyport, O_RDWR | O_NOCTTY | O_NDELAY)) < 0)
    {
        logMsg(LEVELE, "TTY port %s open error:  %s\n", ttyport, strerror(errno));
        ttyfd = -1;
        return -1;
    }

    /* change back to blocking I/O */
    if (fcntl(ttyfd, F_SETFL, fcntl(ttyfd, F_GETFL, 0) & ~O_NDELAY) < 0)
    {
        logMsg(LEVELE, "TTY port fcntl error:  %s\n", strerror(errno));
        return -1;
    }

    return 0;
}

/*
 * save tty settings and switch to cooked/ICANON mode
 *
 * Returns:  0 - OK
 *          -1 - error
 */
int setTTY(void)
{
    /* Flush port */
    if (tcflush(ttyfd, TCIOFLUSH) < 0) return -1;


    /* Save tty port settings */
    if (tcgetattr(ttyfd, &otty) < 0) return -1;

    /* Configure the rtty structure for raw mode */
    rtty.c_iflag = (IGNBRK | IGNPAR);
    rtty.c_cflag = (CS8 | CREAD | HUPCL | CRTSCTS);
    rtty.c_cc[VMIN] = 0;
    rtty.c_cc[VTIME] = 1;
    if (clocal) rtty.c_cflag |= CLOCAL;
    if (cfsetispeed(&rtty, ttyspeed) < 0) return -1;
    if (cfsetospeed(&rtty, ttyspeed) < 0) return -1;

    /* Set the tty port for raw mode from rtty */
    if (tcsetattr(ttyfd, TCSANOW, &rtty) < 0) return -1;

    /* delay needed after changes for some devices */
    usleep(1000);

    /* Save the raw tty port to the ntty structure */
    if (tcgetattr(ttyfd, &ntty) < 0) return -1;

    /* Add the ICANON flag to the ntty structure for ICANON mode */
    ntty.c_lflag = ICANON;

    /* Set the tty port to ICANON mode from ntty structure */
    if (tcsetattr(ttyfd, TCSANOW, &ntty) < 0) return -1;

    /* delay needed after changes for some devices */
    usleep(1000);

    logMsg(LEVEL1, "TTY port initialized in ICANON mode.\n");

    return 0;
}

/*
 * change to raw mode if raw = 1
 * change to icanon mode if raw = 0
 */

void changeTTY(int raw)
{
    if (ttyfd < 0) return;

    if (tcflush(ttyfd, TCIOFLUSH) < 0)
    {
        logMsg(LEVELE, "TTY port flush error:  %s\n", strerror(errno));
    }

    if (raw)
       tcsetattr(ttyfd, TCSANOW, &rtty);
    else
       tcsetattr(ttyfd, TCSANOW, &ntty);

    logMsg(LEVEL1, "TTY port changed to %s mode.\n", raw ? "RAW" : "ICANON");

    /* delay needed after changes for some devices */
    usleep(1000);

}

/*
 * Check modem for Call Waiting (SCID?)
 */
int checkSCID(void)
{
    int ret;
    const char *ptr;
    char *eptr;
    int widflag = 0;

    /* Query Snooping CID (Type II) */
    logMsg(LEVEL3, "\nQuery Snooping CID (Type II)\n");
    if (!(ret = initModem(QUERYATSCID, READTRY)))
    {
        ptr = modembuf + 10;
        while (*ptr == '\r' || *ptr == '\n') ++ptr;
        if ((eptr = strchr(ptr, (int) '\r')))
        {
            *eptr = '\0';
            widflag = atoi(ptr);
            logMsg(LEVEL1, "Modem Snooping CID (off = 0 on = 1 or 2): %s\n\n", ptr);
            *eptr = '\r';
        }
        else logMsg(LEVEL1, "Cannot determine if modem Snooping CID\n\n");
    }
    else
    {
        logMsg(LEVEL1, "Modem Snooping CID query failed: ret =%d\n\n", ret);
    }
    return widflag;
}


/*
 * Check some modem capabilities and hangup options
 */
int checkModem(void)
{
    int ret;
    char countryGCI[GCISIZE], *ptr, *eptr, **gci;

    logMsg(LEVEL1, "Checking Modem Capabilities...\n");

    /* Set the tty port to raw mode */
    changeTTY(RAW);

    /* initialize modem */
    if ((ret = initModem(initstr, READTRY)) == 4)
    {
         startModemLostTimer("No modem found");
         return -1;
    }
    else
    {
        logMsg(LEVEL1, "Modem Initialized\n");
    }

    /* Query modem software */
    logMsg(LEVEL1, "\nQuery modem software\n");
    if (!(ret = initModem(QUERYATI3, READTRY)))
    {
        ptr = modembuf + 7;
        while (*ptr == '\r' || *ptr == '\n') ++ptr;
        if ((eptr = strchr(ptr, (int) '\r')))
        {
            *eptr = '\0';
            logMsg(LEVEL1, "Modem Identifier: %s\n", ptr);
            *eptr = '\r';
        }
        else logMsg(LEVEL1, "Cannot determine Modem Identifier\n");
    }
    else
    {
        logMsg(LEVEL1, "Modem Identifier query failed: ret =%d\n", ret);
    }

    /* Query chipset version */
    logMsg(LEVEL1, "\nQuery chipset version\n");
    if (!(ret = initModem(QUERYATPV, READTRY)))
    {
        ptr = modembuf + 7;
        while (*ptr == '\r' || *ptr == '\n') ++ptr;
        if ((eptr = strchr(ptr, (int) '\r')))
        {
            *eptr = '\0';
            logMsg(LEVEL1, "Modem Chipset Version: %s\n", ptr);
            *eptr = '\r';
        }
        else logMsg(LEVEL1, "Cannot determine Modem Chipset Version\n");
    }
    else
    {
        logMsg(LEVEL1, "Modem Chipset Version query failed: ret =%d\n", ret);
    }

    /* Query modem country setting (United States = B5) */
    logMsg(LEVEL1, "\nQuery modem country setting (United States = B5)\n");
    if (!(ret = initModem(QUERYATGCI, READTRY)))
    {
        if ((ptr = strstr(modembuf, "GCI:")))
        {
            if (strchr(ptr, (int) ' ')) ptr += 5;
            else ptr += 4;
            memset(countryGCI, 0, GCISIZE);
            strncpy(countryGCI, ptr, 2);
            for (gci = modemGCI; *gci != NULL; ++gci)
            {
                if (!strncmp(countryGCI, *gci, 2))
                {
                    strncpy(countryGCI, *gci, GCISIZE - 2);
                    break;
                 }
            }
            logMsg(LEVEL1, "Modem country code: %s\n", countryGCI);
        }
        else logMsg(LEVEL1, "Modem country code cannot be determined\n");
    }
    else
    {
        logMsg(LEVEL1, "Modem country code query failed: ret = %d\n", ret);
    }

    /* Query active profile */
    logMsg(LEVEL1, "\nQuery active profile\n");
    if (!(ret = initModem(QUERYATV, READTRY)) || ret == 3)
    {
        if (!(ptr = strchr(modembuf, (int) '\n') + 1))
            logMsg(LEVEL1, "Cannot retrieve modem Active Profile\n");
        if ((eptr = strstr(modembuf, "OK")))
        {
            *(eptr - 2) = '\0';
            if (strstr(modembuf, "ACTIVE PROFILE"))
               logMsg(LEVEL1, "Modem %s", ptr);
            else
               logMsg(LEVEL1, "Modem Active Profile settings: %s", ptr);
        }
        else logMsg(LEVEL1, "Failed to retrieve the modem Active Profile\n");
    }
    else
    {
        logMsg(LEVEL1, "Modem active profile query failed: ret = %d\n", ret);
    }

    /* Query modem modes supported */
    logMsg(LEVEL1, "\nQuery modem modes supported \n");
    if (!(ret = initModem(QUERYATFCLASS, READTRY)))
    {
        /* Modes start with 0 then optional mode numbers for example: 0,1,1.0,2,8 */
        if (strncmp(modembuf, "0", 1))
        {
            modemModes |= 0x1;
            logMsg(LEVEL1, "Modem supports Data Mode\n");
        }
        if (strstr(modembuf, ",1"))
        {
            modemModes |= 0x2;
            logMsg(LEVEL1, "Modem supports FAX Mode 1\n");
        }
        if (strstr(modembuf, ",2"))
        {
            modemModes |= 0x4;
            logMsg(LEVEL1, "Modem supports FAX Mode 2\n");
        }
        if (strstr(modembuf, ",8"))
        {
            modemModes |= 0x8;
            logMsg(LEVEL1, "Modem supports VOICE Mode\n");
        }
            logMsg(LEVEL1, "\n");
    }
    else
    {
        logMsg(LEVEL1, "Modem modes query failed: ret = %d\n", ret);
    }

    /* Set the tty port to ICANON mode */
    changeTTY(NOTRAW);

    return 0;
}

/*
 * checkVoiceFile
 * mode is reset to 1 if recording file is not found
 * path is filled in with the absolute path to the file
 * return = 0 for a recording file
 * return = 1 for no recording file
 */
int checkVoiceFile(int *mode, char * path, const char * recording)
{
    int rc;
    struct stat statbuf;

    if (!recording || *recording == '\0')
    {
        recording = RECORDING;
    }

    if (*recording == '/')
    {
        sprintf(path, "%s", recording);
    }
    else
    {
        sprintf(path, "%s/%s", ANNDIR, recording);
    }

    if (stat(path, &statbuf))
    {
        *mode = 1;
        logMsg(LEVELE, "WARNING: %s using Normal Hangup (mode 1), no recording file:\n         %s\n",
            mode == &hangup ? "Internal Hangup" : "Hangup Extension", path);
        rc = 1;
    }
    else
    {
        logMsg(LEVEL1, "%s recording file: %s\n",
            mode == &hangup ? "Internal Hangup" : "Hangup Extension", path);
        rc = 0;
    }

    return rc;
}


/*
 * return = 1 for a valid hangup extension
 * errrorExit for a problem
 */
void checkHangupExtension(void)
{
    struct stat statbuf;


    if (*hupname == '/')
    {
      /* separate the name and path */
      char * ptr;
      sprintf(huppath, "%s", hupname);
      ptr = strrchr(huppath, '/');
      hupname = strdup(ptr + 1);
    }
    else sprintf(huppath, "%s/%s", EXTDIR, hupname);

    if (strncmp(hupname, HUPNAME "-", strlen(HUPNAME "-")))
    {
      logMsg(LEVELE, "Hangup Extension name must start with \"%s-\": %s\n",
              HUPNAME, hupname);
      errorExit(-114, 0, 0);
    }
    else if (stat(huppath, &statbuf) != 0)
    {
      logMsg(LEVELE, "Hangup Extension not found: %s\n", huppath);
      errorExit(-114, 0, 0);
    }
    else if (!((statbuf.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)) ==
             (S_IXUSR | S_IXGRP | S_IXOTH)))
    {
      logMsg(LEVELE, "%s must be executable\n", huppath);
      errorExit(-114, 0, 0);
    }
}


/*
 * Check if hangup mode is supported by modem
 * if not, set mode for a normal hangup
 * Log hangup mode
 * If FAX mode, log if pickup is enabled or not
 * If voice mode, log voice parameters
 */
void checkHangupMode(int *mode)
{
    int ret;
    char msgbuf[BUFSIZ], *ptr, *eptr;

    /* if hangup option set, do checks on hangup modes  */
    if (*mode > 0)
    {
        /* determine if FAX or VOICE mode can be used, if not change it */
        switch (*mode)
        {
            case 1: /* normal hangup */
                break;

            case 2: /* FAX mode */
                if (!(modemModes & 0x2) && !(modemModes & 0x4))
                {
                    /* fax not supported */
                    *mode = 1;
                    logMsg(LEVELE, "WARNING: %s using Normal Hangup (mode 1), modem does not support FAX Hangup\n",
                        mode == &hangup ? "Internal Hangup" : "Hangup Extension");
                }
                else
                {
                    logMsg(LEVEL1, "Pickup %s for FAX hangup\n",
                    pickup == 1 ? "enabled" : "not enabled");
                }
                break;

            case 3: /* VOICE mode */
                if (!(modemModes & 0x8))
                {
                    /* voice not supported */
                    *mode = 1;
                    logMsg(LEVELE, "WARNING: %s using Normal Hangup (mode 1), modem does not support Announce Hangup\n",
                        mode == &hangup ? "Internal Hangup" : "Hangup Extension");
                }
                else
                {
                    /* set raw mode */
                    changeTTY(RAW);

                    /* Query Voice Sampling Methods (audio data formats) */
                    ret = initModem(VOICEMODE, READTRY);
                    if (!(ret = initModem(QUERYATFMI, READTRY)) || ret == 3)
                    {
                        if ((ptr = strstr(modembuf, "FMI")))
                        {
                            ptr += (strlen("FMI") + 3);
                            eptr = strchr(ptr, (int) '\r');
                            *(eptr + 2) = '\0'; /* skip over "\r\n" */
                            logMsg(LEVEL1, "Manufacturer: %s", ptr);
                        }
                        else logMsg(LEVEL1, "Manufacturer not found\n");
                    }
                    else logMsg(LEVEL1, "Modem AT+FMI query failed\n");
                    if (!(ret = initModem(QUERYATVSM, READTRY)) || ret == 3)
                    {
                      if ((ptr = strstr(modembuf, "VSM")))
                      {
                        ptr = strchr(modembuf, (int) '\n') + 1;
                        eptr = strrchr(modembuf, (int) ',');
                        eptr = strchr(eptr, (int) '\r');
                        *(eptr + 2) = '\0'; /* skip over "\r\n" */
                        logMsg(LEVEL1, "Modem Voice Sampling Methods:\n%s", ptr);
                      }
                      else logMsg(LEVEL1, "Modem Voice Sampling methods not found\n");
                    }
                    else logMsg(LEVEL1, "Modem AT+VSM=? query failed\n");
                    logMsg(LEVEL1, "Modem Voice Sampling Method selected: %s\n", audiofmt);

                    ret = initModem(DATAMODE, READTRY);

                    /* restore ICANON mode */
                    changeTTY(NOTRAW);
                }
                break;

            default: /* should never happen */
                logMsg(LEVELE, "WARNING: Unknown %s mode (%d), using Normal Hangup (mode 1)\n",
                    mode == &hangup ? "Internal Hangup" : "Hangup Extension", *mode);
                *mode = 1;
                break;
        }

        /* log hangup option used */
        sprintf(msgbuf, "%s Mode = ",
            mode == &hangup ? "Internal Hangup" : "Hangup Extension");
        switch (*mode)
        {
            case 1:
                strcat(msgbuf, "1: hangup");
                break;
            case 2:
                strcat(msgbuf, "2: generate FAX tones then hangup");
                break;
            case 3:
                strcat(msgbuf, "3: play an announcement then hangup");
                break;
        }
        strcat(msgbuf, " on a blacklisted call\n");
        logMsg(LEVEL1, msgbuf);
    }
    else
    {
        logMsg(LEVEL1, "%s mode = %d: disabled\n",
            mode == &hangup ? "Internal Hangup" : "Hangup Extension", *mode);
    }
}

/*
 * Configure the modem
 * returns:  0 if successful
 *          -1 if cannot read from or write to modem
 * exits program if major problem
 */
int setModem(void)
{
    int ret;

    if (*initstr)
    {
        changeTTY(RAW);
        /* Try to initialize modem */
        if ((ret = initModem(initstr, READTRY)))
        {
            switch (ret)
            {
                case 1: /* CONNECT */
                    logMsg(LEVEL1, "Modem returned \"CONNECT\".\n");
                    break;
                case 2: /* ERROR */
                case 3: /* incomplete or unexpected response from modem */
                    startModemLostTimer("Unable to initialize modem");
                    return -1;
                case 4: /* no response */
                    startModemLostTimer("No modem found");
                    return -1;
                case -1: /* cannot read from or write to modem */
                    startModemLostTimer("Unable to initialize modem");
                    return -1;

            }
        }
        else
        {
            /* OK */
            logMsg(LEVEL1, "Modem initialized.\n");
        }
    }
    else
    {
        /* initstr is null */
        logMsg(LEVEL1, "Initialization string for modem is null.\n");
    }

    if (*initcid)
    {
        /* try to initialize modem for CID */
        if ((ret = initModem(initcid, READTRY)) < 0) return -1;

        if (ret && !setcid)
        {
            /*default CID string 1 failed, try default CID string 2 */
            initcid = INITCID2;
            if ((ret = initModem(initcid, READTRY)) < 0) return -1;
        }

        if (ret)
        {
            /* CID initialization failed */
            tcsetattr(ttyfd, TCSANOW, &otty);
            startModemLostTimer("Unable to set modem CallerID");
            return -1;
        }
        else
        {
            /* CID initialization succeeded */
            widflag = checkSCID();
            if (widflag)
            {
              logMsg(LEVEL1,
              "Modem set for CallerID and Call Waiting: AT-SCID=%d\n", widflag);
            }
            else logMsg(LEVEL1, "Modem set for CallerID.\n");
        }
    }
    else
    {
        /* initcid is null */
        logMsg(LEVEL1, "CID initialization string for modem is null.\n");
    }
    changeTTY(NOTRAW);

    return 0;
}

/*
 * Initialize modem
 * expects:  ptr     - initialization string to send
 *           maxwait - 0 or more times to wait READWAIT milliseconds
 * returns:  0 modem returns OK
 *           1 modem returns "CONNECT"
 *           2 modem returns "ERROR"
 *           3 incomplete or unexpected response from modem
 *           4 no response from modem
 *           5 modem returns "NO DIALTONE"
 *           6 modem returns "NO CARRIER"
 *          -1 modem read or write error
 */
int initModem(const char * ptr, int maxwait)
{
    int num, wnum, size, midx, ret = 4;
    int readcnt, waitcnt;
    int was_idle;

    UNUSED(wnum);

    if (ttyfd < 0) return -1;

    /* Flush port */
    if (tcflush(ttyfd, TCIOFLUSH) < 0) return -1;

    /* send string to modem */
    strcat(strncpy(modembuf, ptr, BUFSIZ - 2), CRLF);
    size = strlen(modembuf);
    if ((num = write(ttyfd, modembuf, size)) < 0)
    {
      if (errno != EINTR && errno != EAGAIN)
          logMsg(LEVEL2, "modem write error: %s.\n", strerror(errno));
      return -1;
    }
    logMsg(LEVEL3, "Sent Modem %d of %d characters: \n%s", num, size, modembuf);
    if (verbose >= 7) hexdump(modembuf, size);

    /*
     * read until OK, CONNECT, ERROR, or "Press any key"  response detected
     * or number of tries exceeded
     */
    size = 0;
    readcnt = 0;
    waitcnt = 0;
    was_idle = 0;
    while (waitcnt < maxwait)
    {
        /* use poll to wait and see if there is something to read */
        int pret;
        struct pollfd serialfd;
        serialfd.fd = ttyfd;
        serialfd.events = POLLIN | POLLERR | POLLHUP | POLLNVAL;
        serialfd.revents = 0;
        pret = poll(&serialfd, 1, READWAIT);
        if (pret < 0) {
           if (errno != EINTR) {
              logMsg(LEVEL2, "modem polling error: %s.\n", strerror(errno));
              return -1;
           }
        }
        else if (pret == 0) {
           /* timed out, will allow maxwait*READWAIT milliseconds */
           waitcnt++;
           was_idle = 1;
           logMsg(LEVEL9, "modem poll %d millisecond timeout.\n", READWAIT);
        }
        else  if (serialfd.revents & (POLLERR | POLLNVAL)) {
             logMsg(LEVEL2, "modem polling error: ttyfd %d not open.\n", ttyfd);
             return -1;
        }
        else if (serialfd.revents & POLLIN) {
           /* there is something to read */
           readcnt++;
           was_idle = 0;
           if ((num = read(ttyfd, modembuf + size, BUFSIZ - size - 1)) < 0)
           {
              if (errno != EINTR && errno != EAGAIN) {
                 logMsg(LEVEL2, "modem read error: %s.\n", strerror(errno));
                 return -1;
              }
           }
           if (num > 0) {
              /* remove \0 bytes by changing them to \r
               * in case the tcflush was not recent enough
               */
              for (midx = size; midx < size+num; midx++) {
                 if (modembuf[midx] == '\0') modembuf[midx] = '\r';
              }
           }
           size += num;
       }

       if (size > 0 && was_idle)
       {
           /* check response */
           modembuf[size] = '\0';
           if (strstr(modembuf, "OK"))
           {
               ret = 0;
               break;
           }
           else if (strstr(modembuf, "CONNECT"))
           {
               ret = 1;
               break;
           }
           else if (strstr(modembuf, "ERROR"))
           {
               ret = 2;
               break;
           }
           else if (strstr(modembuf, "NO DIALTONE"))
           {
               ret = 5;
               break;
           }
           else if (strstr(modembuf, "NO CARRIER"))
           {
               ret = 6;
               break;
           }
           else if (strstr(modembuf, "Press any key"))
           {
               wnum = write(ttyfd, CRLF, 2);
           }
           else
           {
               ret = 3;
           }
       }
    }
    modembuf[size] = '\0';

    if (size > 0)
    {
        logMsg(LEVEL3, "Modem response: %d characters in %d %s %d %s:\n%s",
            size,
            readcnt, readcnt == 1 ? "read" : "reads",
            waitcnt, waitcnt == 1 ? "wait" : "waits",
            modembuf);
        if (verbose >= HEXLEVEL) hexdump(modembuf, size);
    }
    else
    {
        /* maxwait can be zero to not read a modem response */
        if (maxwait)
           logMsg(LEVEL3, "No Modem Response\n");
        else
            logMsg(LEVEL3, "Skipped read for a modem response\n");
    }

    return ret;
}


/*
 * open a TCP listen socket for clients
 */

int tcpOpen(void)
{
    int     sd, ret, optval;
    static struct  sockaddr_in bind_addr;
    int socksize = sizeof(bind_addr);

    optval = 1;
    if ((sd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        return sd;
    /* allow immediate reuse of the port number on ncidd restarts */
    if((ret = setsockopt(sd, SOL_SOCKET, SO_REUSEADDR,
        &optval, sizeof(optval))) < 0)
        return ret;
    /* send keep-alive probes when connection is idle for 7200s */
    if((ret = setsockopt(sd, SOL_SOCKET, SO_KEEPALIVE,
        &optval, sizeof(optval))) < 0)
        return ret;


    /* bind to a particular interface or INADDR_ANY */
     bind_addr.sin_family = AF_INET;
    if (interfaceaddr[0] == '\0') {
       /*  0.0.0.0, listen to all interfaces on this host */
       bind_addr.sin_addr.s_addr = INADDR_ANY;
    }
    else {
       /* an interface address was specified, so restrict it to one interface */
       bind_addr.sin_addr.s_addr = inet_addr(interfaceaddr);
       if (bind_addr.sin_addr.s_addr == INADDR_NONE) {
          logMsg(LEVELE, "Unable to parse ifaddr: %s, allowing connections from localhost.\n", interfaceaddr);
           bind_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
       }
       else {
          logMsg(LEVEL2, "Allowing connections from ifaddr: %s\n", interfaceaddr);
       }
    }
    memset(bind_addr.sin_zero, 0, 8);
    bind_addr.sin_port = htons(port);   /* port in network byte order */
    if ((ret = bind(sd, (struct sockaddr *)&bind_addr, socksize)) < 0)
    {
        logMsg(LEVEL2, "unable to bind listen socket: %s.\n", strerror(errno));
        close(sd);
        return ret;
    }

    /* listen for new tcp connections from clients */
    if ((ret = listen(sd, 5)) < 0)
    {
        logMsg(LEVEL2, "unable to create listen socket: %s.\n", strerror(errno));
        close(sd);
        return ret;
    }
    return sd;
}

/*
 *  accept a new TCP connection from a client
 */

int  tcpAccept(void)
{
    int sd;
    int ret;
    char tmpbuf[MAXIPBUF-5];    /* -5 to quiet -Wformat-overflow */

    struct  sockaddr_in sa;
    socklen_t sa_len = sizeof(sa);

    if ((sd = accept(mainsock, (struct sockaddr *) &sa, &sa_len)) != -1)
    {
        strcpy(tmpIPaddr, inet_ntoa(sa.sin_addr));

        /* get the client host name (max domain name is 253) */
        ret = getnameinfo((struct sockaddr *) &sa, sa_len, tmpbuf, sizeof(tmpbuf), NULL, 0, 0);
        if (ret == 0)
        {
           sprintf(tmpHostName, " [%s]", tmpbuf);
        }
        else if (ret == EAI_AGAIN)
        {
           /* being internet offline gives:  -3, Temporary failure in name resolution */
           strcpy(tmpHostName, " [Not Available]");
        }
        else
        {
           sprintf(tmpHostName, " [hostname lookup error %d, %s]", ret, gai_strerror(ret));
        }
    }

    return sd;
}


/*
 * initialize the poll table with -1 to mark the available slots
 */

void initPoll(void)
{
    int pos;

    for (pos = 0; pos < MAXCONNECT; ++pos)
    {
        polld[pos].fd = -1;    /* tells poll to ignore this slot in the table */
        ack[pos] = 0;
        polld[pos].revents = 0;
        polld[pos].events = 0;
    }
}

/*
 * Add client to the poll table
 * slots marked with fd -1 are available ( and are ignored by poll() )
 */

int addPoll(int addfd)
{
    int added = 0, pos;

    for (added = pos = 0; pos < MAXCONNECT; ++pos)
    {
        if ( polld[pos].fd >= 0 ) continue;    /* occupied */
        if ( polld[pos].fd < -2 ) continue;    /* reserved */
        ack[pos] = 0;
        polld[pos].revents = 0;
        polld[pos].fd = addfd;
        polld[pos].events = (POLLIN | POLLPRI);            /*  | POLLOUT, will be set later when needed */
        ++added;
        break;
    }
    return added ? pos : -1;
}

/*
 * reply to a client on a cellular network.
 * Drop in replacement for write()
 */

int replyToSock(int sd, char *buf)
{
    int ret;
    int len;

    alarm(SEND_TIMEOUT);

    len = strlen(buf);
    ret = write(sd, buf, len);

    if (ret < 0)
    {
       logMsg(LEVEL1, "replyToSock %d fails with errno: %d %s [%s]\n", sd, errno, strerror(errno), strdate(ONLYTIME));
    }

    alarm(0);

    return ret;
}

/*
 * handle the events from devices and sockets after poll() wakes up
 */

void doPoll(int events)
{
  static int nodatacnt = 0;
  int num, pos, sd = 0;
  char buf[BUFSIZ], msgbuf[BUFSIZ];
  char *ptr;

  /*
   * Poll is configured for POLLIN and POLLPRI events
   * POLLERR, POLLHUP, POLLNVAL events can also happen
   * Poll may also be configured for the POLLOUT event
   * when the outgoing client socket would have blocked.
   */

  for (pos = 0; events && pos < MAXCONNECT; ++pos)
  {
    if (!polld[pos].revents) continue; /* no events */

    /* log event flags */
    logMsg(LEVEL9, "polld[%d].revents: 0x%X, fd: %d\n",
            pos, polld[pos].revents, polld[pos].fd);

    if (polld[pos].revents & POLLHUP) /* Other end hung up, stream may still have data */
    {
      /* was it the modem? */
      if (polld[pos].fd == ttyfd)
      {
        logMsg(LEVEL2, "tty device fd %d pos %d Disconnected %s\n",
                polld[pos].fd, pos, strdate(WITHSEP));
        startModemLostTimer("tty device disconnected");
      }
      else
      {
        /* or a client */
        logMsg(LEVEL2, "Client %d pos %d Disconnected %s\n",
                polld[pos].fd, pos, strdate(WITHSEP));
        disconnectClient(pos);
        --events;
        continue;
      }
    }

    if (polld[pos].revents & POLLERR) /* Poll Error */
    {
      /* was it the modem? */
      if (polld[pos].fd == ttyfd)
      {
        logMsg(LEVEL2, "tty device fd %d pos %d Poll Error %s\n",
                polld[pos].fd, pos, strdate(WITHSEP));
        startModemLostTimer("tty device error");
      }
      else
      {
        /* or a client */
        logMsg(LEVEL1, "Poll Error, closed Client %d pos %d %s\n",
                polld[pos].fd, pos, strdate(WITHSEP));
        disconnectClient(pos);
        --events;
        continue;
      }
    }

    if (polld[pos].revents & POLLNVAL) /* Invalid Request, output fd not open */
    {
      /* was it the modem? */
      if (polld[pos].fd == ttyfd)
      {
        logMsg(LEVEL2, "tty device fd %d pos %d Invalid Request %s\n",
                polld[pos].fd, pos, strdate(WITHSEP));
        startModemLostTimer("Invalid Request from Serial device");
      }
      else
      {
        /* or a client */
        logMsg(LEVEL2, "Removed Client %d pos %d, invalid request %s\n",
                polld[pos].fd, pos, strdate(WITHSEP));
        disconnectClient(pos);
        --events;
        continue;
      }
    }

    if (polld[pos].revents & POLLOUT) /* Write Event */
    {
      /* a previously blocked socket has become writable */
      send_blocked(pos);
    }

    if (polld[pos].revents & (POLLIN | POLLPRI)) /* there is data that can be read */
    {
      /* was it the modem? */
      if (polld[pos].fd == ttyfd)
      {
        if (!modem_locked)
        {
          /* Modem or device has data to read */
          modem_lidx = 0;
          modem_ilines[modem_lidx] = NULL;
          do {
             if ((num = read(ttyfd, buf, BUFSIZ-1)) < 0)
             {
               sprintf(msgbuf, "Serial device %d pos %d read error: %s\n",
                       ttyfd, pos, strerror(errno));
               startModemLostTimer(msgbuf);
             }

             /* Modem or device returned no data */
             else if (!num)
             {
               nodatacnt++;
               /* if no data 10 times in a row, something wrong */
               if (nodatacnt >= 10)
               {

                 logMsg(LEVEL2, "tty device fd %d pos %d returned no data in %d tries. %s\n",
                         polld[pos].fd, pos, nodatacnt, strdate(WITHSEP));
                 startModemLostTimer("Serial device returned no data");
               }
               else
               {
                   logMsg(LEVEL2, "Serial device %d pos %d returned no data in try #%d %s\n", ttyfd, pos, nodatacnt, strdate(WITHSEP));
               }
             }
             else
             {
               /* Modem or device returned data */

               nodatacnt = 0;
               if (verbose >= 7) hexdump(buf, num);

               /* Terminate String */
               buf[num] = '\0';

               /* strip <CR> and <LF> */
               if ((ptr = strchr(buf, '\r'))) *ptr = '\0';
               if ((ptr = strchr(buf, '\n'))) *ptr = '\0';

               logMsg(LEVEL5, "Serial fd %d pos %d sent %d bytes  %s\n", ttyfd, pos, num, strdate(WITHSEP));
               writeLog(datalog, buf);

               modem_ilines[modem_lidx] = strdup(buf);
               /* logMsg(LEVEL5, "line %d %p\n", modem_lidx, modem_ilines[modem_lidx]); */
               modem_lidx++;
               modem_ilines[modem_lidx] = NULL;
               if (modem_lidx == MAXMODEMLINES-1) {
                  /* pass a null terminated array of lines */
                  formatCID(modem_ilines);
                  modem_lidx = 0;
               }
             }
             if (num > 0) {
                num = 0;
                /* use poll to see if there is more */
                struct pollfd serialfd;
                serialfd.fd = ttyfd;
                serialfd.events = POLLIN;
                serialfd.revents = 0;
                poll(&serialfd, 1, 200);
                num = (serialfd.revents & POLLIN);
             }
          } while (num > 0);
          /* pass a null terminated array of lines */
          if (modem_lidx > 0)
              formatCID(modem_ilines);
        }
        else
        {
           logMsg(LEVEL5, "Serial fd %d pos %d is locked\n", ttyfd, pos);
        }
      }
      /* or the listen socket eg 3333 */
      else if (polld[pos].fd == mainsock)
      {
        /* accept a TCP/IP Client Connection */
        if ((sd = tcpAccept()) < 0)
        {
          logMsg(LEVEL2, "Connect Error: %s, sd: %d\n", strerror(errno), sd);
        }
        else
        {
          /* Client connected */

          /* set the client socket to non-blocking */
          if (fcntl(sd, F_SETFL, O_NONBLOCK) < 0)
          {
            logMsg(LEVEL2, "NONBLOCK Error: %s, sd: %d\n",
              strerror(errno), sd);
            close(sd);
          }
          else
          {
            if ((pos = addPoll(sd)) < 0)
            {
              /* sadly there are too many connections for the polld table */
              logMsg(LEVEL1, "Client %d trying to connect.\n", sd);
              logMsg(LEVEL1, TOOMANY, (cidinput & 0x2) ? MAXCLIENTS + 1 : MAXCLIENTS,
                      strdate(WITHSEP), NL);
              sprintf(buf, TOOMANY, (cidinput & 0x2) ? MAXCLIENTS + 1 : MAXCLIENTS,
                      strdate(WITHSEP), CRLF);
              replyToSock(sd, buf);
              close(sd);
            }
            else
            {
              /* set up the new client connection info */
              strncpy(IPinfo[pos].addr, tmpIPaddr, MAXIPBUF -1);
              strncpy(IPinfo[pos].name, tmpHostName, MAXIPBUF -1);
              IPinfo[pos].addr[MAXIPBUF-1] = '\0';
              IPinfo[pos].name[MAXIPBUF-1] = '\0';
              logMsg(LEVEL2, "\nClient %d pos %d from %s%s %s connected %s\n",
                      sd, pos, IPinfo[pos].addr, IPinfo[pos].name,
                      IPinfo[pos].dialout ? "dialout" : "nodialout", strdate(WITHSEP));

              QueuedMsg_begin_connect(pos);

              /* send a welcome message */
              logMsg(LEVEL2, "%s%s %s [%s:%d] %s\n", PRETTYINDENT, ANNOUNCE,
                     progname, hostname, port, VERSION);
              sprintf(buf, "%s %s [%s:%d] %s%s", ANNOUNCE, progname, hostname,
                      port, VERSION, CRLF);
              safeSend(pos, buf);

              logMsg(LEVEL2, "%s%s%s\n", PRETTYINDENT, APIANNOUNCE, API);
              sprintf(buf, "%s%s%s", APIANNOUNCE, API, CRLF);
              safeSend(pos, buf);

              /* Look for HELLO to come within 10 milliseconds */
              doHELLO(pos);

              /* client might have sent no_log or send_log to override the config setting */
              if ((sendlog && sendlog_override == -1) || sendlog_override == 1 )
              {
                sendLog(pos);
              }
              else
              {
                /* CID log not sent */
                logMsg(LEVEL2, "%s%s: %s\n", PRETTYINDENT, NOLOGSENT, cidlog);
                sprintf(msgbuf, "%s%s", NOLOGSENT, CRLF);
                safeSend(pos, msgbuf);
              }

              /* Send important server options to clients */
              doOptLine(pos);

              /* End of startup messages */
              logMsg(LEVEL2, "%s%s\n\n", PRETTYINDENT, ENDSTARTUP);
              sprintf(msgbuf, "%s%s", ENDSTARTUP, CRLF);
              safeSend(pos, msgbuf);
            }
          }
        }
      }
      /* or data has arrived from a client */
      else
      {
          if ((num = read(polld[pos].fd, buf, BUFSIZ-1)) < 0)
          {
            logMsg(LEVEL2, "Client %d pos %d read error %d: %s %s\n",
                    polld[pos].fd, pos, errno, strerror(errno), strdate(WITHSEP));
            /* drop the client for any read error */
            logMsg(LEVEL2, "Client %d pos %d from %s%s %s removed %s\n",
                    polld[pos].fd, pos, IPinfo[pos].addr, IPinfo[pos].name,
                    IPinfo[pos].dialout ? "dialout" : "nodialout",
                    IPinfo[pos].ident, strdate(WITHSEP));
            disconnectClient(pos);
          }
          /* read will return 0 for a disconnect */
          else if (num == 0)
          {
            /* TCP/IP Client End Connection */
            logMsg(LEVEL2, "Client %d pos %d from %s%s %s {%s} disconnected %s\n",
                    polld[pos].fd, pos, IPinfo[pos].addr, IPinfo[pos].name,
                    IPinfo[pos].dialout  ? "dialout" : "nodialout",
                    IPinfo[pos].ident, strdate(WITHSEP));
            disconnectClient(pos);
          }
          else
          {
            /*
             * Client sent message to server
             */

            /* Terminate String */
            buf[num] = '\0';

            /*
             * there is no guarantee that the message is complete, it's
             * a stream, so call handle_incoming_message() to
             * accumulate data until a newline character is found
             * which will then call handle_accumulated_message()
             */
            handle_incoming_message(pos, buf, num);
          }
      }
    }

    polld[pos].revents = 0;
    --events;
  }
}


/*
 * Process a line from a client.
 * safesend limits the length of the incoming buf to acc_size=1024
 */
void handle_accumulated_message(int pos, const char * buf)
{
  char tmpbuf[BUFSIZ];
  char **svrtag;
  /* additional info obtained from unformatted phone number in MSG and NOT */
  char * infoBuffer;

  /* Make sure there is data in the message line */
  if ((strlen(buf)) != 0)
  {
    /* Look for CALL, CALLINFO, or MSG lines */
    if (strncmp(buf, CALL, strlen(CALL)) == 0)
    {
      /*
       * Found CALL Line
       * See comments for gatewayCID for line format
       */

      logMsg(LEVEL3, "Gateway %d pos %d sent CALL data %s\n",
              polld[pos].fd, pos, strdate(WITHSEP));

      writeLog(datalog, buf);
      if (ack[pos]) sendACK(pos, buf);
      gatewayCID(buf + strlen(CALL));
      finishCID(&gwcid);
    }
    else if (strncmp(buf, CALLINFO, strlen(CALLINFO)) == 0)
    {
      /* Found a CALLINFO Line */

      logMsg(LEVEL3, "Gateway %d pos %d sent CALLINFO %s\n",
              polld[pos].fd, pos, strdate(WITHSEP));

       handleGWCallinfo(pos, buf);

    }
    else if (buf[0] == '+' && (buf[4] == ':' || buf[8] == ':'))
    {
      /*
       * Found line from a server gateway
       * for example: "+CID: *", "+HUP: *", "+CIDINFO: *
       * messages require two checks:
       *      "MSG: " and "***",  "NOT: " and "***"
       */

      for (svrtag = serverTags; *svrtag; svrtag++)
      {
          if (!strncmp(buf, *svrtag, strlen(*svrtag)))
          {
              logMsg(LEVEL3, "Server Gateway (sd %d) sent %s %s\n",
                  polld[pos].fd, buf, strdate(WITHSEP));
              if (strncmp(buf, "+CIDINFO", strlen("+CIDINFO")))
              {
                  writeLog(cidlog, buf + 1);
              }
              writeClients(buf + 1);
              break;
          }
      }
      if (*svrtag == NULL)
      {
          logMsg(LEVEL3, "Unknown Server Gateway (sd %d) line %s %s\n",
                  polld[pos].fd, buf, strdate(WITHSEP));
      }
    }
    else if (!strncmp(buf, MSGLINE, strlen(MSGLINE)))
    {
      /*
       * Found MSG: line
       * MSG: <message>
       * or
       * MSG: <message> ###DATE*mmddyyyy*TIME*hhmm*NAME*<name>*NMBR*<number>*LINE*<id>*MTYPE*IN|OUT|SYS|USER*
       *
       * Format of a MSG: line passed to TCP/IP clients by ncidd:
       *
       * MSG: <message> ***DATE*<mmddyyyy>*TIME*<hhmm>*LINE*<id>*NMBR*<nmbr>*MTYPE*IN|OUT|SYS|USER*FNMBR*<formatted number*NTYPE*FIX/CELL*CTRY*<cc>*LOCA*<location>*CARI*<carrier>*NAME*<name>*
       *
       * Write message to cidlog and all clients
       */

      logMsg(LEVEL3, "Client %d sent text message %s\n", polld[pos].fd, strdate(WITHSEP));
      writeLog(datalog, buf);
      getINFO(buf);      /* fills the global struct mesg */

      /*
       * numbers coming from a mobile phone/gateway are often not dialable
       * (have country code but no international prefix)
       * this will take care of it, before we write them in the cidlog
       */
      normalizeToDialableNumberFromFixedLine(mesg.nmbr, CIDSIZE);
      infoBuffer = GetCompleteInfoFromNumber(mesg.nmbr);

      sprintf(tmpbuf, MESSAGE, buf, mesg.date, mesg.time, mesg.line, mesg.nmbr, mesg.type, infoBuffer, mesg.name);

      writeLog(cidlog, tmpbuf);
      writeClients(tmpbuf);
    }
    else if (!strncmp(buf, NOTLINE, strlen(NOTLINE)))
    {
      /*
       * Found NOT: (remote notification) line from a gateway
       * NOT: <message> ###DATE*mmddyyyy*TIME*hhmm*NAME*<name>*NMBR*<number>*LINE*<id>*MTYPE*<IN|OUT|USER>*
       *
       * Format of a NOT: line passed to TCP/IP clients by ncidd:
       *
       * NOT: <message> ***DATE*<mmddyyyy>*TIME*<hhmm>*LINE*<id>*NMBR*<nmbr>*MTYPE*IN|OUT|SYS|USER*FNMBR*<formatted number*NTYPE*FIX/CELL*CTRY*<cc>*LOCA*<location>*CARI*<carrier>*NAME*<name>*
       *
       * Write notice to cidlog and all clients
       *
       * The longest SMS text message is 918 characters.
       * When it is longer than 160 characters then it will be broken
       * into chunks of 153 characters before being sent to the
       * recipient's handset.
       */

      logMsg(LEVEL3, "Gateway %d pos %d sent a notice %s\n",
              polld[pos].fd, pos, strdate(WITHSEP));
      writeLog(datalog, buf);
      if (ack[pos]) sendACK(pos, buf);
      getINFO(buf);      /* fills the global struct mesg */

      /*
       * numbers coming from a mobile phone/gateway are often not dialable
       * (have country code but no international prefix)
       * this will take care of it, before we write them in the cidlog
       */
      normalizeToDialableNumberFromFixedLine(mesg.nmbr, CIDSIZE);
      infoBuffer = GetCompleteInfoFromNumber(mesg.nmbr);

      sprintf(tmpbuf, MESSAGE, buf, mesg.date, mesg.time, mesg.line, mesg.nmbr, mesg.type, infoBuffer, mesg.name);

      writeLog(cidlog, tmpbuf);
      writeClients(tmpbuf);
    }
    else if (!strncmp(buf, RLYLINE, strlen(RLYLINE)) && strstr(buf, "###"))
    {
      /*
       * Found RLY: (remote relay) line from a gatewaay
       * RLY: <message> ###<info fields>
       * Change ### to *** and write to cidlog and all clients
       */

      logMsg(LEVEL3, "Gateway %d pos %d sent a relay %s\n",
              polld[pos].fd, pos, strdate(WITHSEP));
      writeLog(datalog, buf);
      if (ack[pos]) sendACK(pos, buf);
      strcpy(tmpbuf,buf);
      strncpy(strstr(tmpbuf,"###"),"***",3);
      writeLog(cidlog, tmpbuf);
      writeClients(tmpbuf);
    }
    else if (strncmp (buf, REQLINE, strlen(REQLINE)) == 0)
    {
      /* Found REQ: line */
       handleClientRequest(pos, buf);
    }
    else if (strncmp (buf, WRKLINE, strlen(WRKLINE)) == 0)
    {
      /* Found WRK: line */
       handleClientWork(buf);
    }
    else if (strncmp (buf, GOODBYE, strlen(GOODBYE)) == 0)
    {
      /*
       * Found GOODBYE line
       * Disconnect the client.
       */
      logMsg(LEVEL2, CLIENT_GOODBYE, polld[pos].fd, pos, strdate(WITHSEP));
      disconnectClient(pos);
    }
    else if (strncmp (buf, IDENT, strlen(IDENT)) == 0)
    {
      /*
       * Found HELLO: IDENT: line, which has arrived too late for
       * the usual HELLO processing
       * get the client IDENT: info from it anyway
       */
      doIDENT(pos, buf);
    }
    else if (strncmp (buf, COMMAND, strlen(COMMAND)) == 0)
    {
      /*
       * Found HELLO: CMD: line, which has arrived too late for
       * the usual HELLO processing
       * quietly discard it
       */
      logMsg(LEVEL3, "Client %d sent late %s\n", polld[pos].fd, buf);
    }
    else
    {
      /*
       * Found unknown data
       */

      logMsg(LEVEL3, "Client %d sent unknown data %s\n",
              polld[pos].fd, strdate(WITHSEP));
      writeLog(datalog, buf);
    }
  }
  else
  {
    /*
     * Found empty line
     * the android client does this on a keepalive timer
     */

      logMsg(LEVEL6, "Client %d sent empty line %s\n",
              polld[pos].fd, strdate(WITHSEP));
  }
}


/*
 * WRK: line
 * called from handle_accumulated_message()
 * Perform the requested work on behalf of the client
 */
void handleClientWork(const char * buf)
{
  int ret;
  char msgbuf[BUFSIZ];

  UNUSED(ret);

  logMsg(LEVEL2, "%s\n", buf);
  if (strncmp (buf + strlen(WRKLINE), ACPT_LOG, strlen(ACPT_LOG)) == 0)
  {
    /* WRK: ACCEPT LOG or WRK: ACCEPT LOGS
       accept the changes made by REQ: UPDATE or UPDATES */
    if (strstr (buf + strlen(WRKLINE), ACPT_LOGS)) {
        sprintf(msgbuf, "for f in %s.*[0-9]; do mv $f.new $f; done", cidlog);
        ret = system (msgbuf);
        logMsg(LEVEL2, "%s [%s]\n", msgbuf, strdate(ONLYTIME));
    }
    sprintf(msgbuf, "%s.new", cidlog);
    ret = rename (msgbuf, cidlog);
    logMsg(LEVEL2, "rename %s.new %s [%s]\n", cidlog, cidlog, strdate(ONLYTIME));
  }
  else if (strncmp (buf + strlen(WRKLINE), RJCT_LOG, strlen(RJCT_LOG)) == 0)
  {
    /* WRK: REJECT LOG or WRK: REJECT LOGS
       reject the changes made by REQ: UPDATE or UPDATES */
    if (strstr (buf + strlen(WRKLINE), RJCT_LOGS)) {
      sprintf(msgbuf, "rm %s.*.new",cidlog);
      ret = system (msgbuf);
      logMsg(LEVEL2, "%s [%s]\n", msgbuf, strdate(ONLYTIME));
    }
    sprintf(msgbuf, "%s.new", cidlog);
    ret = unlink (msgbuf);
    logMsg(LEVEL2, "erase %s.new [%s]\n", cidlog, strdate(ONLYTIME));
  }
}


/*
 * REQ: line
 * called from handle_accumulated_message()
 * Perform the requested action and send a response back to the client
 */
void handleClientRequest(int pos, const char * buf)
{
   static int cnt;
   char tmpbuf[BUFSIZ], msgbuf[BUFSIZ], opt[BUFSIZ];
   char *ptr, *sptr;

   logMsg(LEVEL2, "Received: %s    [%s]%s", buf,
            strdate(WITHSEP), NL);
   if (!strncmp(buf, RELOAD, strlen(RELOAD)))
   {
     /*
      * Found REQ: RELOAD line
      */
      long position = 0;

      /* remember the current position in the log file */
      if (logptr)
      {
          position = ftell (logptr);
      }
      /* reload alias, blacklist and whitelist files */
      reload();

      if (logptr)
      {
         /* read the log file to send to the client */
         msgbuf[0] = '\0';
         cnt = 0;
         fseek (logptr, position, SEEK_SET);
         while (fgets (tmpbuf, sizeof (tmpbuf), logptr) != 0)
         {
             cnt += sizeof (INFOLINE) + strlen(tmpbuf);
             if ((unsigned)cnt >= sizeof (msgbuf) - 2) break;
             if ((unsigned)cnt >= BUFSIZ - 2) break;
             strcat (msgbuf, INFOLINE);
             strcat (msgbuf, tmpbuf);
         }
      }
      else
      {
         strcpy(msgbuf, INFOLINE RELOADED NL);
      }
      logMsg(LEVEL2, BEGIN_DATA NL);
      safeSend(pos, BEGIN_DATA CRLF);

      logMsg(LEVEL2, msgbuf);
      safeSend(pos, msgbuf);

      logMsg(LEVEL2, END_DATA NL);
      safeSend(pos, END_DATA CRLF);
   }
   else if (!strncmp(buf, UPDATE, strlen(UPDATE)))
   {
     /*
      * Found REQ: UPDATE or UPDATES
      * invoke cidupdate to update all of the CID call log files
      */
      FILE *respHandle;
      char *ignore;
      char * updatecmd;
      int status;

      (void) ignore;

      updatecmd = (char *) malloc(strlen(DOUPDATE)+strlen(config)+strlen(cidalias)+strlen(cidlog)+strlen(whitelist)+strlen(blacklist)+sizeof(opt)+20);

      opt[0] = '\0';
      /* explicitly add all the path options for ABW */
      sprintf (updatecmd, DOUPDATE, config, cidalias, cidlog, whitelist, blacklist);
      if (!strncmp(buf, UPDATES, strlen(UPDATES))) strcat(updatecmd, " --multi");
      sprintf(opt, " --regex %d", regex);
      if (ignore1) strcat(opt, " --ignore1");
      if (hangup || hupmode) strcat(opt, " -H 1");
                        else strcat(opt, " -H 0");

      strcat(updatecmd, opt);
      strcat(updatecmd, " < /dev/null 2>&1");

      logMsg(LEVEL4, "Begin: Executing %s [%s]\n", NCIDUPDATE, strdate(ONLYTIME));

      /* invoke /bin/sh -c updatecmd */
      respHandle = popen (updatecmd, "r");

      if (!respHandle) {
         /* when sh -c fails to start */
         logMsg(LEVELE, "popen %s failed - %s\n", NCIDUPDATE, strerror(errno));
         safeSend(pos, "cidupdate failed to run" CRLF);
      }
      else
      {
         strcat(updatecmd, "\n");
         logMsg(LEVEL2, updatecmd);
         strcpy (msgbuf, INFOLINE);
         ptr = msgbuf + sizeof (INFOLINE) - 1;
         cnt = sizeof (msgbuf) - sizeof (INFOLINE);
         ignore = fgets (ptr, cnt, respHandle);
         /* "denied" will come from the system errno, e.g. Unable to open new call log: Permission denied
          * "no changes" comes from cidupdate's call to nciddalias.c
          */
         if (strstr(msgbuf, NOCHANGES) || strstr(msgbuf, DENIED))
         {
           /* There were no changes to the call log */
           logMsg(LEVEL2, BEGIN_DATA NL);
           safeSend(pos, BEGIN_DATA CRLF);

           logMsg(LEVEL2, msgbuf);
           safeSend(pos, msgbuf);
         }
         else
         {
           /* There were changes to the call log */
           logMsg(LEVEL2, BEGIN_DATA1 NL);
           safeSend(pos, BEGIN_DATA1 CRLF);

           logMsg(LEVEL2, msgbuf);
           safeSend(pos, msgbuf);

           /* return the output from cidupdate */
           while (fgets(ptr, cnt, respHandle))
           {
              logMsg(LEVEL2, msgbuf);
              safeSend(pos, msgbuf);
           }
         }

         status = pclose(respHandle);

         if (status < 0) {
            logMsg(LEVELE, "%s pclose %s\n", NCIDUPDATE, strerror(errno));
         }
         else
         {
            if ( WIFEXITED(status) ) {
               logMsg(LEVEL4, "%s exited with status %d\n", NCIDUPDATE, WEXITSTATUS(status) );
            }
            else {
               logMsg(LEVEL4, "%s abnormal exit\n", NCIDUPDATE);
            }
         }
         logMsg(LEVEL4, "End: Executing %s [%s]\n", NCIDUPDATE, strdate(ONLYTIME));

         logMsg(LEVEL2, END_DATA NL);
         safeSend(pos, END_DATA CRLF);
         free(updatecmd);
      }
   }
   else if (!strncmp(buf, REREAD, strlen(REREAD)))
   {
     /*
      * Found REQ: REREAD
      */
     sendLog(pos);
   }
   else if (!strncmp(buf, DIAL, strlen(DIAL)))
   {
     /*
      * Found REQ: DIAL <number>&&<name>&&<lineid>
      * or    REQ: DIAL_ABORT <number>&&<name>&&<lineid>
      */
     int mret, abort = 0;
     char *eptr1, *eptr2, mstr[100];

    logMsg(LEVEL2, BEGIN_DATA2 NL);
     safeSend(pos, BEGIN_DATA2 CRLF);
     sprintf(msgbuf, DIAL_WORKING NL);
     safeSend(pos, msgbuf);

     /* make sure the REQ format is OK */
     eptr1 = eptr2 = 0;
     eptr1 = strstr(buf, "&&");
     if (eptr1) eptr2 = strstr(eptr1 + 2, "&&");
     if (!(eptr1 && eptr2))
     {
       logMsg(LEVEL2, END_RESP NL);
       safeSend(pos, END_RESP CRLF);
       /* dial format error */
       logMsg(LEVEL2, DIAL_LINE_ERR NL, buf);
       safeSend(pos, msgbuf);
       /* give up */
       return;
     }

     /* DIAL or DIAL_ABORT */
     sptr = strchr(buf, 'L'), sptr += 2;
     if (*sptr == 'A') abort = 1, sptr += 6;

     /* get number */
     *eptr1 = '\0';
     strncpy(modem_cid.cidnmbr, sptr, CIDSIZE - 1);
     modem_cid.cidnmbr[CIDSIZE-1] = '\0';
     *eptr1 = '&';
     sptr = eptr1 + 2;

     /* get name */
     *eptr2 = '\0';
     strncpy(modem_cid.cidname, sptr, CIDSIZE - 1);
     modem_cid.cidname[CIDSIZE-1] = '\0';
     *eptr2 = '&';
     sptr = eptr2 + 2;

     /* get dial lineid */
     strncpy(modem_cid.cidline, sptr, CIDSIZE - 1);
     modem_cid.cidline[CIDSIZE-1] = '\0';

     if (abort)
     {
       /* user aborted dial request */
       if (hanguptimer)
       {
         logMsg(LEVEL2, END_RESP NL);
         safeSend(dialclientpos, END_RESP CRLF);
         hanguptimer = 1;        /* make it go off right away */
         doHangupTimer();
       }
     }
     else
     {
       /* handle dial request */

       /* dialOUT means a gateway will log the OUT: line */
       if (!dialOUT)
       {
         /* server creates OUT line for dialout, not a gateway */
         userAlias(modem_cid.cidnmbr, modem_cid.cidname, modem_cid.cidline);
         ptr = strdate(NOSEP);
         strncpy(modem_cid.ciddate, ptr, 8);
         modem_cid.ciddate[8] = '\0';
         strncpy(modem_cid.cidtime, ptr + 9, 4);
         modem_cid.cidtime[4] = '\0';
        /* get additional info from phone number */
         if (modem_cid.infoBuffer != NULL) free(modem_cid.infoBuffer);
         modem_cid.infoBuffer = GetCompleteInfoFromNumber(modem_cid.cidnmbr);
       }

       initModem("AT", READTRY);
       sprintf(msgbuf, "ATDT%s;%s", modem_cid.cidnmbr, CRLF);
       if (!(mret = initModem(msgbuf, READTRY)))
       {
         /* start timer to hang up after dialdelay seconds */
         hanguptimer = time(0) + dialdelay;
         dialclientpos = pos;

         logMsg(LEVEL2, "RESP: Dialed %s on line \"%s\" %s\n", modem_cid.cidnmbr, modem_cid.cidline, strdate(WITHSEP));
         sprintf(msgbuf, DIALED_NMBR NL, modem_cid.cidnmbr, modem_cid.cidline);
         safeSend(pos, msgbuf);

         /* server creates OUT line for dialout, otherwise a gateway does it */
         modem_cid.calltype = OUT;
         if (!dialOUT) sendCall(OUTLINE, &modem_cid, modem_cid.cidname);

         logMsg(LEVEL2, DIALOK NL, dialdelay);
         sprintf(msgbuf, DIALOK CRLF, dialdelay);
         safeSend(pos, msgbuf);
         logMsg(LEVEL2, END_RESP NL);
         safeSend(pos, END_RESP CRLF);
       }
       else /* Dial Number Failed */
       {
         logMsg(LEVEL2, END_RESP NL);
         safeSend(pos, END_RESP CRLF);

         switch (mret)
         {
           case 4:
             sprintf(mstr, "NO RESPONSE");
             break;
           case 5:
             sprintf(mstr, "NO DIALTONE");
             break;
           default:
             sprintf(mstr, "error code %d", mret);
             break;
         }
         logMsg(LEVEL2, DIAL_FAIL NL, mstr);
         safeSend(pos, msgbuf);

         mret = initModem(HANGUP, READTRY);
       }
     }
   }
   else if (!strcmp(buf, REQ_ACK) || !strcmp(buf, REQ_YO))
   {
     /*
      * Found REQ: ACK or REQ: YO
      * respond with ACK: and echo the request
      */
      if (strstr(buf, ACK)) ack[pos] = 1;

      logMsg(LEVEL3, "Sent Client %d pos %d: %s%s %s%s",
              polld[pos].fd, pos, ACKLINE, buf, strdate(WITHSEP), NL);
      sprintf(msgbuf, "%s%s%s", ACKLINE, buf, CRLF);
      safeSend(pos, msgbuf);
   }
   else if (!strncmp(buf, REQ_INFO, strlen(REQ_INFO)))
   {
      /* found a REQ: INFO <nmbr>&&<name>&&<line> line */
      char *temp = 0;
      const char * cptr;
      char  name[CIDSIZE], number[CIDSIZE], line[CIDSIZE], dialinfo[2*CIDSIZE+20];
      int   which;

      cptr = buf + strlen(REQ_INFO);

      number[0] = name[0] = line[0] = '\0';
      /* parse out number name line */
      if (strlen(cptr) > 1)
      {
        cptr++;
        /* cheats const char cptr here */
        if ((temp = strstr(cptr, "&&"))) *temp = '\0';
        strncpy (number, cptr, CIDSIZE-1);
        number[CIDSIZE-1] = '\0';
        if (temp)
        {
          cptr += strlen(number) + 2;
          if ((temp = strstr(cptr, "&&"))) *temp = '\0';
          strncpy (name, cptr, CIDSIZE-1);
          name[CIDSIZE-1] = '\0';
        }
        if (temp)
        {
          cptr += strlen(name) + 2;
          strncpy (line, cptr, CIDSIZE-1);
          line[CIDSIZE-1] = '\0';
        }
      }
      else
      {
        /* line was only REQ: INFO\n
         * no reply logs
         * Unable to handle REQ: INFO request - Ignored.
         */
      }

      if (number[0] == '\0' || name[0] == '\0')
      {
         /* number and name must be supplied */
         logMsg(LEVEL1, "Unable to handle request %s - Ignored.\n", buf);
      }
      else
      {
        logMsg(LEVEL4, "Begin: findAlias() [%s]\n", strdate(ONLYTIME));
        temp = findAlias(name, number, line);
        logMsg(LEVEL4, "End: findAlias() [%s]\n", strdate(ONLYTIME));


        logMsg(LEVEL2, BEGIN_DATA3 NL);
        safeSend(pos, BEGIN_DATA3 CRLF);

        /* send back something like INFO: alias NOALIAS */
        logMsg(LEVEL2, INFOLINE "alias %s\n", temp);
        sprintf(msgbuf, INFOLINE "alias %s%s", temp, CRLF);
        safeSend(pos, msgbuf);

        which = onBlackWhite(name, number);
        switch (which)
        {
        case 0:
                temp = "neither";
                break;
        case 1:
                temp = "black name";
                break;
        case 2:
                temp = "white name";
                break;
        case 5:
                temp = "black number";
                break;
        case 6:
                temp = "white number";
                break;
        default:
                temp = "";
                break;
        }
        /* send back something like INFO: neither */
        logMsg(LEVEL2, INFOLINE "%s\n", temp);
        sprintf (msgbuf, INFOLINE "%s\r\n", temp);
        safeSend(pos, msgbuf);

        /* tell client if it can use dial */
        ptr = number;
        if (!(cidinput & 0x1))
        {
          /* 0x0 and 0x2 */
          while(*ptr != '\0')
          {
             if (!isdigit((int) *ptr)) break;
             ptr++;
          }
        }
        if ((cidinput & 0x1) || *ptr)
        {
          /* disable dial in the client */
          strcpy(dialinfo, "dial NODIAL");
        }
        else
        {
          /* enable dial in the client */
          sprintf(dialinfo, "dial %s&&%s", number, name);
        }

        /* send back  411 End of response */
        logMsg(LEVEL2, INFOLINE "%s\n" END_RESP NL, dialinfo);
        sprintf (msgbuf, INFOLINE "%s\r\n" END_RESP CRLF, dialinfo);
        safeSend(pos, msgbuf);

      }
   }
   else
   {
     /*
      * Found REQ: black/white/alias add/remove
      * which will be passed to ncidutil
      */
      const char * filename = "";
      const char * type = "";
      char multi[BUFSIZ];
      const char * cptr;

      multi[0] = opt[0] = '\0';
      sprintf(opt, " --regex %d", regex);
      if (ignore1) strcat(opt, " --ignore1");
      cptr = buf + strlen(REQLINE);
      if (strncmp(cptr, BLK_LST , strlen(BLK_LST)) == 0)
      {
        /*
         * Received: REQ: black add "<string>" "[<comment>]"
         * Received: REQ: black remove "<string>" ""
         */
         filename = blacklist;
         cptr += strlen(BLK_LST);
         type = "Blacklist";
      }
      else if (strncmp(cptr, ALIAS_LST , strlen(ALIAS_LST)) == 0)
      {
        /*
         * Received: REQ: alias add "<number>&&<alias>" "<type>&&<name>"
         * Received: REQ: alias modify "<number>&&<alias>" "<type>&&<name>"
         * Received: REQ: alias remove "<number>&&" "<type>&&<name>"
         */
         filename = cidalias;
         cptr += strlen(ALIAS_LST);
         type = "Alias";
         sprintf (multi, "--multi \"%s %s\"", blacklist, whitelist);
      }
      else if (strncmp(cptr, WHT_LST , strlen(WHT_LST)) == 0)
      {
       /*
        * Received: REQ: white add "<string>" "[<comment>]"    [<date> <time>]
        * Received: REQ: white remove "<string>" ""    [<date> <time>]
        */
         filename = whitelist;
         cptr += strlen(WHT_LST);
         type = "Whitelist";
      }
      else
      {
         /* none of white/black/alias */
         logMsg(LEVEL1, "Unable to handle %s request - Ignored.\n", cptr);
         type = 0;
      }

      /*
       * any of REQ: white, black, alias
       * filename is the path to the ncidd.whitelist, ncidd.blacklist etc
       * cptr points to the rest eg  add "6138285800" "Fri Jul 14 13:22:57 2017"
       */
      if (type != 0 && strlen(cptr) > 4)
      {
          char * ptr;
          char * utilcmd;
          int    status;

          /* invoke ncidutil to adjust whitelist, blacklist or alias files */
          FILE        *respHandle;

          cptr++;
          utilcmd = (char *) malloc(strlen(DOUTIL)+strlen(opt)+strlen(multi)+strlen(filename)+strlen(type)+strlen(cptr)+10);

          sprintf (utilcmd, DOUTIL, opt, multi, filename, type, cptr);

          logMsg(LEVEL4, "Begin: Executing %s [%s]\n", NCIDUTIL, strdate(ONLYTIME));

          /* invoke /bin/sh -c utilcmd */
          respHandle = popen (utilcmd, "r");

          if (!respHandle) {
             /* when sh -c fails to start */
             logMsg(LEVELE, "popen %s failed - %s\n", NCIDUTIL, strerror(errno));
             safeSend(pos, "ncidutil failed to run" CRLF);
          }
          else {
             strcat(utilcmd, "\n");
             logMsg(LEVEL2, utilcmd);

             logMsg(LEVEL2, BEGIN_DATA2 NL);
             safeSend(pos, BEGIN_DATA2 CRLF);

             strcpy(msgbuf, RESPLINE);
             ptr = msgbuf + sizeof (RESPLINE) - 1;
             cnt = sizeof (msgbuf) - sizeof (RESPLINE);
             while (fgets(ptr, cnt, respHandle))
             {
               logMsg(LEVEL2, msgbuf);
               safeSend(pos, msgbuf);
             }

             status = pclose(respHandle);

             if (status < 0) {
                logMsg(LEVELE, "%s pclose %s\n", NCIDUTIL, strerror(errno));
             }
             else
             {
                if ( WIFEXITED(status) ) {
                   logMsg(LEVEL4, "%s exited with status %d\n", NCIDUTIL, WEXITSTATUS(status) );
                }
                else {
                   logMsg(LEVEL4, "%s abnormal exit\n", NCIDUTIL );
                }
             }
             logMsg(LEVEL4, "End: Executing %s [%s]\n", NCIDUTIL, strdate(ONLYTIME));

             logMsg(LEVEL2, END_RESP NL);
             safeSend(pos, END_RESP CRLF);
             free(utilcmd);
         }
      }
   }
}


/*
 * Process HELLO lines if gateway or client sent HELLO <string> at connect.
 *
 * A message line is permitted without any HELLO lines or after any HELLO lines.
 *
 */
void doHELLO(int pos)
{
  int size, doingHello = 1, got_hello = 0;
  char *sptr, *eptr, buf[BUFSIZ];

  /*
   * use the config sendlog preference, unless this client says otherwise
   * 1- client wants cidcall data, 0- client wants quiet
   */
  sendlog_override = -1;

  while (doingHello)
  {

    usleep(10000); /* Wait 10 ms for connect data */

    /* read data from socket without removing the data from buffer */
    size = recv(polld[pos].fd, buf, BUFSIZ, MSG_PEEK);

    /* Check for error */
    if (size == -1)
    {
      if (errno == EAGAIN)
      {
         logMsg(LEVEL2, "%s%s\n", PRETTYINDENT,
              got_hello ? "End of client HELLO data" :
                          "No client HELLO data received");
      }
      else
      {
         logMsg(LEVEL2, "%srecv returned %d: %s\n", PRETTYINDENT, size, strerror(errno));
         disconnectClient(pos);
         /* a dead client won't need the logs */
         sendlog_override = 0;
      }
      doingHello = 0;
    }

    /* Check for data */
    if (size > 0)
    {
      buf[size] = '\0';

      /* remove whitespace from start and end of string */
      sptr = trimWhitespace(buf, buf + size - 2);


      /* look for the HELLO label in buf */
      if (!strncmp(sptr, HELLO, strlen(HELLO)))
      {
        if ((eptr = strchr(sptr, (int) '\n')))
        {
          if (*(eptr - 1) == '\r') *(eptr - 1) = '\0';
          *eptr = '\0';
          while (*(eptr + 1) == '\n' || *(eptr + 1) == '\r' || *(eptr + 1) == ' ') eptr++;
          size = (eptr - buf) + 1;
        }
        /* hello line found */
        got_hello = 1;
        logMsg(LEVEL2, "%sClient Sent: %s\n", PRETTYINDENT, buf);

        if (!strncmp(buf, COMMAND, strlen(COMMAND)))
        {
          if ((sptr = strstr(buf, "no_log")))
          {
            /* got "no_log" command */
            sendlog_override = 0;
            logMsg(LEVEL2, "%sClient Command Accepted: %s\n", PRETTYINDENT, sptr);

          }
          else if ((sptr = strstr(buf, "send_log")))
          {
            /* got "send_log" command */
            sendlog_override = 1;
            logMsg(LEVEL2, "%sClient Command Accepted: %s\n", PRETTYINDENT, sptr);
          }
          else
          {
            /*  unknown command */
            logMsg(LEVEL2, "%sUnknown Client Command: %s\n", PRETTYINDENT, buf);
          }
        }
        else if (!strncmp(buf, IDENT, strlen(IDENT)))
        {
          doIDENT(pos, buf);
        }
        else
        {
          logMsg(LEVEL2, "%sUnknown HELLO Line: %s\n", PRETTYINDENT, buf);
        }

        /* read and discard HELLO line from socket buffer */
        size = recv(polld[pos].fd, buf, size, 0);
      }
      else
      {
        /* indicate non-hello line found */
        doingHello = 0;
      }
    }
    else
    {
      doingHello = 0;
    }
  }
  return;
}

/*
 * Process IDENT lines if gateway or client sent IDENT <string>.
*/

void doIDENT(int pos, const char * buf)
{
  strncpy(IPinfo[pos].ident, buf + strlen(IDENT), MAXIPBUF -1);
  IPinfo[pos].ident[MAXIPBUF-1] = '\0';
  logMsg(LEVEL2, "%sClient Ident: %s\n", PRETTYINDENT, buf + strlen(IDENT));

  /* needed for "dial" in case it is enabled */
  if (strstr(IPinfo[pos].ident, GWOUT))
  {
    IPinfo[pos].dialout = 1;
    if (dialOUT >= 0) ++dialOUT;
    logMsg(LEVEL2, "%s       Gateway can handle outgoing calls: dialOUT=%d\n",
           PRETTYINDENT, dialOUT);
  }
  else IPinfo[pos].dialout = 0;
}

/*
 * Send important server options to clients
 */
void doOptLine(int pos)
{
    char msgbuf[BUFSIZ];

    logMsg(LEVEL2, "%sSent 'country: %s' option to client\n", PRETTYINDENT, country);
    sprintf (msgbuf, "%scountry: %s\r\n", OPTLINE, country);
    safeSend(pos, msgbuf);

    if (!(cidinput & 0x1))
    {
        /* 0x0 and 0x2 */
        logMsg(LEVEL2, "%sSent LineIDS to client: %s\n", PRETTYINDENT, LineIDS);
        sprintf (msgbuf, "%sLineIDS: %s\r\n", OPTLINE, LineIDS);
        safeSend(pos, msgbuf);
    }
    if (hangup)
    {
        logMsg(LEVEL2, "%sSent 'hangup-%d' option to client\n", PRETTYINDENT, hangup);
        sprintf (msgbuf, "%shangup-%d\r\n", OPTLINE, hangup);
        safeSend(pos, msgbuf);
    }

    if (hupmode)
    {
        logMsg(LEVEL2, "%sSent 'hupmode-%d' option to client\n", PRETTYINDENT, hupmode);
        sprintf (msgbuf, "%shupmode-%d\r\n", OPTLINE, hupmode);
        safeSend(pos, msgbuf);
    }

    if (ignore1)
    {
        logMsg(LEVEL2, "%sSent 'ignore1' option to client\n", PRETTYINDENT);
        sprintf (msgbuf, "%signore1\r\n", OPTLINE);
        safeSend(pos, msgbuf);
    }

    logMsg(LEVEL2, "%sSent 'regex-%d' option to client\n", PRETTYINDENT, regex);
    sprintf (msgbuf, "%sregex-%d\r\n", OPTLINE, regex);
    safeSend(pos, msgbuf);
}

/*
 * Send a ACKLINE.
 */
void sendACK(int pos, const char * buf)
{
    char msgbuf[BUFSIZ];

    logMsg(LEVEL3, "Gateway %d pos %d %s%s %s%s",
            polld[pos].fd, pos, ACKLINE, buf, strdate(WITHSEP), NL);
    sprintf(msgbuf, "%s%s%s", ACKLINE, buf, CRLF);
    safeSend(pos, msgbuf);
}

/*
 * Get or create the INFO from a MSG: or NOT:
 * fills the global struct mesg
 */
void getINFO(const char * bufptr)
{
    int len;
    char * ptr;

    *mesg.date = *mesg.time = '\0';
    len = strlen(bufptr);
    /* this cheats const bufptr here */
    if ((ptr = strstr(bufptr, " ###")) && len > (ptr + 8) - ptr && (*(ptr + 8) == '*'))
    {
        getField(ptr, "DATE", mesg.date, "");
        getField(ptr, "TIME", mesg.time, "");
        getField(ptr, "NAME", mesg.name, NONAME);
        getField(ptr, "NMBR", mesg.nmbr, NONMBR);
        getField(ptr, "LINE", mesg.line, NOLINE);
        getField(ptr, "MTYPE", mesg.type, NOTYPE);
        /* cut the string in bufptr at ### to leave the message part */
        *ptr = '\0';

    }
    else
    {
        /* fill in missing fields */
        strcpy(mesg.name, NONAME);
        strcpy(mesg.nmbr, NONMBR);
        strcpy(mesg.line, NOLINE);
        strcpy(mesg.type, NOTYPE);
    }
    userAlias(mesg.nmbr, mesg.name, mesg.line);

    if (!*mesg.date || !*mesg.time)
    {
        /* no date and time, create both */
        ptr = strdate(NOSEP);
        strncpy(mesg.date, ptr, 8);
        mesg.date[8] = '\0';
        strncpy(mesg.time, ptr + 9, 4);
        mesg.time[4] = '\0';
    }
}

/*
 * Get a INFO field from a MSG: or NOT:
 */
void getField(const char * bufptr, const char * field_name, char * mesgptr, const char * noval)
{
    char tmpbuf[BUFSIZ], *ptr;

    tmpbuf[BUFSIZ -1] = '\0';
    if ((ptr = strstr(bufptr, field_name)))
    {
        if (*(ptr + strlen(field_name) + 1) == '*')
        {
            /* no value for the field, return a default */
            strncpy(mesgptr, noval, CIDSIZE - 1);
        }
        else
        {
            strncpy(tmpbuf, ptr + strlen(field_name) + 1, BUFSIZ -1);
            ptr = strchr(tmpbuf, '*');
            if (ptr) *ptr = '\0';
            strncpy(mesgptr, tmpbuf, CIDSIZE - 1);
        }
    }
    else
    {
        /* field is absent, return a default */
        strncpy(mesgptr, noval, CIDSIZE - 1);
    }
    mesgptr[CIDSIZE-1] = '\0';
}


/*
 * Format of data from NetCallerID for three types of calls and a message
 *
 *      ###DATE03301423...NMBR4075551212...NAMEWIRELESS CALL+++\r
 *      ###DATE03301423...NMBR...NAME-UNKNOWN CALLER-+++\r
 *      ###DATE03301423...NMBR...NAME+++\r
 *      ###DATE...NMBR...NAME   -MSG OFF-+++\r
 */
/*
 * Gateway CALL Line Format:
 *      where type is one of: BLK CID HUP IN MWI OUT PID PUT RID WID
 *
 *      CALL: ###DATEmmddhhss...CALLtype...LINEidentifier...NMBRnumber...NAMEwords+++\r
 */

void gatewayCID(const char * buf)
{
   /* Process Caller ID information from a gateway */
   if (strncmp(buf, "###", 3) == 0)
   {
        /*
         * Found NetCallerID box, or a Gateway
         * All information received on one line
         * The Gateway creates a CID, HUP, OUT, PID, PUT Message Line
         * The Gateway contains a LINE and a CALL field
         * The NetCallerID box does not have a LINE or CALL field
         */
        char cidbuf[BUFSIZ];
        char *ptr;

        logMsg(LEVEL4, "Detected NetCallerID or gateway format\n");

        /* Make sure the status field and cidsent are zero */
        gwcid.calltype = CID;
        gwcid.status = 0;
        gwcid.cidsent = 0;
        /* the serial NetCallerID device has the ncidd.conf lineid */
        strcpy(gwcid.cidline, serial_lineid);

        if ((ptr = strstr(buf, "DATE")))
        {
            if (*(ptr + 4) == '.')
            {
                /* no date and time, create both */
                ptr = strdate(NOSEP);
                strncpy(gwcid.ciddate, ptr, 8);
                gwcid.ciddate[8] = '\0';
                gwcid.status |= CIDDATE;
                strncpy(gwcid.cidtime, ptr + 9, 4);
                gwcid.cidtime[4] = '\0';
                gwcid.status |= CIDTIME;
            }
            else
            {
                strncpy(gwcid.cidtime, ptr + 8, 4); /* HHMM */
                gwcid.cidtime[4] = '\0';
                gwcid.status |= CIDTIME;

                strncpy(gwcid.ciddate, ptr + 4, 4); /* MMDD */
                gwcid.ciddate[4] = '\0';

                /* need to get year */
                ptr = strdate(ONLYYEAR);        /* returns: YYYY */
                strcat(gwcid.ciddate, ptr);       /* MMDDYYYY */

                gwcid.status |= CIDDATE;
            }
        }

        /*
         * this field is only from a Gateway
         * calltype will be one of:
         * CALLBLK, CALLCID, CALLHUP, CALLIN, CALLMWI,
         * CALLOUT, CALLPID, CALLPUT, CALLRID, CALLWID
         */
        if ((ptr = strstr(buf, CALLIN)) || (ptr = strstr(buf, CALLCID)))
        {
            gwcid.calltype = CID; /* this is an incoming call */
        }
        else if ((ptr = strstr(buf, CALLBLK)))
        {
            gwcid.calltype = BLK; /* this is a blocked call */
        }
        else if ((ptr = strstr(buf, CALLHUP)))
        {
            gwcid.calltype = HUP; /* this is a blacklisted call hangup */
        }
        else if ((ptr = strstr(buf, CALLMWI)))
        {
            gwcid.calltype = MWI; /* this indicates voicemail message waiting */
        }
        else if ((ptr = strstr(buf, CALLOUT))) /* taa: added 'else' to beginning */
        {
            gwcid.calltype = OUT; /* this is an outgoing call */
        }
        else if ((ptr = strstr(buf, CALLPID)))
        {
            gwcid.calltype = PID; /* this is a incoming call from a smart phone */
        }
        else if ((ptr = strstr(buf, CALLPUT)))
        {
            gwcid.calltype = PUT; /* this is a outgoing call from a smart phone */
        }
        else if ((ptr = strstr(buf, CALLRID)))
        {
            gwcid.calltype = RID; /* this is a ringback */
        }
        else if ((ptr = strstr(buf, CALLWID)))
        {
            gwcid.calltype = WID; /* this is call waiting */
        }

        if ((ptr = strstr(buf, "LINE")))
        {
            /* this field is only from a Gateway */
            if (*(ptr + 5) == '.')
            {
                /* the gateway said LINE without giving a lineid */
                strcpy(gwcid.cidline, "-");
            }
            else
            {
                strncpy(gwcid.cidline, ptr + 4, CIDSIZE - 1);
                gwcid.cidline[CIDSIZE-1] = '\0';
                ptr = strchr(gwcid.cidline, '.');
                if (ptr) *ptr = '\0';
            }
        }

        if ((ptr = strstr(buf, "NMBR")))
        {
            if (*(ptr + 5) == '.') strcpy(gwcid.cidnmbr, NONMBR);
            else
            {
                strncpy(cidbuf, ptr + 4, BUFSIZ - 1);
                cidbuf[BUFSIZ-1] = '\0';
                ptr = strchr(cidbuf, '.');
                if (ptr) *ptr = '\0';
                builtinAlias(gwcid.cidnmbr, cidbuf);
            }
            gwcid.status |= CIDNMBR;
            /* in case PID type (smartphone gateway (format from "international-only-digit" --> "(in)national-dialable-only-digit" ) ) */
            normalizeToDialableNumberFromFixedLine(gwcid.cidnmbr, CIDSIZE);

        }

        if ((ptr = strstr(buf, "NAME")))
        {
            if (*(ptr + 5) == '+') strcpy(gwcid.cidname, NONAME);
            else
            {
                strncpy(cidbuf, ptr + 4, BUFSIZ - 1);
                cidbuf[BUFSIZ-1] = '\0';
                ptr = strchr(cidbuf, '+');
                if (ptr) *ptr = '\0';
                builtinAlias(gwcid.cidname, cidbuf);
            }
            gwcid.status |= CIDNAME;
        }
   }
}


void handleGWCallinfo(int pos, const char * buf)
{
      /*
       * Found CALLINFO Line
       *
       * CALLINFO Line Format:
       *  CALLINFO: ###CANCEL...DATE%s...SCALL%S...ECALL%s...CALLIN...LINE%s...NMBR%s...NAME%s+++
       *  CALLINFO: ###CANCEL...DATE%s...SCALL%S...ECALL%s...CALLOUT...LINE%s...NMBR%s...NAME%s+++
       *  CALLINFO: ###BYE...DATE%s...SCALL%S...ECALL%s...CALLIN...LINE%s...NMBR%s...NAME%s+++
       *  CALLINFO: ###BYE...DATE%s...SCALL%S...ECALL%s...CALLOUT...LINE%s...NMBR%s...NAME%s+++
       */

  char msgbuf[BUFSIZ];
  char *ptr, *sptr, *eptr, *label;

      writeLog(datalog, buf);
      if (ack[pos]) sendACK(pos, buf);

      /* get end of call date and time */
      label = "DATE";
      if ((sptr = strstr(buf, label)) != NULL)
      {
          /* points to DATEmmddhhmm... */
          sptr += (strlen(label));        /* MMDDHHMM */
          ptr = strdate(ONLYYEAR);        /* returns: YYYY */
          strncpy(endcall.date, sptr, 4); /* MMDD */
          endcall.date[4] = '\0';
          strcat(endcall.date, ptr);      /* MMDDYYYY */

          sptr += (4);                    /* HHMM */
          strncpy(endcall.time, sptr, 4);
          endcall.time[4] = '\0';
      }
      else
      {
          strcpy(endcall.date, "-");
          strcpy(endcall.time, "-");
      }

      /* get end of call start date and extended time */
      label = "SCALL";
      if ((sptr = strstr(buf, label)) != NULL)
      {
          sptr += strlen(label);
          if (!(eptr = strstr(sptr, "...")))
              eptr = strstr(sptr, "+++");
          strncpy(endcall.scall, sptr, eptr - sptr);
          endcall.scall[eptr - sptr] = '\0';
      }
      else  strcpy(endcall.scall, "-");

      /* get end of call end date and extended time */
      label = "ECALL";
      if ((sptr = strstr(buf, label)) != NULL)
      {
          sptr += strlen(label);
          if (!(eptr = strstr(sptr, "...")))
              eptr = strstr(sptr, "+++");
          strncpy(endcall.ecall, sptr, eptr - sptr);
          endcall.ecall[eptr - sptr] = '\0';
      }
      else  strcpy(endcall.ecall, "-");

      /* get end of call type */
      label = ".CALL";
      if ((sptr = strstr(buf, label)) != NULL)
      {
          sptr += strlen(label);
          if (!(eptr = strstr(sptr, "...")))
              eptr = strstr(sptr, "+++");
          strncpy(endcall.ctype, sptr, eptr - sptr);
          endcall.ctype[eptr - sptr] = '\0';
      }
      else  strcpy(endcall.ctype, "-");


      /* get end of call line label */
      label = "LINE";
      if ((sptr = strstr(buf, label)) != NULL)
      {
          sptr += strlen(label);
          if (!(eptr = strstr(sptr, "...")))
              eptr = strstr(sptr, "+++");
          strncpy(endcall.line, sptr, eptr - sptr);
          endcall.line[eptr - sptr] = '\0';
      }
      else
      {
          /* LINE tag with no info */
          strcpy(endcall.line, "-");
      }

      /* get end of call telephone number */
      label = "NMBR";
      if ((sptr = strstr(buf, label)) != NULL)
      {
          sptr += strlen(label);
          if (!(eptr = strstr(sptr, "...")))
              eptr = strstr(sptr, "+++");
          strncpy(endcall.nmbr, sptr, eptr - sptr);
          endcall.nmbr[eptr - sptr] = '\0';
      }
      else  strcpy(endcall.nmbr, "-");

      /* get end of call name */
      label = "NAME";
      if ((sptr = strstr(buf, label)) != NULL)
      {
          sptr += strlen(label);
          if (!(eptr = strstr(sptr, "...")))
              eptr = strstr(sptr, "+++");
          strncpy(endcall.name, sptr, eptr - sptr);
          endcall.name[eptr - sptr] = '\0';
      }
      else  strcpy(endcall.name, "-");

      userAlias(endcall.nmbr, endcall.name, endcall.line);

      /* get and process end of call termination */
      if (strstr(buf, CANCEL))
      {
          strcpy(endcall.htype, CANCEL);
          sendInfo(-1, endcall.line);
      }
      else if (strstr(buf, BYE))
      {
          strcpy(endcall.htype, BYE);
          sendInfo(-2, endcall.line);
      }
      else if (strstr(buf, BUSY))
      {
          strcpy(endcall.htype, BUSY);
          sendInfo(-3, endcall.line);
      }
      else strcpy(endcall.htype, "-");

      /*
       * This sprintf() probably needs the optional blacklist
       * match name for NAME if ncidd does hangup but it is
       * not simple to add because of other possible calls
       * before this one ends.
       */
      sprintf(msgbuf, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
          ENDLINE,
          HTYPE, endcall.htype,
          DATE,  endcall.date,
          TIME,  endcall.time,
          SCALL, endcall.scall,
          ECALL, endcall.ecall,
          CTYPE, endcall.ctype,
          LINE,  endcall.line,
          NMBR,  endcall.nmbr,
          NAME,  endcall.name,
          STAR);

      /* Log the end of call "END:" line */
      writeLog(cidlog, msgbuf);

      /* Send END text line to clients */
      writeClients(msgbuf);
}


/*
 * message from a modem or serial device
 *
 * A USB modem                               (AT+VCID=1)
 *        call modemCID()
 *
 * a USB Modem sending MDMF or SDMF messages (AT+VCID=2)
 *        call xdmfCID(buf)
 *
 * serial TCI multi-pots-line devices
 *        call TCIserial(buf)
 */
void formatCID(char ** lines)
{
   const char * buf;

   if (cidinput == 0)
   {
      /* modem connected */
      while (*lines != NULL) {
         buf = *lines;
         if (!strncmp(buf, "80", 2) || /* MDMF CID */
             !strncmp(buf, "04", 2) || /* SDMF CID */
             !strncmp(buf, "82", 2))   /* MDMF MWI */
         {
             /* USB Modem sending MDMF or SDMF messages (AT+VCID=2) */
             xdmfCID(buf);
         }
         else
         {
             /* serial or USB modem sending multiline CID */
             modemCID(buf);
         }
         /* logMsg(LEVEL5, "free %p\n", *lines); */
         free(*lines);
         lines++;
      }
      finishCID(&modem_cid);
   }

   else if (cidinput == 0x1)
   {
      /* serial device connected */
      while (*lines != NULL) {
         buf = *lines;
         if (strlen(buf) == 70 &&
             isdigit((int) *buf) &&
             *(buf + 9) == '/' &&
             *(buf + 24) == 'M')
         {
             /* Found TCI serial device */
             TCIserial(buf);
             finishCID(&modem_cid);
         }

         else if (strncmp(buf, "###", 3) == 0)
         {
             /* Found a serial NetCallerID device */
             gatewayCID(buf);
             finishCID(&gwcid);
         }
         free(*lines);
         lines++;
      }
   }

   else  /* cidinput 2 and 3 */
   {
      /* serial device is being ignored */
      while (*lines != NULL) {
         free(*lines);
         lines++;
      }
   }
}


/*
 * a message from a modem
 *
 * Format of data from modem:
 *
 *      DATE = 0330            -or-  DATE=0330
 *      TIME = 1423            -or-  TIME=1423
 *      NMBR = 4075551212      -or-  NMBR=4075551212
 *      NAME = WIRELESS CALL   -or-  NAME=WIRELESS CALL
 *
 * kortex Modems  France Telecom:
 *      CID RING
 *
 *      CALLING MSG
 *
 *      DATE TIME=11/24 22:10
 *
 *      NBR=0632727751
 *
 *      END MSG
 *
 */
void modemCID(const char * buf)
{
    const char * ptr;


    /*
     * At a RING
     * Could be "RING", "RING X" or "CID RING"
     *
     * US systems send Caller ID between the 1st and 2nd ring
     * Some non-US systems send Caller ID before 1st ring.
     * e.g. BT in UK, KPN in Netherlands
     *
     * If generate Caller ID set and Caller ID not received, generate
     * a generic Caller ID at RING 2.
     *
     * Clear Caller ID info between rings.
     */
    if (strstr(buf, "RING"))
    {
        /*
         * If distinctive ring, save line indicator, it will be
         * "RING A", "RING B", etc.
         * append the letter to the lineid
         * RING A/B is discarded if the system sends CID ahead of the 1st RING
         */
         if (ringcnt==0)
         {
             if (strlen(buf) == 6)
             {
                 strcpy(modem_cid.cidline, serial_lineid);
                 strcat(modem_cid.cidline, buf + 5);
             }
             else
             {
                 strcpy(modem_cid.cidline, serial_lineid);
             }
         }
         modem_cid.calltype = CID;

        /*
         * If ring information is wanted, send it to the clients.
         */
        ++ringcnt;
        sendInfo(ringcnt, modem_cid.cidline);
        /* timer to announce that ringing has stopped */
        ringtimer = time(0) + RINGWAIT;

        if (modem_cid.status != 0)
        {
            /* caller ID has arrived before this RING */
            if (!(modem_cid.status & CIDDATE) ||
                !(modem_cid.status & CIDTIME)) {
               /* generate date and time, if missing */
               char * sptr;
               ptr = strdate(NOSEP);     /* returns: MMDDYYYY HHMM */
               for (sptr = modem_cid.ciddate; *ptr && *ptr != ' ';) *sptr++ = *ptr++;
               *sptr = '\0';
               for (sptr = modem_cid.cidtime, ptr++; *ptr;) *sptr++ = *ptr++;
               *sptr = '\0';
               modem_cid.status |= (CIDDATE | CIDTIME);
            }
            finishCID(&modem_cid);
        }
        else if (gencid && modem_cid.cidsent == 0 && ringcnt == 2 )
        {
            /*
             * gencid = 1: generate a Caller ID if it is not received
             * gencid = 0: do not generate a Caller ID
             *
             * CID information always received between before RING 2
             * no CID information received, so create one.
             */
            char * sptr;
            ptr = strdate(NOSEP);     /* returns: MMDDYYYY HHMM */
            for (sptr = modem_cid.ciddate; *ptr && *ptr != ' ';) *sptr++ = *ptr++;
            *sptr = '\0';
            for (sptr = modem_cid.cidtime, ptr++; *ptr;) *sptr++ = *ptr++;
            *sptr = '\0';
            strcpy(modem_cid.cidnmbr, "RING");
            strcpy(modem_cid.cidname, NOCID);
            modem_cid.status = (CIDDATE | CIDTIME | CIDNMBR | CIDNAME);
            /* process the generated caller ID */
            finishCID(&modem_cid);
        }
        else
        {
            /*
             * CID not here yet or already processed
             * Make sure status and cidsent are zero
             */
            modem_cid.status = 0;
            modem_cid.cidsent = 0;
            return;
        }
    } /* RING */

    /*
     * A Mac Mini Motorola Jump  modem sends "^PR^PX\n\n" before the CID
     * information.  It also sends "^P.^XNAME"
     */
    else if ((ptr = strstr(buf, "\020R\020X")))
    {
        modem_cid.status = 0;
        modem_cid.cidsent = 0;
        strcpy(modem_cid.cidline, serial_lineid);
    }

    else if (strncmp(buf, "DATE", 4) == 0)
    {
        if ((ptr = strchr(buf, '='))) ++ptr;
        else ptr = buf + 4; /* there was no equals */
        if (*ptr == ' ') ++ptr;

        /* NetTalk sends no date but modem has line: "DATE=" */
        if (*ptr)
        {
            if (strstr(buf, "TIME"))
            {
                /* kortex Modems send
                 * DATE TIME=11/24 22:10 */
                strncpy(modem_cid.ciddate, ptr, 2);
                strncpy(modem_cid.ciddate + 2, ptr + 3, 2);
                modem_cid.ciddate[4] = '\0';
                strncpy(modem_cid.cidtime, ptr + 6, 2);
                strncpy(modem_cid.cidtime + 2, ptr + 9, 3);
                modem_cid.status |= CIDTIME;
            }
            else
            {
                /* Conexant and other modems send
                 * DATE = 0330 */
                strncpy(modem_cid.ciddate, ptr, CIDSIZE - 1);
                modem_cid.ciddate[CIDSIZE-1] = '\0';
            }

            /* need to get year */
            ptr = strdate(ONLYYEAR);        /* returns: YYYY */
            strcat(modem_cid.ciddate, ptr);       /* MMDDYYYY */

            modem_cid.status |= CIDDATE;
        }
        modem_cid.cidsent = 0;
    }
    else if (strncmp(buf, "TIME", 4) == 0)
    {
        if ((ptr = strchr(buf, '='))) ++ptr;
        else ptr = buf + 4; /* there was no equals */
        if (*ptr == ' ') ++ptr;
        if (*ptr)
        {
            /* NetTalk sends no time but modem has line: "TIME=" */
            strncpy(modem_cid.cidtime, ptr, CIDSIZE - 1);
            modem_cid.cidtime[CIDSIZE-1] = '\0';
            modem_cid.status |= CIDTIME;
        }
        modem_cid.cidsent = 0;
    }
    /*
     * Some modems send DDN_NMBR, DDN, or NBR instead of NMBR.
     * This will catch the 4 cases.
     */
    else if (!strncmp(buf, "NMBR", 4) ||
             !strncmp(buf, "DDN", 3) ||
             !strncmp(buf, "NBR", 3))
    {
        /* some telcos send NMBR = ##########, then NMBR = O to mask it */
        if (!(modem_cid.status & CIDNMBR))
        {
            if ((ptr = strchr(buf, '='))) ++ptr;
            else while (*ptr && !isblank((int) *ptr)) ++ptr; /* this should never happen */
            if (*ptr == ' ') ++ptr;
            builtinAlias(modem_cid.cidnmbr, ptr);
            modem_cid.status |= CIDNMBR;
            modem_cid.cidsent = 0;
        }

        /*
         * Check if this is a Call Waiting CID (WID).
         * If CID arrives more than 200ms before RING, or more than 6 seconds after RING
         * and modem snooping is on, it will marked as a WID.
         */
        if (widflag == 1 && ringcnt == 0)
        {
           logMsg(LEVEL4, "Number without RING is Call Waiting (WID)\n");
           modem_cid.calltype = WID;
        }
    }
    /*
     * Using strstr() instead of strncmp() because a Mac
     * Mini Jump modem sent '^P.^PXNAME' instead of 'NAME'.
     * At this point the string was converted to '?.?XNAME'
     */
    else if ((ptr = strstr(buf, "NAME")))
    {
        /*
         * if NAME already received, discard the second one
         */
        if (!(modem_cid.status & CIDNAME))
        {
            char * sptr;
            char tempname[100];

            if (*(ptr + 4) == '=') ptr += 5;
            else ptr += 7;
            if (*ptr == ' ') ++ptr;
            /* remove any trailing spaces */
            strncpy(tempname, ptr, sizeof(tempname));
            tempname[sizeof(tempname)-1]=  '\0';
            for (sptr = tempname; *sptr; ++sptr);
            for (--sptr; *sptr && *sptr == ' '; --sptr) *sptr = '\0';

            builtinAlias(modem_cid.cidname, tempname);
            modem_cid.status |= CIDNAME;
            modem_cid.cidsent = 0;
        }
    }
    else if (strncmp(buf, "MESG", 4) == 0)
    {
        int curlen, addlen;

        ptr = buf;
        if (*(ptr + 4) == '=') ptr += 5;
        else ptr += 7;
        if (*ptr == ' ') ++ptr;
        /* Comb out some of the noise where the modem has not handled valid info.
         * The modems appear to process these
         *     040150 - P private
         *     04014F - O unavailable/out-of-area BT/Bell
         * but don't process
         *     110101 - BT voice call
         *     110102 - BT ring-back-when-free call
         *     110181 - BT message waiting call
         *     06014C - Bell Canada Long Distance
         */
        while (*ptr != '\0') {
           if (!strncmp(ptr, "110101", 6)) {
              ptr += 6;
              logMsg(LEVEL3, "110101 is a BT voice call\n");
           }
           else if (!strncmp(ptr, "110102", 6)) {
              ptr += 6;
              logMsg(LEVEL3, "110102 is a BT ring-back-when-free call\n");
           }
           else if (!strncmp(ptr, "110181", 6)) {
              ptr += 6;
              logMsg(LEVEL3, "110181 is a BT message waiting call\n");
           }
           else if (!strncmp(ptr, "06014C", 6)) {
              ptr += 6;
              logMsg(LEVEL3, "06014C is Bell Canada Long Distance\n");
           }
           /* stop if the code is unknown, though it should be code-length-content */
           else break;
        }
        addlen = strlen(ptr);
        if (addlen > 0) {
           if (!(modem_cid.status & CIDMESG))
           {
              if (CIDSIZE - addlen > 0) {
                 strcpy(modem_cid.cidmesg, ptr);
              }
              else {
                 logMsg(LEVEL3, "MESG discarded %s\n", ptr);
              }
           }
           else
           {
              /* there is already a previous code MESG */
              curlen = strlen(modem_cid.cidmesg);
              if (CIDSIZE - (curlen + addlen + 2) > 0) {
                 modem_cid.cidmesg[curlen] = ',';
                 strcpy(modem_cid.cidmesg+curlen+1, ptr);
              }
              else {
                 logMsg(LEVEL3, "MESG discarded %s\n", ptr);
              }
           }
           modem_cid.cidmesg[CIDSIZE-1] = '\0';
           modem_cid.status |= CIDMESG;
           modem_cid.cidsent = 0;
        }
    }
}


/*
 * TCI serial device
 * All information received on one line
 * Only inbound calls
 * No MESG field
 *
 * Format of data from serial TCI multi-pots-line devices. Should also
 * work for serial Whozz Calling devices that have the output format
 * (usually switch 7) set to ON for TCI.
 *
 * Fixed field format, should always 70 characters per line.
 *
 *                1         2         3         4         5         6
 *      0123456789012345678901234567890123456789012345678901234567890123456789
 *      01      9/03      2:00 PM            PRIVATE
 *      02      9/03      2:15 PM            PRIVATE           PRIVATE
 *      03      9/03      2:23 PM       407-555-1212           WIRELESS CALLER
 */
void TCIserial(const char * buf)
{
    char cidbuf[BUFSIZ], tmpbuf[BUFSIZ];
    char *ptr, *sptr, *tptr;
    int i;

        logMsg(LEVEL4, "Detected TCI serial device format\n");

        /* Make sure the status field and cidsent are zero */
        modem_cid.status = 0;
        modem_cid.cidsent = 0;

        /* copy data to working buffer and init pointer */
        strncpy(cidbuf, buf, BUFSIZ - 1);
        cidbuf[BUFSIZ-1] = '\0';
        ptr = cidbuf;

       /*
        * TELEPHONE LINE
        * positions: 0-1
        */
        *(ptr + 2) = '\0';
        /* if the lineid is still the default, use the line number */
        if (strcmp(serial_lineid, ONELINE) == 0)
        {
            strcpy(modem_cid.cidline, ptr);
        }
        else {
            /* otherwise append the line number */
            strcpy(modem_cid.cidline, serial_lineid);
            strcat(modem_cid.cidline, ptr);
        }

        /*
         * MONTH
         * positions 7-8
         * force leading zero
         */
        *(ptr + 9) = '\0';
        sprintf(modem_cid.ciddate, "%02d", atoi(ptr + 7));

        /*
         * DAY
         * positions 10-11
         */
        *(ptr + 12) = '\0';
        strcat(modem_cid.ciddate, ptr + 10);

        /* need to get year */
        tptr = strdate(ONLYYEAR);        /* returns: YYYY */
        strcat(modem_cid.ciddate, tptr);       /* MMDDYYYY */

        modem_cid.status |= CIDDATE;

        /*
         * HOUR
         * positions 17-18
         * AM/PM positions 23-24
         */
        *(ptr + 19) = '\0';
        i = atoi(ptr + 17);
        if (strncmp(ptr + 23, "PM", 2) == 0) { if (i < 12) i = i + 12; }
        else if (i == 12) i = 0;
        sprintf(modem_cid.cidtime, "%02d", i);

        /*
         * MINUTE
         * positions 20-21
         */
        *(ptr + 22) = '\0';
        strcat(modem_cid.cidtime, ptr + 20);
        modem_cid.status |= CIDTIME;

        /*
         * NMBR
         * positions 29-43
         * right justified (15 characters)
         */
        *(ptr + 44) = '\0';
        sptr = trimWhitespace(ptr + 29, ptr + 43);
        if (isdigit((int) *sptr))
        {
            /* remove hyphens from a number, builtinAlias() not needed */
            tptr = tmpbuf;
            while(*sptr)
            {
                if (*sptr != '-') *tptr++ = *sptr++;
                else ++sptr;
            }
            *tptr = '\0';
            strncpy(modem_cid.cidnmbr, tmpbuf, CIDSIZE - 1);
            modem_cid.cidnmbr[CIDSIZE-1] = '\0';
        }
        else builtinAlias(modem_cid.cidnmbr, sptr);
        modem_cid.status |= CIDNMBR;

        /*
         * NAME
         * positions 55-69
         * left justified (15 characters)
         */
        *(ptr + 54) = '\0';
        sptr = trimWhitespace(ptr + 55, ptr + 69);
        if (*sptr) builtinAlias(modem_cid.cidname, ptr + 55);
        else strcpy(modem_cid.cidname, NONAME);
        modem_cid.status |= CIDNAME;
}


/*
 * USB Modem sending MDMF or SDMF messages (AT+VCID=2)
 * All information received as one HEX string
 * Inbound calls (MDMF and SDMF) and Message Waiting (MDMF)
 * No MESG field
*/
void xdmfCID(const char * buf)
{
    unsigned char ucbuf[BUFSIZ], *ptr, chksum, reclen, nmss;
    char *tptr;
    size_t msglen;

      logMsg(LEVEL4, "Detected MDMF or SDMF message\n");

      bzero(ucbuf, BUFSIZ);
      hex2uchar(buf, ucbuf);
      if (verbose >= HEXLEVEL) hexdump(ucbuf, strlen(buf) / 2);

      ptr = ucbuf;
      modem_cid.calltype = CID;
      strcpy(modem_cid.cidline, serial_lineid);
      modem_cid.cidsent = 0;

      chksum = checksum(ptr, (msglen = strlen(buf) / 2));
      logMsg(LEVEL4, "Checksum %s: 0x%.2x\n", chksum ? "bad" : "good", chksum);

      modem_cid.ciddate[0] = modem_cid.cidtime[0] = modem_cid.cidname[0] = modem_cid.cidnmbr[0] = '\0';

      if (*ptr == SDMF_MESG_TYPE)
      {
        logMsg(LEVEL4, "Detected SDMF message\n");

        /* SDMF date and time are at fixed byte positions */
        reclen = 8;

        /* start at second byte */
        ptr += 2;

        strncat(modem_cid.ciddate, (char *) ptr, 4);
        modem_cid.status |= CIDDATE;

        strncat(modem_cid.cidtime, (char *) ptr + 4, 4);
        modem_cid.status |= CIDTIME;

        logMsg(LEVEL4, "Date & Time: %s %s\n", modem_cid.ciddate, modem_cid.cidtime);

        /* need to get year */
        tptr = strdate(ONLYYEAR);        /* returns: YYYY */
        strcat(modem_cid.ciddate, tptr);       /* MMDDYYYY */

        ptr += reclen;

        /*
         * SDMF has NMBR only, no NAME
         * starts at a fixed position
         * length varies
         */

        strcpy(modem_cid.cidname, NONAME);
        modem_cid.status |= CIDNAME;

        reclen = msglen - reclen - 3;

        strncat(modem_cid.cidnmbr, (char *) ptr, reclen);
        modem_cid.status |= CIDNMBR;

        if (reclen == 1)
        {
          logMsg(LEVEL4, "Why No Number: %c\n", *ptr);
          switch (*ptr)
          {
             case LETTER_P:
               strcpy(modem_cid.cidnmbr, WORDFOR_P);
               break;
             case LETTER_O:
               strcpy(modem_cid.cidnmbr, WORDFOR_O);
               break;
          }
        }

        logMsg(LEVEL4, "Calling Number: %s\n", modem_cid.cidnmbr);

        logMsg(LEVEL4, "Message Checksum: 0x%.2x\n", *(ucbuf + msglen - 1));
      }
      else
      {
        logMsg(LEVEL4, "Detected MDMF message\n");

        /* Check packet ID  for call type*/
        if (*ptr == MDMF_MSG_WAITING) modem_cid.calltype = MWI;

        /* start at second parameter */
        ptr += 2;

        while (*ptr != 0)
        {
          switch (*ptr)
          {
            case MDMF_VISUAL_IND:
              logMsg(LEVEL4, "Parameter 0x%.2x: US Visual Indicator\n", *ptr);
              ++ptr;
              reclen = *ptr++;
              if (*ptr) strcpy(modem_cid.cidname, "Message(s) Waiting");
              else strcpy(modem_cid.cidname, "No Messages Waiting");
              strcpy(modem_cid.cidnmbr, "Message");
              modem_cid.status = CIDNAME | CIDNMBR;
              ptr += reclen;
              break;
            case CALL_TYPE_UK:
              ++ptr;
              reclen = *ptr++;
              logMsg(LEVEL4, "Parameter 0x%.2x: UK Call Type 0x%.2x; ",
                     *(ptr - 2), *ptr);
              switch (*ptr)
              {
                case VOICE_UK:
                  modem_cid.calltype = CID;
                  logMsg(LEVEL4, "CallType 0x%.2x\n", modem_cid.calltype);
                  break;
                case RINGBACK_UK:
                  modem_cid.calltype = RID;
                  logMsg(LEVEL4, "CallType 0x%.2x\n", modem_cid.calltype);
                  break;
                case MSG_WAITING_UK:
                  modem_cid.calltype = MWI;
                  ++ptr;
                  nmss = *ptr++;
                  reclen = *ptr++;
                  switch (*ptr)
                  {
                    case 0:
                      strcpy(modem_cid.cidname, "No Messages Waiting");
                      break;
                    case 1:
                      strcpy(modem_cid.cidname, "1 Message Waiting");
                      break;
                    default:
                      sprintf(modem_cid.cidname, "%u Messages Waiting", *ptr);
                      break;
                  }
                  logMsg(LEVEL4, "CallType 0x%.2x; NMSS 0x%.2x\n",
                         modem_cid.calltype, nmss);
                  strcpy(modem_cid.cidnmbr, "Voicemail");
                  modem_cid.status = CIDNAME | CIDNMBR;
                  break;
              }
              ptr += reclen;
              break;
            case MDMF_DATETIME:
              logMsg(LEVEL4, "Parameter 0x%.2x: Date & Time\n", *ptr);
              ++ptr;
              reclen = *ptr++;
              strncat(modem_cid.ciddate, (char *) ptr, 4);
              modem_cid.status |= CIDDATE;

              strncat(modem_cid.cidtime, (char *) ptr + 4, 4);
              modem_cid.status |= CIDTIME;

              /* need to get year */
              tptr = strdate(ONLYYEAR);        /* returns: YYYY */
              strcat(modem_cid.ciddate, tptr);       /* MMDDYYYY */

              ptr += reclen;
              break;
            case MDMF_CALLER_NMBR:
              logMsg(LEVEL4, "Parameter 0x%.2x: Calling Number\n", *ptr);
              ++ptr;
              reclen = *ptr++;
              if (reclen)
              {
                strncat(modem_cid.cidnmbr, (char *) ptr, reclen);
                modem_cid.status |= CIDNMBR;
              }
              ptr += reclen;
              break;
            case MDMF_WHY_NO_NMBR:
              logMsg(LEVEL4, "Parameter 0x%.2x: Why No Number\n", *ptr);
              ++ptr;
              reclen = *ptr++;
              if (*ptr == LETTER_P) strcpy(modem_cid.cidnmbr, WORDFOR_P);
              else if (*ptr == LETTER_O) strcpy(modem_cid.cidnmbr, WORDFOR_O);
              else strcpy(modem_cid.cidnmbr, NONMBR);
              modem_cid.status |= CIDNMBR;
              ++ptr;
              break;
            case MDMF_CALLER_NAME:
              logMsg(LEVEL4, "Parameter 0x%.2x: Calling Name\n", *ptr);
              ++ptr;
              reclen = *ptr++;
              if (reclen)
              {
                strncat(modem_cid.cidname, (char *) ptr, reclen);
                modem_cid.status |= CIDNAME;
              }
              ptr += reclen;
              break;
            case MDMF_WHY_NO_NAME:
              logMsg(LEVEL4, "Parameter 0x%.2x: Why No Name\n", *ptr);
              ++ptr;
              reclen = *ptr++;
              if (*ptr == LETTER_P) strcpy(modem_cid.cidname, WORDFOR_P);
              else if (*ptr == LETTER_O) strcpy(modem_cid.cidname, WORDFOR_O);
              else strcpy(modem_cid.cidname, NONAME);
              modem_cid.status |= CIDNAME;
              ++ptr;
              break;
            default:
              ++ptr;
              if (ptr == ucbuf + msglen)
              {
                logMsg(LEVEL4, "Parameter 0x%.2x: Message Checksum\n",
                       *(ptr - 1));
              }
              else
              {
                logMsg(LEVEL4, "Parameter 0x%.2x: Unknown\n", *(ptr - 1));
                /* skip parameter data */
                reclen = *ptr++;
                ptr += reclen;
              }
              break;
          }
        }

        /* make sure there is a date, time, number and name */
        if (modem_cid.cidnmbr[0] == '\0')
        {
          strcpy(modem_cid.cidnmbr, NONMBR);
          modem_cid.status |= CIDNMBR;
        }
        if (modem_cid.cidname[0] == '\0')
        {
          strcpy(modem_cid.cidname, NONAME);
          modem_cid.status |= CIDNAME;
        }
        if (modem_cid.ciddate[0] == '\0' || modem_cid.cidtime[0] == '\0')
        {
          tptr = strdate(NOSEP);
          strncat(modem_cid.ciddate, tptr, 8);
          strncat(modem_cid.cidtime, tptr + 9, 4);
          modem_cid.status |= (CIDDATE | CIDTIME);
        }
      }
}


/* see if all of the call information has arrived */
void finishCID(struct cid * fcid)
{
    const char * ptr;
    const char * linelabel;
    const char * nameptr;
    const char * announceptr;

   if (nothex(fcid->cidmesg) && fcid->status == (CIDALL3 | CIDMESG))
    {
        /*
         * date, time, nmbr and mesg were received
         * name is in the mesg field
         * copy mesg to name
         */
        fcid->status |= CIDNAME;
        strncpy(fcid->cidname, fcid->cidmesg, CIDSIZE - 1);
        fcid->cidname[CIDSIZE-1] = '\0';
        logMsg(LEVEL4, "received date, time, nmbr, mesg\n");
    }

    if (CIDALL3 == (fcid->status & IGNHIGH))
    {
        /* date, time and nmbr were received, set No NAME */
        fcid->status |= CIDNAME;
        strcpy(fcid->cidname,  NONAME);
        logMsg(LEVEL4, "received date, time, nmbr\n");
    }

    if (CIDALT3 == (fcid->status & IGNHIGH))
    {
        /* date, time and name were received, set No Number */
        fcid->status |= CIDNMBR;
        strcpy(fcid->cidnmbr,  NONMBR);
        logMsg(LEVEL4, "received date, time, name\n");
    }

    if (fcid->status == CIDNODT || fcid->status == (CIDNODT | CIDMESG))
    {
        /*
         * number and name received but no date and time
         * mesg may have been received
         */

        logMsg(LEVEL4,"No date and time found -- using current\n");
        ptr = strdate(NOSEP);           /* returns: MMDDYYYY HHMM */
        strncpy(fcid->ciddate, ptr, 8);   /*          0123456789012 */
        fcid->ciddate[8] = '\0';
        strncpy(fcid->cidtime, ptr + 9, 4);
        fcid->cidtime[4] = '\0';
        fcid->status |= (CIDDATE | CIDTIME);
    }

    if (CIDALL4 == (fcid->status & IGNHIGH))
    {
        /*
         * All Caller ID or outgoing call information received.
         *
         * Create the proper line type text line (CID, HUP, OUT, etc.)
         *
         * For OUT text lines (outgoing calls):
         *     the MESG field is not used
         *     the NAME field will be generic if no alias
         *
         * For HUP server generated text lines (hungup call):
         *     the CID label is replaced by a HUP label
         */

       logMsg(LEVEL4, "received %s\n", fcid->status == CIDALL4 ?
           "date, time, nmbr, name" : "date, time, nmbr, name, mesg");

        userAlias(fcid->cidnmbr, fcid->cidname, fcid->cidline);

        /* get additional info from phone number */
        if (fcid->infoBuffer != NULL) free(fcid->infoBuffer);
        fcid->infoBuffer = GetCompleteInfoFromNumber(fcid->cidnmbr);

        linelabel = getLineLabel(fcid->calltype);

        nameptr = fcid->cidname;
        announceptr = annpath;

        /* determine if a hangup check is wanted on proper calltype */
        if ((hangup || hupmode) && fcid->calltype == CID)
        {
            /* can only hangup on CID calls */
            logMsg(LEVEL3,
              "calltype=CID hangup=%d hupmode=%d cidline=%s lineid=%s\n",
                 hangup, hupmode, fcid->cidline, serial_lineid);
            /* fcid->cidline must match serial_lineid to hangup */
            if (strcmp(fcid->cidline, serial_lineid) == 0)
            {
                logMsg(LEVEL3, "cidline == lineid: check whitelist\n");
                if (checkWhitelist(fcid->cidname, fcid->cidnmbr))
                {
                   /* the name or number was on the whitelist */
                    logMsg(LEVEL3,
                      "Name or number on the whitelist: skip hangup check\n");
                   /* add the name found after #= in the whitelist */
                   if (listname) nameptr = listname;
                }
                else
                {
                    int dohangup = 0;
                    /* it wasn't on the white list */
                    if (hangup && checkBlacklist(fcid->cidname, fcid->cidnmbr))
                    {
                       /*
                       * hangup phone if on blacklist but not whitelist
                       */
                       dohangup = hangup;
                       /* add the name found after #= in the blacklist */
                       if (listname) nameptr = listname;
                    }
                    else if (hupmode)
                    {
                       /*
                       * hangup phone if external hangup extension says to
                       */
                       dohangup = runExternalHangup(&linelabel, fcid, &nameptr, &announceptr);
                       /* dohangup = hupmode, but may have been downgraded
                        * nameptr may have been extended with the hangup reason
                        * announceptr = rmdpath -or- hermd
                        */
                    }

                    if (dohangup > 0)
                    {
                       /* try to hangup the call */
                       logMsg(LEVEL5, "Processing hangup request\n");
                       if (hangupCall_offhook(dohangup) == 0)
                       {
                          /* change CID to HUP if it can go offhook */
                          linelabel = HUPLINE;
                          fcid->calltype = HUP;
                          sendCall(linelabel, fcid, nameptr);
                          fcid->cidsent = 1;
                          if (hangupCall(dohangup, announceptr) == 0)
                          {
                             /* make the ring timer go off right away to reset the cid */
                             if (ringtimer) ringtimer = 1;
                          }
                          else
                          {
                             logMsg(LEVEL3, "hangupCall may have left line off hook\n");
                          }
                          /* indicate hangup completed even if hangup failed for modems providing CID */
                          sendInfo(-4, fcid->cidline);
                       }
                       else
                       {
                          logMsg(LEVEL3, "hangupCall could not go off hook\n");
                       }
                    }
                 }
            }
            else logMsg(LEVEL3, "cidline != lineid: skip hangup check\n");
        }

        /* log the call and tell the clients/gateways */
        if (!fcid->cidsent)
           sendCall(linelabel, fcid, nameptr);

        /* Reset caller id info in modem_cid */
        strcpy(fcid->cidmesg, NOMESG); /* default message */
        strcpy(fcid->cidline, serial_lineid); /* default line indicator */
        fcid->status = 0;
        fcid->cidsent = 1;  /* don't do gencid on 2nd RING */
        fcid->calltype = 0;
        if (fcid->infoBuffer!=NULL) free(fcid->infoBuffer);
        fcid->infoBuffer = NULL;
    }
    else if (fcid->status != 0) {
       logMsg(LEVEL2, "finishCID did not receive full call info 0x%04x\n", fcid->status);
    }
}


/*
 * call the external hangup script(s) to see if hangup is wanted
 */
int runExternalHangup(const char ** linelabel, const struct cid * ecid, const char ** nameptr, const char ** rmdptr)
{
   char msgbuf[BUFSIZ], exthupcmd[BUFSIZ+(5*CIDSIZE)+60+strlen(ecid->infoBuffer)];
   char *ptr, *sptr;

   FILE *respHandle;
   int len, status, ehupmode, rethupmode;
   char *eptr;
   /* for the announcement file name */
   static char hermd[BUFSIZ];
   /* for adding the hangup reason to the caller name */
   static char hreason[BUFSIZ];

   /*
    * run an external script to decide if hangup is wanted
    */
   sprintf (exthupcmd, DOEXTEND, huppath, ecid->ciddate, ecid->cidtime,
            ecid->cidline, ecid->cidnmbr, hupmode, ecid->infoBuffer?ecid->infoBuffer:" ", ecid->cidname);
   logMsg(LEVEL4, "Begin: Executing %s [%s]\n", hupname, strdate(ONLYTIME));

   /* invoke /bin/sh -c exthupcmd */
   respHandle = popen (exthupcmd, "r");

   if (!respHandle) {
      /* when sh -c fails to start */
      logMsg(LEVELE, "popen %s failed - %s\n", hupname, strerror(errno));
      return 0; /* don't hang up */
   }

   logMsg(LEVEL3, "%s\n", exthupcmd);
   logMsg(LEVEL3, "default hangup mode for %s: %d\n", hupname, hupmode);

   ehupmode = hupmode;  /* effective hangup mode */
   *rmdptr = rmdpath;   /* default announcement for extensions */
   rethupmode = -1;     /* return hangup mode */

   /* read the lines returned from running the extension */
   while (fgets(msgbuf, sizeof(msgbuf), respHandle))
   {
      logMsg(LEVEL3, "%s: %s", hupname, msgbuf);
      if (ehupmode == 3)
      {
         /* check for recording name in hangup extension response */
         if ((ptr = strstr(msgbuf, RECFILE)))
         {
            /* skip over RECFILE */
            sptr = ptr + strlen(RECFILE);
            /* eliminate spaces between RECFILE and recording file name */
            while (sptr && *sptr == ' ') sptr++;
            if ((len = strlen(sptr) - 1))
            {
               /* eliminate '\n' at end of string */
               *(eptr = sptr + len) = '\0';
               /* eliminate spaces at the end of the recording file name */
               for (--eptr; *eptr && *eptr == ' '; --eptr) *eptr = '\0';

               /* if the extension doesn't give a valid recording, just hang up */
               checkVoiceFile(&ehupmode, hermd, sptr);
               *rmdptr = hermd;
            }
            else
            {
               /* if the extension doesn't give a valid recording, just hang up */
               ehupmode = 1;
               logMsg(LEVEL3, "WARNING: no recording file, using Normal Hangup (mode 1)\n%s\n", huppath);
            }
         }
      }

      /* check for hangup reason in hangup extension response */
      if ((ptr = strstr(msgbuf, HUPREASON)))
      {
         /* skip over HUPREASON */
         sptr = ptr + strlen(HUPREASON);
         /* eliminate spaces between HUPREASON and reason text */
         while (sptr && *sptr == ' ') sptr++;
         if ((len = strlen(sptr) - 1))
         {
            /* eliminate '\n' at end of string */
            *(eptr = sptr + len) = '\0';
            /* eliminate spaces at the end of the reason text */
            for (--eptr; *eptr && *eptr == ' '; --eptr) *eptr = '\0';
            /* add the reason to the caller name */
            snprintf(hreason, sizeof(hreason)-1, "%s - %s", *nameptr, sptr);
            /* make sure name and reason does not exceed "CIDSIZE" */
            hreason[CIDSIZE - 1] = '\0';

            *nameptr = hreason;
         }
      }

      /* extension script responds OK or "hangup" */
      if (!strncmp(msgbuf, HUPNAME, strlen(HUPNAME)))
      {
         logMsg(LEVEL3, "using hangup mode: %d\n", ehupmode);
         if (ehupmode == 3)
            logMsg(LEVEL3, "using recording: %s\n", strrchr(*rmdptr, (int) '/'));
         rethupmode = ehupmode;
         break;
      }
      else if (strstr(msgbuf, "OK"))
      {
         rethupmode = 0;  /* don't hang up */
         break;
      }
   }
   if (rethupmode < 0) {
      logMsg(LEVELE, "%s did not return hangup or OK\n", hupname);
      rethupmode = 0;  /* don't hang up */
   }

   status = pclose(respHandle);

   if (status < 0) {
      logMsg(LEVELE, "%s pclose %s\n", hupname, strerror(errno));
   }
   else
   {
      if ( WIFEXITED(status) ) {
         if (WEXITSTATUS(status) != 0)
            logMsg(LEVELE, "%s exited with status %d\n", hupname, (char)(WEXITSTATUS(status)) );
         if (WEXITSTATUS(status) == 127 ||        /* extension not found */
             WEXITSTATUS(status) == 126 ||        /* extension not executable */
             (char)(WEXITSTATUS(status)) < 0) {
            rethupmode = 0; /* don't hang up */
         }
      }
      else {
         logMsg(LEVEL4, "%s abnormal exit\n", hupname );
         rethupmode = 0; /* don't hang up */
      }
   }
   logMsg(LEVEL4, "End: Executing %s [%s]\n", hupname, strdate(ONLYTIME));

   return rethupmode;
}


/*
 * Convert hex string to unsigned characters
 */
void hex2uchar(const char * buf, unsigned char * ucbuf)
{
    int buf_len = strlen(buf)/2, i;
    char tmpchar[5];

    bzero(tmpchar, sizeof(tmpchar));
    for (i = 0; i < buf_len; i++)
    {
        strncpy(tmpchar, buf + 2 * i, 2);
        ucbuf[i] = (unsigned char) strtol(tmpchar, NULL, 16);
    }
}

/*
 * 2's complement 8-bit checksum
 * Modified from: https://stackoverflow.com/questions/3463976/c-file-checksum
 */
unsigned char checksum (unsigned char *ptr, size_t size) {
  unsigned int sum = 0;
  while (size-- != 0)
  {
    sum += *ptr++;
  }
  sum = (sum ^ 0xFF) + 1;
  return (unsigned char)sum;
}

/*
 * Send the call line to clients
 *
 * Format of a CALL line passed to TCP/IP clients by ncidd:
 *      where CALL is one of: BLK CID HUP MWI OUT PID PUT RID WID
 *
 *      CALL: *DATE*mmddyyyy*TIME*hhmm*LINE*<id>*NMBR*number*MESG*NONE*FNMBR*<formatted number>*NTYPE*FIX/CELL*CTRY*<country code>*LOCA*<location>*CARI*<carrier>*NAME*<name>*
 *
 * Format of a MSG or NOT line passed to TCP/IP clients by ncidd:
 *
 *      MSG: <mesg> ***DATE*<date>*TIME*<time>*NAME*<name>*NMBR*<nmbr>*LINE*<line>*MTYPE*<type>*
 */
void sendCall(const char * linelabel, const struct cid * scid, const char * nameptr)
{
    char cidbuf[BUFSIZ];

    sprintf(cidbuf, "%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
        linelabel,
        DATE, scid->ciddate,
        TIME, scid->cidtime,
        LINE, scid->cidline,
        NMBR, scid->cidnmbr,
        MESG, scid->cidmesg,
              scid->infoBuffer?scid->infoBuffer:"",
        NAME, nameptr,
        STAR);

    /* Log the CID, OUT, HUP, etc. text line */
    writeLog(cidlog, cidbuf);

    /*
     * Send the CID, OUT, HUP, etc. text line to clients
     */
    writeClients(cidbuf);

}


/*
 * called from safesend disconnectClient to finish up anything
 * that is specific to ncidd
 */
void finishDisconnect(int pos)
{
   /* remove the client from the IPinfo tables and turn off the ack flag */
   if (pos >= 0 && pos < MAXCONNECT)
   {
      ack[pos] = 0;
      if (IPinfo[pos].dialout == 1)
      {
         if (dialOUT > 0) --dialOUT;
         IPinfo[pos].dialout = 0;
      }
      IPinfo[pos].addr[0] = '\0';
      IPinfo[pos].name[0] = '\0';
      IPinfo[pos].ident[0] = '\0';
   }
}


/*
 * Check if MESG is a hexadecimal number
 * http://ncid.sourceforge.net/doc/NCID-API.html#modem-mesg-hex
 *
 * Returns: 0 - if all characters are hexadecimal
 *          1 - if any character is not hexadecimal
 */

int nothex(const char * mesg)
{
    unsigned int pos;
    int ret = 0;

    for (pos = 0; pos < strlen(mesg); ++pos)
    {
        if (!(ret = isxdigit((int)mesg[pos]))) break;
    }

    return !ret;
}

/*
 * remove whitespace from the start and end of a string
 */
char *trimWhitespace(char *sptr, char *eptr)
{
    while (*sptr && isspace((int) *sptr)) ++sptr;
    while (*eptr && isspace((int) *eptr)) *eptr-- = '\0';

return sptr;
}

/*
 * Send string to all TCP/IP CID clients.
 */

void writeClients(const char * inbuf)
{
    int pos, ret;
    char buf[BUFSIZ];

    (void) ret;


    logMsg(LEVEL6, "writeClients: %s\n", inbuf);

    strcat(strcpy(buf, inbuf), CRLF);
    for (pos = 0; pos < MAXCONNECT; ++pos)
    {
        if (polld[pos].fd <= 0 || polld[pos].fd == ttyfd ||
            polld[pos].fd == mainsock)
        {
            /* empty polld slots, the modem and listen port are not clients */
            continue;
        }

        logMsg(LEVEL6, "Sending data to Client %d pos %d at %s%s %s\n",
            polld[pos].fd, pos, IPinfo[pos].addr, IPinfo[pos].name, strdate(WITHSEP));

        safeSend(pos, buf);
    }
}

/*
 * Send log, if log file exists and is under cidlogmax bytes.
 */

void sendLog(int pos)
{
  struct stat statbuf;
  int linecount;
  const char **ptr;
  char *iptr, *optr, input[BUFSIZ], logbuf[BUFSIZ], msgbuf[BUFSIZ];
  FILE *fp;

  if (stat(cidlog, &statbuf) == 0)
  {
    if ((long unsigned int) statbuf.st_size > cidlogmax)
    {
      /* don't send a really big cidlog (default 110000 bytes) */
      logMsg(LEVEL1, LOG2BIG, (long unsigned int) statbuf.st_size,
              cidlogmax, strdate(WITHSEP), NL);
      sprintf(logbuf, LOG2BIG, (long unsigned int) statbuf.st_size,
              cidlogmax, strdate(WITHSEP), CRLF);
      safeSend(pos, logbuf);
      logMsg(LEVEL1, "%s%s", NOLOGSENT, NL);
      sprintf(msgbuf, "%s%s", NOLOGSENT, CRLF);
      safeSend(pos, msgbuf);
      /* return early */
      return;
    }
  }

    if ((fp = fopen(cidlog, "r")) == NULL)
    {
      logMsg(LEVEL6, "cidlog: %d %s [%s]\n", errno, strerror(errno), strdate(ONLYTIME));
      logMsg(LEVEL1, "%s: %s%s", NOLOG, cidlog, NL);
      sprintf(msgbuf, "%s%s", NOLOG, CRLF);
      safeSend(pos, msgbuf);
    }
    else
    {
      /*
       * read each line of file, one line at a time
       * add "LOG" to line tag (CID: becomes CIDLOG:)
       * send line to clients
       */
      iptr = 0;
      linecount = 0;
      logMsg(LEVEL4, "%sBegin: Send call log: %s [%s]\n", PRETTYINDENT, cidlog, strdate(ONLYTIME));

       /* Indicate start of Call Log */
       logMsg(LEVEL2, "%s%s%s", PRETTYINDENT, LOGSTART, NL);
       sprintf(msgbuf, "%s%s", LOGSTART, CRLF);
       safeSend(pos, msgbuf);

      while (fgets(input, BUFSIZ - sizeof(LINETYPE), fp) != NULL)
      {
        /* strip <CR> and <LF> */
        if ((iptr = strchr( input, '\r')) != NULL) *iptr = '\0';
        if ((iptr = strchr( input, '\n')) != NULL) *iptr = '\0';

        optr = logbuf;
        iptr = input;
        if (strstr(input, ": ") != NULL)
        {
          /* possible line tag found */
          for (ptr = lineTags; *ptr; ++ptr)
          {
            if (!strncmp(input, *ptr, strlen(*ptr)))
            {
              /* copy line tag, skip ": " */
              for (iptr = input; *iptr != ':';) *optr++ = *iptr++;
              iptr += 2;
              break;
            }
          }
        }
        /*
         * if line "<label>: " found, line begins with "<label>LOG: "
         * if line label not found, line begins with "LOG: "
         */
        strcat(strncat(strcpy(optr, LOGLINE), iptr, BUFSIZ - (iptr - input - 1)), CRLF);
        safeSend(pos, logbuf);
        linecount++;

        optr = logbuf;
      }

      (void) fclose(fp);

        /* Determine if a Call Log was sent */
        if (iptr)
        {
          /* Indicate end of the Call Log */
          logMsg(LEVEL2, "%sSent call log: %s %d lines\n%s%s\n", PRETTYINDENT, cidlog, linecount, PRETTYINDENT, LOGEND);
          sprintf(msgbuf, "%s%s", LOGEND, CRLF);
          safeSend(pos, msgbuf);
        }
        else
        {
          logMsg(LEVEL2, "%s%s: %s%s", PRETTYINDENT, EMPTYLOG, cidlog, NL);
          sprintf(msgbuf, "%s%s", EMPTYLOG, CRLF);
          safeSend(pos, msgbuf);
        }

    }

  logMsg(LEVEL4, "%sEnd: Send call log: %s [%s]\n", PRETTYINDENT, cidlog, strdate(ONLYTIME));

}

/*
 * Write log, if logfile exists.
 */

void writeLog(const char * logf, const char * logbuf)
{
    int logfd, ret;
    char msgbuf[BUFSIZ];

    (void) ret;

    /* write to server log */
    logMsg(LEVEL3, "%s\n", logbuf);

    if ((logfd = open(logf, O_WRONLY | O_APPEND)) < 0)
    {
        logMsg(LEVEL6, "%s: %s\n", logf, strerror(errno));
    }
    else
    {
        /* write log entry */
        sprintf(msgbuf, "%s\n", logbuf);
        ret = write(logfd, msgbuf, strlen(msgbuf));
        close(logfd);
    }
}

/*
 * Send call information
 *
 * Format of CIDINFO line passed to TCP/IP clients by ncidd:
 *
 * CIDINFO: *LINE*<label>*RING*<number>*TIME<hh:mm:ss>
 */

void sendInfo(int ring, const char * lineid)
{
    char buf[BUFSIZ];
    char infoline[CIDSIZE];

    if (!sendinfo) return;

    strcpy(infoline, lineid);

    userAlias(0, 0, infoline);
    sprintf(buf, "%s%s%s%s%d%s%s%s",CIDINFO, LINE, infoline, \
            RING, ring, TIME, strdate(ONLYTIME), STAR);

    logMsg(LEVEL3, "%s\n", buf);
    writeClients(buf);
}

/*
 * Check if lockfile present and has an active process number in it
 * ret == 0 if lockfile is not present or lockfile has stale process number
 * ret == 1 if lockfile present and process number active, or error
 */

int CheckForLockfile(void)
{
    int kret, ret = 0, lockp;
    static unsigned int sentmsg = 0;
    FILE *fp;
    char lockbuf[BUFSIZ];
    struct stat statbuf;

    if (lockfile != 0)
    {
        if (stat(lockfile, &statbuf) == 0)
        {
            ret = 1;
            if ((fp = fopen(lockfile, "r")) == NULL)
            {
                if (!(sentmsg & 1))
                {
                    logMsg(LEVEL1, "%s: %s\n", lockfile, strerror(errno));
                    sentmsg |= 1;
                }
            }
            else
            {
                if (fgets(lockbuf, BUFSIZ - 1, fp) != NULL)
                {
                    lockbuf[BUFSIZ - 1] = '\0';
                    lockp = atoi(lockbuf);
                    if (lockp)
                    {
                        /* lockfile contains a process number */
                        kret = kill(lockp, 0);
                        if (kret && errno != EPERM)
                        {
                            /* the error is not permission denied */
                            if (unlink(lockfile))
                            {
                                if (!(sentmsg & 2))
                                {
                                    logMsg(LEVEL1,
                                        "Failed to remove stale lockfile: %s\n",
                                        lockfile);
                                    sentmsg |= 2;
                                }
                            }
                            else
                            {
                                logMsg(LEVEL1, "Removed stale lockfile: %s\n",
                                        lockfile);
                                ret = 0;
                            }
                        }
                    }
                }
                fclose(fp);
            }
        }
    }

    /* clear all locally latched failures if lockfile no longer present */
    if (ret == 0) sentmsg = 0;

    return ret;
}


/*
 * touch the TTY lockfile
 */
int makeLockfile(void)
{
    FILE *lockptr;

    if (CheckForLockfile()) return 1;

    /* Create TTY port lockfile */
    if ((lockptr = fopen(lockfile, "w")) == NULL)
    {
        logMsg(LEVELE, "%s: %s\n", lockfile, strerror(errno));
        return 1;
    }
    else
    {
        /* write process number to lockfile */
        fprintf(lockptr, "%d\n", getpid());

        fclose(lockptr);
    }

    return 0;
}

/*
 * remove the TTY lockfile
 */
int removeLockfile(void)
{
    if (unlink(lockfile))
    {
        logMsg(LEVELE, "Failed to remove my lockfile: %s\n", lockfile);
        return 1;
    }

    return 0;
}

/*
 * test the lockfile path by trying to make a lockfile
 */

int CheckLockfilePath(void)
{
    FILE *lockptr;

    /* Create TTY port lockfile */
    if ((lockptr = fopen(lockfile, "w")) == NULL)
    {
        logMsg(LEVELE, "%s: %s\n", lockfile, strerror(errno));
        return 1;
    }
    else
    {
        /* write process number to lockfile */
        fprintf(lockptr, "%d\n", getpid());

        fclose(lockptr);
    }
    /* remove TTY port lockfile */
    if (unlink(lockfile))
    {
        logMsg(LEVELE, "Failed to remove my lockfile: %s\n", lockfile);
        return 1;
    }

    return 0;
}

/*
 * The modem lockfile timer checks frequently to see if a lockfile
 * has been created, probably by minicom.
 * When the lockfile exists, the modem file descriptor is closed,
 * and when the lockfile has been removed, the modem is again added to the
 * poll table.
 *
 */
void doModemLockTimer(void)
{

   /* TTY port lockfile */
   if (CheckForLockfile())
   {
       if (!modem_locked)
       {
           /* lockfile just found */
           /* stop polling, but reserve the slot in polld[] */
           if (pollpos >= 0) reservePoll(pollpos);
           if (ttyfd >= 0) close(ttyfd);
           ttyfd = -1;
           logMsg(LEVEL1, "TTY in use: releasing modem %s\n",
               strdate(WITHSEP));
           modem_locked = 1;
           ringtimer = 0;
       }
   }
   else if (modem_locked)
   {
       /* lockfile just went away */
       logMsg(LEVEL1, "TTY free: using modem again %s\n",
           strdate(WITHSEP));
       modem_locked = 0;

       /* restore tty poll events */
       reAttachModem();
   }
}


/*
 * return a string for the call type
 */
const char * getLineLabel(int calltype)
{
    const char * linelabel;
    switch(calltype)
    {
        case BLK:
            linelabel = BLKLINE;
            break;
        case CID:
            linelabel = CIDLINE;
            break;
        case HUP:
            linelabel = HUPLINE;
            break;
        case MWI:
            linelabel = MWILINE;
            break;
        case OUT:
            linelabel = OUTLINE;
            break;
        case PID:
            linelabel = PIDLINE;
            break;
        case PUT:
            linelabel = PUTLINE;
            break;
        case RID:
            linelabel = RIDLINE;
            break;
        case WID:
            linelabel = WIDLINE;
            break;
        default: /* should not happen */
            linelabel = CIDLINE;
            break;
    }
    return linelabel;
}


/*
 * Close all file descriptors and restore tty parameters.
 */
void cleanup(void)
{
    int pos;

    if (ttyfd > 2)
    {
        /* restore tty parameters for serial device or modem */
        tcflush(ttyfd, TCIOFLUSH);
        tcsetattr(ttyfd, TCSANOW, &otty);

        /*
         * send hangup to the modem if hangup option given
         * this makes sure modem hangs up if off-hook when ncidd dies
         */
        if (hangup || hupmode) (void) initModem(HANGUP, 0);
    }

    /* close open files and sockets */
    for (pos = 0; pos < MAXCONNECT; ++pos)
        if (polld[pos].fd > 0) close(polld[pos].fd);

    /* remove pid file, if it was created */
    if (pid)
    {
        unlink(pidfile);
        logMsg(LEVEL1, "Removed pidfile: %s\n", pidfile);
    }

    /* close log file, if open */
    if (logptr) fclose(logptr);
}

/*
 * exit on a SIGINT, SIGQUIT, SIGABRT, SIGTERM signal
 */

void TERMhandler(int sig)
{
   /* set a flag so ncidd can close everything neatly before leaving */
   TERM_flag = sig;
}

void TERMexit(void)
{
   /* got signal SIGTERM via TERMhandler, clean up and leave */
   logMsg(LEVELE, "Received Signal %d: %s\nTerminated: %s\n",
           TERM_flag, strsignal(TERM_flag), strdate(WITHSEP));
   /* close open files and sockets */
   cleanup();
   exit(0);
}

/*
 * exit on a SIGSEGV signal
 * do minimal clean up after a memory exception
 */

void SEGVhandler(int sig)
{
    /* restore tty parameters */
    if (ttyfd > 2)
    {
        tcflush(ttyfd, TCIOFLUSH);
        tcsetattr(ttyfd, TCSANOW, &otty);
    }

    /* remove pid file, if it was created */
    if (pid) {
        unlink(pidfile);
    }

    /* and exit */
    _exit(sig);
}

/*
 * reload alias, blacklist and whitelist files
 */
void reload(void)
{
    logMsg(LEVEL1, "Reloading alias, blacklist and whitelist files [%s]\n",
      strdate(WITHSEP));

    /*
     * Decided not to do a reconfig because it seems like too much work
     * for a small gain:
     *   - open configurable files need to be closed and opened again
     *   - configuration verbose indicators, need to be output again
     * if (doConf()) errorExit(-104, 0, 0);
     */

    /* remove existing aliases to free memory used */
    rmaliases();

    /* reload alias file, but quit on error */
    if (readAliasFile()) errorExit(-109, 0, 0);

    /* remove existing whitelist, blacklist entries to free memory used */
    rmBWlists();

    /* reload whitelist, blacklist entries, but quit on error */
    readBWlists();

    /* time of last ABW reads */
    abw_lastread  = time(0);
    checkABWtimer = time(0) + ABWCHECK;

    logMsg(LEVEL1,
      "End: Reloaded alias, blacklist and whitelist files [%s]\n", strdate(ONLYTIME));

}

/*
 * show connected clients
 */
void showConnected(void)
{
    int pos;
    int cnt = 0;

    logMsg(LEVELR, "Received Signal: show connected at %s\n", strdate(WITHSEP));

    for (pos = 0; pos < MAXCONNECT; ++pos)
    {
        if (polld[pos].fd <= 0 || polld[pos].fd == ttyfd ||
            polld[pos].fd == mainsock)
        {
            /* empty polld slots, the modem and listen port are not clients */
            continue;
        }

        logMsg(LEVELR, "Client %03d pos %03d from %s%s %s {%s} is connected\n",
            polld[pos].fd, pos, IPinfo[pos].addr, IPinfo[pos].name,
            IPinfo[pos].dialout ? "dialout" : "nodialout", IPinfo[pos].ident);
        cnt++;
    }
    if (cnt == 0)
    {
        logMsg(LEVELR, "There are no clients connected at this time\n" );
    }
}

/*
 * check the dates on the ncidd.alias, blacklist and whitelist files
 * to see if something like a manual edit has changed them outside of ncidd
 * if so, reload the ABW caches
 */
void doCheckABWTimer(void)
{
   if (checkABWtimer < time(0)) {
      if (check_ABW_dates(whitelist) ||
          check_ABW_dates(blacklist) ||
          check_ABW_dates(cidalias))
      {
          reload();
          return;
      }
      else {
          /* start the timer again */
          checkABWtimer = time(0) + ABWCHECK;
      }
   }
}


/*
 * return true if an ABW file is newer than the last ABW reload
 * and has been stable for half of the ABWCHECK time
 */
int check_ABW_dates(const char * filename)
{
   struct stat statbuf;

   if (stat(filename, &statbuf) == 0) {
      /* logMsg(LEVEL4, "%s ABW check %d\n", filename, abw_lastread); */
      if ((abw_lastread < statbuf.st_mtime) && (time(0) - statbuf.st_mtime > (ABWCHECK/2))) {
         logMsg(LEVEL4, "%s file changed since %d: size %d timestamp %d\n", filename, abw_lastread, statbuf.st_size, statbuf.st_mtime);
         return 1;
      }
   }
   else {
      /* the file is not readable */
      logMsg(LEVEL1, "%s: %d %s [%s]\n", filename, errno, strerror(errno), strdate(ONLYTIME));
   }
   return 0;
}

/*
 * USRhandler
 */

void    USRhandler(int sig)
{
   if ( sig == SIGHUP ) {
      /* SIGHUP reloads the alias, whitelist, blacklist */
      HUP_flag = 1;
   }
   else if ( sig == SIGUSR1 ) {
      /* SIGUSR1 updates the cidcall log by replacing cidcall.log file with cidcall.log.new*/
      USR1_flag = 1;
   }
   else if ( sig == SIGUSR2 ) {
      /* SIGUSR2 lists the connected clients */
      USR2_flag = 1;
   }
   else if ( sig == SIGALRM ) {
      /* SIGALRM when a read or write went too long */
      logMsg(LEVEL6, "Received Signal %d: %s\nIgnored: %s\n", sig, strsignal(sig), strdate(WITHSEP));
   }
   else {
      /* well that's odd */
      logMsg(LEVELE, "Received unexpected Signal %d: %s\nIgnored: %s\n", sig, strsignal(sig), strdate(WITHSEP));
   }
}

/*
 * set up signal handlers for HUP, USR1, USR2, INT, etc
 */

void setup_signal_handlers(void)
{
   /* SIGHUP reloads the alias, whitelist, blacklist */
   /* SIGUSR1 updates the cidcall log */
   /* SIGUSR2 lists the connected clients */
   HUP_flag = 0;
   USR1_flag = 0;
   USR2_flag = 0;
   static struct sigaction user_sigaction;
   user_sigaction.sa_handler = USRhandler;
   sigfillset( &user_sigaction.sa_mask);
   user_sigaction.sa_flags = 0;
   sigaction(SIGHUP, &user_sigaction, NULL);
   sigaction(SIGUSR1, &user_sigaction, NULL);
   sigaction(SIGUSR2, &user_sigaction, NULL);
   sigaction(SIGALRM, &user_sigaction, NULL);

   /* SIGINT, etc, will stop ncidd */
   TERM_flag = 0;
   static struct sigaction leave_sigaction;
   leave_sigaction.sa_handler = TERMhandler;
   sigfillset( &leave_sigaction.sa_mask);
   leave_sigaction.sa_flags = 0;
   sigaction(SIGINT, &leave_sigaction, NULL);    /* Interrupt from keyboard */
   sigaction(SIGQUIT, &leave_sigaction, NULL);    /* Quit from keyboard */
   sigaction(SIGABRT, &leave_sigaction, NULL);    /* Abort signal from abort(3) */
   sigaction(SIGTERM, &leave_sigaction, NULL);    /* Termination signal */

   /* SIGSEGV, will try to remove the lock file */
   static struct sigaction segv_sigaction;
   segv_sigaction.sa_handler = SEGVhandler;
   sigfillset( &segv_sigaction.sa_mask);
   segv_sigaction.sa_flags = 0;
   sigaction(SIGSEGV, &segv_sigaction, NULL);    /* Invalid memory reference */

   /* ignore a bunch of them */
   static struct sigaction ignore_sigaction;
   ignore_sigaction.sa_handler = SIG_IGN;
   sigemptyset( &ignore_sigaction.sa_mask);
   ignore_sigaction.sa_flags = 0;
   sigaction(SIGTSTP, &ignore_sigaction, NULL);
   sigaction(SIGPIPE, &ignore_sigaction, NULL);    /* Broken pipe: write to pipe with no readers */
   sigaction(SIGTTIN, &ignore_sigaction, NULL);
   sigaction(SIGTTOU, &ignore_sigaction, NULL);
   sigaction(SIGIO, &ignore_sigaction, NULL);    /* io is waiting */
   sigaction(SIGWINCH, &ignore_sigaction, NULL); /* Window resize signal */

   /* set these back to the default */
   static struct sigaction default_sigaction;
   default_sigaction.sa_handler = SIG_DFL;
   sigfillset( &default_sigaction.sa_mask);
   default_sigaction.sa_flags = 0;
   /* sigaction(SIGSEGV, &default_sigaction, NULL);*/
   sigaction(SIGXCPU, &default_sigaction, NULL);
   sigaction(SIGXFSZ, &default_sigaction, NULL);
}

/*
 * exit after some fatal error
 */

void errorExit(int error, const char * msg, const char * arg)
{
    if (error == -1)
    {
        /* should not happen */
        if (msg == 0) msg = "oops";

        /*
         * system error
         * print msg, arg should be zero
         */
        error = errno;
        logMsg(LEVELE, "%s: %s\n", msg, strerror(errno));
    }
    else
    {
        /* should not happen */
        if (msg != 0 && arg == 0) arg = "oops";

        /*
         * internal program error
         * print msg and arg if both are not 0
         */
        if (msg != 0 && arg != 0)
        {
            logMsg(LEVELE, "%s: %s\n", msg, arg);
        }
    }

    /* do not print terminated message, or cleanup, if option error */
    if (error != -100 && error != -101 && error != -106 && error != -107 &&
        error != -108 && error != -113)
    {
        logMsg(LEVELE, "Terminated:  %s\n", strdate(WITHSEP));
        cleanup();
    }

    exit(error);
}

/*
 * Normal exit
 */

void normalExit(void)
{
    logMsg(LEVEL1, "Terminated by Verbose Level 8:  %s\n", strdate(WITHSEP));
    cleanup();
    exit(0);
}

/*
 * Tell clients about a server issue
 */
void sysMesg(const char * message, const char * sysline)
{
    char sysmessage[BUFSIZ];
    char sysdate[CIDSIZE];
    char systime[CIDSIZE];
    char * tptr;
    const char * sysname = "ncidd";
    const char * sysnmbr = NONMBR;
    const char * systype = "SYS";

    tptr = strdate(NOSEP);
    strncpy(sysdate, tptr, 8);
    sysdate[8] = '\0';
    strncpy(systime, tptr + 9, 4);
    systime[4] = '\0';

    sprintf(sysmessage, SYSMSG, message, sysdate, systime, sysnmbr,
            sysline, systype, sysname);
    writeLog(cidlog, sysmessage);
    writeClients(sysmessage);
}

/*
 * turn polling off and reserve the slot in the table
 */
void reservePoll(int pos)
{
   /* remove TTY poll events */
   pollevents = polld[pos].events;
   polld[pos].events = polld[pos].revents = 0;

   /* addPoll will not reuse this slot */
   polld[pos].fd = -111;
}

/*
 * turn polling on for the reserved slot in the table
 */
void restorePoll(int pos, int fd)
{
   polld[pos].fd = fd;
   polld[pos].events = pollevents;
}
